<ul class="social-media-list">
    {%- if site.dribbble_username -%}
    <li>
        <a href="https://dribbble.com/{{ site.dribbble_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#dribbble' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.dribbble_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.facebook_username -%}
    <li>
        <a href="https://www.facebook.com/{{ site.facebook_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#facebook' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.facebook_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.flickr_username -%}
    <li>
        <a href="https://www.flickr.com/photos/{{ site.flickr_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#flickr' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.flickr_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.github_username -%}
    <li>
        <a href="https://github.com/{{ site.github_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#github' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.github_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.github -%}
    <li><a href="https://github.com/{{ site.github }}">
        <svg class="svg-icon">
            <use xlink:href="{{ '/assets/minima-social-icons.svg#github' | relative_url }}"></use>
        </svg>
        <span class="username">{{ site.github | escape }}</span></a></li>
    {%- endif -%}
    {%- if site.instagram_username -%}
    <li>
        <a href="https://instagram.com/{{ site.instagram_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#instagram' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.instagram_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.linkedin_username -%}
    <li>
        <a href="https://www.linkedin.com/in/{{ site.linkedin_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#linkedin' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.linkedin_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.pinterest_username -%}
    <li>
        <a href="https://www.pinterest.com/{{ site.pinterest_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#pinterest' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.pinterest_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- for mst in site.mastodon -%}{%- if mst.username and mst.instance -%}
    <li>
        <a href="https://{{ mst.instance| cgi_escape | escape}}/@{{mst.username}}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#mastodon' | relative_url }}"></use>
            </svg>
            <span class="username">{{ mst.username|escape }}</span></a></li>
    {%- endif -%}{%- endfor -%}
    {%- if site.twitter_username -%}
    <li>
        <a href="https://www.twitter.com/{{ site.twitter_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#twitter' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.twitter_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.youtube_username -%}
    <li>
        <a href="https://youtube.com/{{ site.youtube_username| cgi_escape | escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#youtube' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.youtube_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.googleplus_username -%}
    <li>
        <a href="https://plus.google.com/{{ site.googleplus_username| escape }}">
            <svg class="svg-icon">
                <use xlink:href="{{ '/assets/minima-social-icons.svg#googleplus' | relative_url }}"></use>
            </svg>
            <span class="username">{{ site.googleplus_username| escape }}</span></a>
    </li>
    {%- endif -%}
    {%- if site.rss -%}
    <li><a href="{{ 'feed.xml' | relative_url }}">
        <svg class="svg-icon">
            <use xlink:href="{{ '/assets/minima-social-icons.svg#rss' | relative_url }}"></use>
        </svg>
        <span>{{ site.rss | escape }}</span></a></li>
    {%- endif -%}
    <li><a href="https://discord.gg/erBPnKwz7R">
        <img style="height: 1.75em"
             src="https://assets-global.website-files.com/6257adef93867e50d84d30e2/62594fdd4b264a31eafa1e5d_92ad040ed5143bfb541ea61f5c3bb18f.svg"/>
    </a></li>
</ul>

<script data-goatcounter="https://lnav.goatcounter.com/count"
        async src="//gc.zgo.at/count.js"></script>
