---
layout: default
---

<div class="post">

    <header class="post-header">
        <h1 class="post-title">{{ page.title }}</h1>
    </header>

    <article class="post-content">
        {{ content }}
    </article>

    <script src="/assets/js/codeblock.js"></script>

</div>
