/* override table width restrictions */
@media screen and (min-width: 767px) {

    .wy-table-responsive table td {
        /* !important prevents the common CSS stylesheets from overriding
           this as on RTD they are loaded after this stylesheet */
        white-space: normal !important;
    }

}

th p {
    margin-bottom: 0px;
}

.wy-nav-content {
    max-width: 900px;
    overflow-x: auto;
}

table.query-results p {
    font-size: 0.9em !important;
}

DL DT:target, :target > H2, :target > H3, span:target + H2, span:target + H3 {
    border: 0 !important;
    border-bottom: 1px solid #d3d381 !important;
    background: #ffc !important;
    /* padding: 1em !important;*/
}

DL DT {
    border: 0 !important;
    background: inherit !important;
    border-bottom: 1px solid #c3c0ee !important;
    /* display: block !important; */
}

kbd {
    padding: 0.1em 0.6em;
    border: 1px solid #ccc;
    font-size: 11px;
    font-family: Arial, Helvetica, sans-serif;
    background-color: #f7f7f7;
    color: #333;
    -moz-box-shadow: 0 1px 0px rgba(0, 0, 0, 0.2), 0 0 0 2px #ffffff inset;
    -webkit-box-shadow: 0 1px 0px rgba(0, 0, 0, 0.2), 0 0 0 2px #ffffff inset;
    box-shadow: 0 1px 0px rgba(0, 0, 0, 0.2), 0 0 0 2px #ffffff inset;
    -moz-border-radius: 3px;
    -webkit-border-radius: 3px;
    border-radius: 3px;
    display: inline-block;
    margin: 0 0.1em 0.1em 0.1em;
    text-shadow: 0 1px 0 #fff;
    line-height: 1.4;
    white-space: nowrap;
}
