/* Generated by re2c 3.0 on Mon Sep 19 01:36:56 2022 */
#line 1 "../../lnav/src/data_scanner_re.re"
/**
 * Copyright (c) 2015, Timothy Stack
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * * Neither the name of Timothy Stack nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include "base/date_time_scanner.hh"
#include "config.h"
#include "data_scanner.hh"

nonstd::optional<data_scanner::tokenize_result> data_scanner::tokenize2()
{
    data_token_t token_out = DT_INVALID;
    capture_t cap_all;
    capture_t cap_inner;
#   define YYCTYPE unsigned char
#   define CAPTURE(tok) { \
        if (YYCURSOR.val == EMPTY) { \
            this->ds_next_offset = this->ds_input.length(); \
        } else { \
            this->ds_next_offset = YYCURSOR.val - this->ds_input.udata(); \
        } \
        cap_all.c_end = this->ds_next_offset; \
        cap_inner.c_end = this->ds_next_offset; \
        token_out = tok; \
    }

#   define RET(tok) { \
        CAPTURE(tok); \
        return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()}; \
    }
    static const unsigned char *EMPTY = (const unsigned char *) "";

    struct _YYCURSOR {
        YYCTYPE operator*() const {
            if (this->val < this->lim) {
                return *val;
            }
            return '\0';
        }

        operator const YYCTYPE *() const {
            if (this->val < this->lim) {
                return this->val;
            }
            return EMPTY;
        }

        const YYCTYPE *operator=(const YYCTYPE *rhs) {
            this->val = rhs;
            return rhs;
        }

        const YYCTYPE *operator+(int rhs) {
            return this->val + rhs;
        }

        const _YYCURSOR *operator-=(int rhs) {
            this->val -= rhs;
            return this;
        }

        _YYCURSOR& operator++() {
            this->val += 1;
            return *this;
        }

        const YYCTYPE *val{nullptr};
        const YYCTYPE *lim{nullptr};
    } YYCURSOR;
    YYCURSOR = (const unsigned char *) this->ds_input.udata() + this->ds_next_offset;
    _YYCURSOR yyt1;
    _YYCURSOR yyt2;
    _YYCURSOR yyt3;
    _YYCURSOR yyt4;
    const YYCTYPE *YYLIMIT = (const unsigned char *) this->ds_input.end();
    const YYCTYPE *YYMARKER = YYCURSOR;

    YYCURSOR.lim = YYLIMIT;

    cap_all.c_begin = this->ds_next_offset;
    cap_all.c_end = this->ds_next_offset;
    cap_inner.c_begin = this->ds_next_offset;
    cap_inner.c_end = this->ds_next_offset;

    
#line 117 "../../lnav/src/data_scanner_re.cc"
{
	YYCTYPE yych;
	unsigned int yyaccept = 0;
	static const unsigned char yybm[] = {
		/* table 1 .. 8: 0 */
		  0, 231, 231, 231, 231, 231, 231, 231, 
		231, 239, 231, 231, 231, 239, 231, 231, 
		231, 231, 231, 231, 231, 231, 231, 231, 
		231, 231, 231, 231, 231, 231, 231, 231, 
		239, 231,  34, 231, 231, 231, 231,  36, 
		231, 231, 231, 231, 231, 175, 231, 183, 
		175, 175, 175, 175, 175, 175, 175, 175, 
		175, 175, 175, 231, 231, 231, 128, 183, 
		231, 175, 175, 175, 175, 175, 175, 175, 
		175, 175, 175, 175, 175, 175, 175, 175, 
		175, 175, 175, 175, 175, 175, 175, 175, 
		175, 175, 175, 231,   1, 231, 231, 175, 
		231, 175, 175, 175, 175, 175, 175, 175, 
		175, 175, 175, 175, 175, 175, 175, 175, 
		175, 175, 175, 175, 175, 175, 175, 175, 
		175, 175, 175, 231, 231, 231, 231, 231, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 9 .. 16: 256 */
		  0, 249, 249, 249, 249, 249, 249, 249, 
		249, 248, 248, 249, 249, 248, 249, 249, 
		249, 249, 249, 249, 249, 249, 249, 249, 
		249, 249, 249, 248, 249, 249, 249, 249, 
		248, 249, 176, 249, 249, 252, 252, 104, 
		248, 248, 249, 249, 249,  28, 249,  28, 
		 28,  28,  28,  28,  28,  28,  28,  28, 
		 28,  28,  25, 249, 249, 252,  25,  28, 
		249,  30,  30,  30,  30,  30,  30,  30, 
		 30,  30,  30,  30,  30,  30,  30,  30, 
		 30,  30,  30,  30,  30,  30,  30,  30, 
		 30,  30,  30, 248,  33, 248, 249,  25, 
		249,  30,  30,  30,  30,  30,  30,  30, 
		 30,  30,  30,  30,  30,  30,  30,  30, 
		 30,  30,  30,  30,  30,  30,  30,  30, 
		 30,  30,  30, 248, 249, 248, 249, 249, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 17 .. 24: 512 */
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  1,   0,   0,   0,   0,   0,   0,  32, 
		  0,   0,   0,   0,   0,  19,  64,   0, 
		159, 159, 159, 159, 159, 159, 159, 159, 
		155, 155,  16,   0,   0,   0,   0,   0, 
		  0, 155, 155, 155, 155, 155, 155, 147, 
		147, 147, 147, 147, 147, 147, 147, 147, 
		147, 147, 147, 147, 147, 147, 147, 147, 
		147, 147, 147,   0,   0,   0,   0, 145, 
		  0, 154, 154, 154, 154, 154, 154, 146, 
		146, 146, 146, 146, 146, 146, 146, 146, 
		146, 146, 146, 146, 146, 146, 146, 146, 
		146, 146, 146,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 25 .. 32: 768 */
		  0,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   0,   2,   2,   2,   2, 
		 34,  10,   2,  10,  26,  11,  10,   0, 
		 10,  10,  10,   3,   2,  27,  27,  26, 
		 95,  95,  95,  95,  95,  95,  95,  95, 
		 95,  95,   2,   2,   2,   2,   2,   2, 
		 26,  91,  91,  91,  91,  91,  91,  91, 
		 91,  91,  91,  91,  91,  91,  91,  91, 
		 91,  91,  91,  91,  91,  91,  91,  91, 
		 91,  91,  91,   2,   8,   2,  10,  27, 
		  2, 219, 219, 219, 219, 219, 219, 219, 
		219, 219, 219, 219, 219, 219, 219, 219, 
		219, 219, 219, 219, 219, 219, 219, 219, 
		219, 219, 219,   2,   2,   2,  10,   2, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 33 .. 37: 1024 */
		  0, 136, 136, 136, 136, 136, 136, 136, 
		136, 160, 128, 136, 136, 160, 136, 136, 
		136, 136, 136, 136, 136, 136, 136, 136, 
		136, 136, 136,   0, 136, 136, 136, 136, 
		160, 128,   0, 128, 136, 128, 128, 128, 
		128, 128, 128, 128, 128, 152, 152, 136, 
		216, 216, 216, 216, 216, 216, 216, 216, 
		216, 216, 128, 192, 128, 192, 128, 192, 
		136, 152, 152, 152, 152, 152, 152, 152, 
		152, 152, 152, 152, 152, 152, 152, 152, 
		152, 152, 152, 152, 152, 152, 152, 152, 
		152, 152, 152, 128,   0, 128, 128, 152, 
		128, 152, 152, 152, 152, 152, 152, 152, 
		152, 152, 152, 152, 152, 152, 152, 152, 
		152, 152, 152, 152, 152, 152, 152, 152, 
		152, 152, 152, 128, 128, 128, 128, 136, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
	};
	yych = *YYCURSOR;
	if (yych <= 'E') {
		if (yych <= '(') {
			if (yych <= 0x1F) {
				if (yych <= '\n') {
					if (yych <= 0x00) goto yy2;
					if (yych <= 0x08) goto yy3;
					if (yych <= '\t') goto yy6;
					goto yy8;
				} else {
					if (yych <= '\r') {
						if (yych <= '\f') goto yy3;
						goto yy10;
					} else {
						if (yych == 0x1B) goto yy11;
						goto yy3;
					}
				}
			} else {
				if (yych <= '#') {
					if (yych <= ' ') goto yy6;
					if (yych == '"') goto yy14;
					goto yy13;
				} else {
					if (yych <= '%') {
						if (yych <= '$') goto yy3;
						goto yy15;
					} else {
						if (yych <= '&') goto yy13;
						if (yych <= '\'') goto yy16;
						goto yy17;
					}
				}
			}
		} else {
			if (yych <= '1') {
				if (yych <= ',') {
					if (yych <= ')') goto yy18;
					if (yych <= '*') goto yy13;
					if (yych <= '+') goto yy15;
					goto yy19;
				} else {
					if (yych <= '.') {
						if (yych <= '-') goto yy20;
						goto yy21;
					} else {
						if (yych <= '/') goto yy23;
						if (yych <= '0') goto yy25;
						goto yy27;
					}
				}
			} else {
				if (yych <= '<') {
					if (yych <= '9') {
						if (yych <= '2') goto yy28;
						goto yy29;
					} else {
						if (yych <= ':') goto yy30;
						if (yych <= ';') goto yy32;
						goto yy33;
					}
				} else {
					if (yych <= '>') {
						if (yych <= '=') goto yy35;
						goto yy36;
					} else {
						if (yych <= '?') goto yy13;
						if (yych <= '@') goto yy3;
						goto yy37;
					}
				}
			}
		}
	} else {
		if (yych <= 'n') {
			if (yych <= 'Z') {
				if (yych <= 'Q') {
					if (yych <= 'F') goto yy38;
					if (yych == 'N') goto yy40;
					goto yy39;
				} else {
					if (yych <= 'S') {
						if (yych <= 'R') goto yy41;
						goto yy39;
					} else {
						if (yych <= 'T') goto yy42;
						if (yych <= 'U') goto yy41;
						goto yy39;
					}
				}
			} else {
				if (yych <= '_') {
					if (yych <= '\\') {
						if (yych <= '[') goto yy43;
						goto yy44;
					} else {
						if (yych <= ']') goto yy45;
						if (yych <= '^') goto yy13;
						goto yy46;
					}
				} else {
					if (yych <= 'e') {
						if (yych <= '`') goto yy13;
						goto yy47;
					} else {
						if (yych <= 'f') goto yy48;
						if (yych <= 'm') goto yy49;
						goto yy50;
					}
				}
			}
		} else {
			if (yych <= '}') {
				if (yych <= 't') {
					if (yych == 'r') goto yy51;
					if (yych <= 's') goto yy49;
					goto yy52;
				} else {
					if (yych <= 'z') {
						if (yych <= 'u') goto yy53;
						goto yy49;
					} else {
						if (yych <= '{') goto yy54;
						if (yych <= '|') goto yy13;
						goto yy55;
					}
				}
			} else {
				if (yych <= 0xE0) {
					if (yych <= 0x7F) {
						if (yych <= '~') goto yy13;
						goto yy3;
					} else {
						if (yych <= 0xC1) goto yy1;
						if (yych <= 0xDF) goto yy56;
						goto yy57;
					}
				} else {
					if (yych <= 0xF0) {
						if (yych <= 0xEF) goto yy58;
						goto yy59;
					} else {
						if (yych <= 0xF3) goto yy60;
						if (yych <= 0xF4) goto yy61;
					}
				}
			}
		}
	}
yy1:
	YYCURSOR = YYMARKER;
	if (yyaccept <= 17) {
		if (yyaccept <= 8) {
			if (yyaccept <= 4) {
				if (yyaccept <= 2) {
					if (yyaccept <= 1) {
						if (yyaccept == 0) {
							goto yy5;
						} else {
							goto yy7;
						}
					} else {
						goto yy9;
					}
				} else {
					if (yyaccept == 3) {
						goto yy12;
					} else {
						goto yy22;
					}
				}
			} else {
				if (yyaccept <= 6) {
					if (yyaccept == 5) {
						goto yy24;
					} else {
						goto yy26;
					}
				} else {
					if (yyaccept == 7) {
						goto yy31;
					} else {
						goto yy34;
					}
				}
			}
		} else {
			if (yyaccept <= 13) {
				if (yyaccept <= 11) {
					if (yyaccept <= 10) {
						if (yyaccept == 9) {
							goto yy72;
						} else {
							goto yy106;
						}
					} else {
						goto yy109;
					}
				} else {
					if (yyaccept == 12) {
						goto yy113;
					} else {
						goto yy125;
					}
				}
			} else {
				if (yyaccept <= 15) {
					if (yyaccept == 14) {
						goto yy161;
					} else {
						goto yy222;
					}
				} else {
					if (yyaccept == 16) {
						yyt3 = yyt4;
						goto yy222;
					} else {
						goto yy199;
					}
				}
			}
		}
	} else {
		if (yyaccept <= 26) {
			if (yyaccept <= 22) {
				if (yyaccept <= 20) {
					if (yyaccept <= 19) {
						if (yyaccept == 18) {
							yyt2 = yyt1;
							goto yy199;
						} else {
							goto yy338;
						}
					} else {
						goto yy345;
					}
				} else {
					if (yyaccept == 21) {
						goto yy354;
					} else {
						goto yy371;
					}
				}
			} else {
				if (yyaccept <= 24) {
					if (yyaccept == 23) {
						goto yy428;
					} else {
						goto yy450;
					}
				} else {
					if (yyaccept == 25) {
						goto yy435;
					} else {
						goto yy217;
					}
				}
			}
		} else {
			if (yyaccept <= 31) {
				if (yyaccept <= 29) {
					if (yyaccept <= 28) {
						if (yyaccept == 27) {
							goto yy310;
						} else {
							goto yy315;
						}
					} else {
						goto yy643;
					}
				} else {
					if (yyaccept == 30) {
						goto yy654;
					} else {
						goto yy674;
					}
				}
			} else {
				if (yyaccept <= 33) {
					if (yyaccept == 32) {
						goto yy943;
					} else {
						goto yy971;
					}
				} else {
					if (yyaccept == 34) {
						goto yy1017;
					} else {
						goto yy1091;
					}
				}
			}
		}
	}
yy2:
	++YYCURSOR;
#line 141 "../../lnav/src/data_scanner_re.re"
	{ return nonstd::nullopt; }
#line 586 "../../lnav/src/data_scanner_re.cc"
yy3:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
yy4:
	if (yybm[1024+yych] & 8) {
		goto yy3;
	}
	if (yych <= 0xE0) {
		if (yych <= ':') {
			if (yych >= '-') goto yy62;
		} else {
			if (yych <= 0xC1) goto yy5;
			if (yych <= 0xDF) goto yy56;
			goto yy57;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy58;
			goto yy59;
		} else {
			if (yych <= 0xF3) goto yy60;
			if (yych <= 0xF4) goto yy61;
		}
	}
yy5:
#line 266 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_SYMBOL);
       }
#line 616 "../../lnav/src/data_scanner_re.cc"
yy6:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy64;
	if (yych <= '9') goto yy65;
	if (yych <= ':') goto yy66;
	goto yy64;
yy7:
#line 271 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_WHITE); }
#line 627 "../../lnav/src/data_scanner_re.cc"
yy8:
	yyaccept = 2;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy9;
	if (yych <= 'Z') goto yy67;
yy9:
#line 270 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LINE); }
#line 636 "../../lnav/src/data_scanner_re.cc"
yy10:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '\n') goto yy68;
		goto yy64;
	} else {
		if (yych <= '9') goto yy65;
		if (yych <= ':') goto yy66;
		goto yy64;
	}
yy11:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '[') goto yy69;
yy12:
#line 274 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_GARBAGE); }
#line 655 "../../lnav/src/data_scanner_re.cc"
yy13:
	++YYCURSOR;
	goto yy12;
yy14:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x1B) {
		if (yych <= 0x00) goto yy12;
		if (yych <= 0x1A) goto yy71;
		goto yy12;
	} else {
		if (yych <= 0x7F) goto yy71;
		if (yych <= 0xC1) goto yy12;
		if (yych <= 0xF4) goto yy71;
		goto yy12;
	}
yy15:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy81;
			goto yy12;
		} else {
			if (yych == ',') goto yy12;
			if (yych <= '.') goto yy81;
			goto yy12;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy81;
			if (yych <= '?') goto yy12;
			if (yych <= 'Z') goto yy81;
			goto yy12;
		} else {
			if (yych == '`') goto yy12;
			if (yych <= 'z') goto yy81;
			goto yy12;
		}
	}
yy16:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x1B) {
		if (yych <= 0x00) goto yy12;
		if (yych <= 0x1A) goto yy84;
		goto yy12;
	} else {
		if (yych <= 0x7F) goto yy84;
		if (yych <= 0xC1) goto yy12;
		if (yych <= 0xF4) goto yy84;
		goto yy12;
	}
yy17:
	yych = *++YYCURSOR;
	if (yych == ')') goto yy93;
#line 225 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LPAREN); }
#line 714 "../../lnav/src/data_scanner_re.cc"
yy18:
	++YYCURSOR;
#line 226 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RPAREN); }
#line 719 "../../lnav/src/data_scanner_re.cc"
yy19:
	++YYCURSOR;
#line 218 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_COMMA); }
#line 724 "../../lnav/src/data_scanner_re.cc"
yy20:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '0') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy94;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy95;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy46;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy21:
	yyaccept = 4;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy81;
			if (yych >= '+') goto yy81;
		} else {
			if (yych <= ',') goto yy22;
			if (yych <= '-') goto yy81;
			if (yych <= '.') goto yy97;
			goto yy98;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy81;
			if (yych <= '?') goto yy22;
			if (yych <= 'Z') goto yy81;
		} else {
			if (yych == '`') goto yy22;
			if (yych <= 'z') goto yy81;
		}
	}
yy22:
#line 272 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_DOT); }
#line 782 "../../lnav/src/data_scanner_re.cc"
yy23:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy4;
				if (yych >= '\v') goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych >= 0x0E) goto yy4;
				} else {
					if (yych <= 0x1B) goto yy24;
					if (yych <= 0x1F) goto yy4;
				}
			}
		} else {
			if (yych <= '$') {
				if (yych == '"') goto yy24;
				if (yych <= '#') goto yy99;
				goto yy101;
			} else {
				if (yych <= '\'') {
					if (yych <= '&') goto yy99;
				} else {
					if (yych <= '*') goto yy99;
					if (yych >= '-') goto yy101;
				}
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= 'P') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy24;
				if (yych <= 'O') goto yy101;
				goto yy102;
			} else {
				if (yych <= '[') {
					if (yych <= 'Z') goto yy101;
				} else {
					if (yych != ']') goto yy99;
				}
			}
		} else {
			if (yych <= '}') {
				if (yych == '`') goto yy24;
				if (yych <= 'z') goto yy101;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy99;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy24;
					if (yych <= 0xF4) goto yy4;
				}
			}
		}
	}
yy24:
#line 171 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_PATH); }
#line 846 "../../lnav/src/data_scanner_re.cc"
yy25:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= ':') {
		if (yych <= '$') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych >= 0x01) goto yy4;
				} else {
					if (yych <= '\n') goto yy26;
					if (yych <= '\f') goto yy4;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy4;
				} else {
					if (yych <= 0x1F) goto yy4;
					if (yych >= '$') goto yy4;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '*') {
					if (yych <= '%') goto yy105;
				} else {
					if (yych <= '+') goto yy80;
					if (yych >= '-') goto yy46;
				}
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy107;
					goto yy4;
				} else {
					if (yych <= '7') goto yy108;
					if (yych <= '9') goto yy110;
					goto yy111;
				}
			}
		}
	} else {
		if (yych <= 'd') {
			if (yych <= 'F') {
				if (yych <= '@') {
					if (yych >= '@') goto yy96;
				} else {
					if (yych == 'E') goto yy114;
					goto yy112;
				}
			} else {
				if (yych <= '^') {
					if (yych <= 'Z') goto yy115;
				} else {
					if (yych <= '_') goto yy46;
					if (yych >= 'a') goto yy112;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'f') {
					if (yych <= 'e') goto yy114;
					goto yy112;
				} else {
					if (yych == 'x') goto yy116;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych >= 0x7F) goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
				}
			}
		}
	}
yy26:
#line 257 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_NUMBER); }
#line 928 "../../lnav/src/data_scanner_re.cc"
yy27:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy107;
					if (yych <= '/') goto yy4;
					goto yy117;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy111;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy112;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy114;
					goto yy112;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy112;
				} else {
					if (yych <= 'e') goto yy114;
					if (yych <= 'f') goto yy112;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy28:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '5') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy26;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy107;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '4') goto yy117;
					goto yy118;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') {
					if (yych <= '9') goto yy110;
					goto yy111;
				} else {
					if (yych <= '?') goto yy26;
					if (yych <= '@') goto yy96;
					goto yy112;
				}
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy114;
					goto yy112;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy112;
				} else {
					if (yych <= 'e') goto yy114;
					if (yych <= 'f') goto yy112;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy29:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy107;
					if (yych <= '/') goto yy4;
					goto yy110;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy111;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy112;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy114;
					goto yy112;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy112;
				} else {
					if (yych <= 'e') goto yy114;
					if (yych <= 'f') goto yy112;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy30:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy119;
yy31:
#line 216 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_COLON); }
#line 1180 "../../lnav/src/data_scanner_re.cc"
yy32:
	++YYCURSOR;
#line 219 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_SEMI); }
#line 1185 "../../lnav/src/data_scanner_re.cc"
yy33:
	yyaccept = 8;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '>') {
		if (yych <= '-') {
			if (yych == '!') goto yy120;
			if (yych >= '-') goto yy121;
		} else {
			if (yych <= '.') goto yy34;
			if (yych <= '/') goto yy122;
			if (yych <= ':') goto yy121;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy123;
			if (yych <= '@') goto yy34;
			if (yych <= 'Z') goto yy121;
		} else {
			if (yych == '`') goto yy34;
			if (yych <= 'z') goto yy121;
		}
	}
yy34:
#line 227 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LANGLE); }
#line 1211 "../../lnav/src/data_scanner_re.cc"
yy35:
	++YYCURSOR;
#line 217 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EQUALS); }
#line 1216 "../../lnav/src/data_scanner_re.cc"
yy36:
	++YYCURSOR;
#line 228 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RANGLE); }
#line 1221 "../../lnav/src/data_scanner_re.cc"
yy37:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '\'') goto yy124;
				goto yy4;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy80;
				goto yy4;
			} else {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				goto yy126;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy127;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy128;
				goto yy129;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy130;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy38:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '\'') goto yy124;
				goto yy4;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy126;
				goto yy127;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= 'A') {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy132;
			} else {
				if (yych <= 'F') goto yy128;
				if (yych <= 'Z') goto yy129;
				goto yy4;
			}
		} else {
			if (yych <= 'a') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				goto yy133;
			} else {
				if (yych <= 'f') goto yy130;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy39:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '\'') {
			if (yych == '%') goto yy80;
			if (yych <= '&') goto yy4;
			goto yy124;
		} else {
			if (yych <= '+') {
				if (yych <= '*') goto yy4;
				goto yy80;
			} else {
				if (yych <= ',') goto yy4;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				if (yych <= '9') goto yy115;
				goto yy134;
			} else {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy129;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy40:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '\'') goto yy124;
				goto yy4;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy115;
				goto yy134;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= 'T') {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy129;
			} else {
				if (yych <= 'U') goto yy135;
				if (yych <= 'Z') goto yy129;
				goto yy4;
			}
		} else {
			if (yych <= 'n') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				goto yy131;
			} else {
				if (yych <= 'o') goto yy136;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy41:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '\'') {
			if (yych == '%') goto yy80;
			if (yych <= '&') goto yy4;
			goto yy137;
		} else {
			if (yych <= '+') {
				if (yych <= '*') goto yy4;
				goto yy80;
			} else {
				if (yych <= ',') goto yy4;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				if (yych <= '9') goto yy115;
				goto yy134;
			} else {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy129;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy42:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '\'') goto yy124;
				goto yy4;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy115;
				goto yy134;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= 'Q') {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy129;
			} else {
				if (yych <= 'R') goto yy138;
				if (yych <= 'Z') goto yy129;
				goto yy4;
			}
		} else {
			if (yych <= 'q') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				goto yy131;
			} else {
				if (yych <= 'r') goto yy139;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy43:
	yych = *++YYCURSOR;
	if (yych == ']') goto yy93;
#line 223 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LSQUARE); }
#line 1486 "../../lnav/src/data_scanner_re.cc"
yy44:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '[') {
			if (yych == '\n') goto yy12;
			goto yy140;
		} else {
			if (yych <= '\\') goto yy98;
			if (yych == 'n') goto yy68;
			goto yy140;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy12;
			if (yych <= 0xDF) goto yy141;
			if (yych <= 0xE0) goto yy142;
			goto yy143;
		} else {
			if (yych <= 0xF0) goto yy144;
			if (yych <= 0xF3) goto yy145;
			if (yych <= 0xF4) goto yy146;
			goto yy12;
		}
	}
yy45:
	++YYCURSOR;
#line 224 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RSQUARE); }
#line 1516 "../../lnav/src/data_scanner_re.cc"
yy46:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[1024+yych] & 16) {
		goto yy46;
	}
	if (yych <= ',') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy5;
				if (yych <= 0x08) goto yy3;
				goto yy5;
			} else {
				if (yych == '\r') goto yy5;
				if (yych <= 0x1A) goto yy3;
				goto yy5;
			}
		} else {
			if (yych <= '$') {
				if (yych <= 0x1F) goto yy3;
				if (yych <= '#') goto yy5;
				goto yy3;
			} else {
				if (yych <= '%') goto yy80;
				if (yych == '+') goto yy80;
				goto yy5;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '?') {
				if (yych <= '/') goto yy3;
				if (yych <= ':') goto yy62;
				goto yy5;
			} else {
				if (yych <= '@') goto yy96;
				if (yych == 0x7F) goto yy3;
				goto yy5;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy56;
				if (yych <= 0xE0) goto yy57;
				goto yy58;
			} else {
				if (yych <= 0xF0) goto yy59;
				if (yych <= 0xF3) goto yy60;
				if (yych <= 0xF4) goto yy61;
				goto yy5;
			}
		}
	}
yy47:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '\'') goto yy124;
				goto yy4;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy80;
				goto yy4;
			} else {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				goto yy126;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy147;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy128;
				goto yy129;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy130;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy48:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '\'') goto yy124;
				goto yy4;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy126;
				goto yy147;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy128;
				if (yych <= 'Z') goto yy129;
				goto yy4;
			}
		} else {
			if (yych <= 'a') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				goto yy133;
			} else {
				if (yych <= 'f') goto yy130;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy49:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '\'') {
			if (yych == '%') goto yy80;
			if (yych <= '&') goto yy4;
			goto yy124;
		} else {
			if (yych <= '+') {
				if (yych <= '*') goto yy4;
				goto yy80;
			} else {
				if (yych <= ',') goto yy4;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				if (yych <= '9') goto yy115;
				goto yy148;
			} else {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy129;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy50:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '\'') goto yy124;
				goto yy4;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy80;
				goto yy4;
			} else {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				goto yy115;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') goto yy148;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy129;
				goto yy4;
			}
		} else {
			if (yych <= 't') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				goto yy131;
			} else {
				if (yych <= 'u') goto yy149;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy51:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '&') {
			if (yych <= '"') {
				if (yych <= '!') goto yy4;
				goto yy70;
			} else {
				if (yych == '%') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= '+') {
				if (yych <= '\'') goto yy150;
				if (yych <= '*') goto yy4;
				goto yy80;
			} else {
				if (yych <= ',') goto yy4;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '9') goto yy115;
				if (yych <= ':') goto yy148;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy129;
				goto yy4;
			}
		} else {
			if (yych <= 'd') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				goto yy131;
			} else {
				if (yych <= 'e') goto yy151;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy52:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '\'') goto yy124;
				goto yy4;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy80;
				goto yy4;
			} else {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				goto yy115;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') goto yy148;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy129;
				goto yy4;
			}
		} else {
			if (yych <= 'q') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				goto yy131;
			} else {
				if (yych <= 'r') goto yy139;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy53:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= '&') {
			if (yych <= '"') {
				if (yych <= '!') goto yy4;
				goto yy70;
			} else {
				if (yych == '%') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '\'') goto yy150;
				goto yy4;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy4;
				goto yy46;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '9') {
				if (yych <= '/') goto yy4;
				goto yy115;
			} else {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy4;
				goto yy96;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy129;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy131;
				goto yy4;
			}
		}
	}
yy54:
	yych = *++YYCURSOR;
	if (yych == '}') goto yy93;
#line 221 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LCURLY); }
#line 1879 "../../lnav/src/data_scanner_re.cc"
yy55:
	++YYCURSOR;
#line 222 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RCURLY); }
#line 1884 "../../lnav/src/data_scanner_re.cc"
yy56:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy3;
	goto yy1;
yy57:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy56;
	goto yy1;
yy58:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy56;
	goto yy1;
yy59:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy58;
	goto yy1;
yy60:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy58;
	goto yy1;
yy61:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy58;
	goto yy1;
yy62:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy152;
	goto yy1;
yy63:
	yych = *++YYCURSOR;
yy64:
	if (yybm[1024+yych] & 32) {
		goto yy63;
	}
	goto yy7;
yy65:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy153;
	goto yy1;
yy66:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy154;
	goto yy1;
yy67:
	yych = *++YYCURSOR;
	if (yych == '\n') goto yy1;
	goto yy156;
yy68:
	++YYCURSOR;
	goto yy9;
yy69:
	yych = *++YYCURSOR;
	if (yybm[1024+yych] & 64) {
		goto yy69;
	}
	if (yych <= '@') goto yy1;
	if (yych <= 'Z') goto yy157;
	if (yych <= '`') goto yy1;
	if (yych <= 'z') goto yy157;
	goto yy1;
yy70:
	yych = *++YYCURSOR;
yy71:
	if (yybm[1024+yych] & 128) {
		goto yy70;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x1B) goto yy1;
		} else {
			if (yych <= '\\') goto yy73;
			if (yych <= 0xC1) goto yy1;
			goto yy74;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy75;
			if (yych <= 0xEF) goto yy76;
			goto yy77;
		} else {
			if (yych <= 0xF3) goto yy78;
			if (yych <= 0xF4) goto yy79;
			goto yy1;
		}
	}
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '"') goto yy70;
yy72:
#line 143 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 1;
           cap_inner.c_end -= 1;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 1994 "../../lnav/src/data_scanner_re.cc"
yy73:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy70;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy70;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy75;
			if (yych <= 0xEF) goto yy76;
			goto yy77;
		} else {
			if (yych <= 0xF3) goto yy78;
			if (yych <= 0xF4) goto yy79;
			goto yy1;
		}
	}
yy74:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy70;
	goto yy1;
yy75:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy74;
	goto yy1;
yy76:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy74;
	goto yy1;
yy77:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy76;
	goto yy1;
yy78:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy76;
	goto yy1;
yy79:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy76;
	goto yy1;
yy80:
	yych = *++YYCURSOR;
yy81:
	if (yybm[768+yych] & 1) {
		goto yy80;
	}
	if (yych != '@') goto yy1;
yy82:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy158;
	goto yy159;
yy83:
	yych = *++YYCURSOR;
yy84:
	if (yybm[768+yych] & 2) {
		goto yy83;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x1B) goto yy1;
		} else {
			if (yych <= '\\') goto yy86;
			if (yych <= 0xC1) goto yy1;
			goto yy87;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy88;
			if (yych <= 0xEF) goto yy89;
			goto yy90;
		} else {
			if (yych <= 0xF3) goto yy91;
			if (yych <= 0xF4) goto yy92;
			goto yy1;
		}
	}
yy85:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= 'R') {
			if (yych == '\'') {
				yyt2 = YYCURSOR;
				goto yy162;
			}
			yyt2 = YYCURSOR;
			goto yy160;
		} else {
			if (yych <= 'S') goto yy1;
			if (yych == 's') goto yy1;
			yyt2 = YYCURSOR;
			goto yy160;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy163;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy164;
			}
			yyt2 = YYCURSOR;
			goto yy165;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy166;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy167;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy168;
			}
			goto yy1;
		}
	}
yy86:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy83;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy83;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy88;
			if (yych <= 0xEF) goto yy89;
			goto yy90;
		} else {
			if (yych <= 0xF3) goto yy91;
			if (yych <= 0xF4) goto yy92;
			goto yy1;
		}
	}
yy87:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy83;
	goto yy1;
yy88:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy87;
	goto yy1;
yy89:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy87;
	goto yy1;
yy90:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy89;
	goto yy1;
yy91:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy89;
	goto yy1;
yy92:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy89;
	goto yy1;
yy93:
	++YYCURSOR;
#line 220 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EMPTY_CONTAINER); }
#line 2182 "../../lnav/src/data_scanner_re.cc"
yy94:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= ':') {
		if (yych <= '$') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) goto yy26;
					goto yy4;
				} else {
					if (yych <= '\n') goto yy26;
					if (yych <= '\f') goto yy4;
					goto yy26;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy4;
					goto yy26;
				} else {
					if (yych <= 0x1F) goto yy4;
					if (yych <= '#') goto yy26;
					goto yy4;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '*') {
					if (yych <= '%') goto yy105;
					goto yy26;
				} else {
					if (yych <= '+') goto yy80;
					if (yych <= ',') goto yy26;
					goto yy46;
				}
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy169;
					goto yy4;
				} else {
					if (yych <= '7') goto yy170;
					if (yych >= ':') goto yy4;
				}
			}
		}
	} else {
		if (yych <= 'd') {
			if (yych <= 'F') {
				if (yych <= '@') {
					if (yych <= '?') goto yy26;
					goto yy96;
				} else {
					if (yych == 'E') goto yy172;
					goto yy171;
				}
			} else {
				if (yych <= '^') {
					if (yych <= 'Z') goto yy46;
					goto yy26;
				} else {
					if (yych <= '_') goto yy46;
					if (yych <= '`') goto yy26;
					goto yy171;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'f') {
					if (yych <= 'e') goto yy172;
					goto yy171;
				} else {
					if (yych == 'x') goto yy173;
					goto yy46;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy95:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '@') {
		if (yych <= '$') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) goto yy26;
					goto yy3;
				} else {
					if (yych <= '\n') goto yy26;
					if (yych <= '\f') goto yy3;
					goto yy26;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy3;
					goto yy26;
				} else {
					if (yych <= 0x1F) goto yy3;
					if (yych <= '#') goto yy26;
					goto yy3;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '*') {
					if (yych <= '%') goto yy105;
					goto yy26;
				} else {
					if (yych <= '+') goto yy80;
					if (yych <= ',') goto yy26;
					goto yy46;
				}
			} else {
				if (yych <= '9') {
					if (yych <= '.') goto yy169;
					if (yych <= '/') goto yy3;
					goto yy95;
				} else {
					if (yych <= ':') goto yy62;
					if (yych <= '?') goto yy26;
				}
			}
		}
	} else {
		if (yych <= 'f') {
			if (yych <= '^') {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy171;
					goto yy172;
				} else {
					if (yych <= 'F') goto yy171;
					if (yych <= 'Z') goto yy46;
					goto yy26;
				}
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy26;
				} else {
					if (yych == 'e') goto yy172;
					goto yy171;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= '~') {
					if (yych <= 'z') goto yy46;
					goto yy26;
				} else {
					if (yych <= 0x7F) goto yy3;
					if (yych <= 0xC1) goto yy26;
					goto yy56;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) goto yy57;
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy26;
				}
			}
		}
	}
yy96:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ',') goto yy4;
		if (yych == '/') goto yy4;
		goto yy174;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy4;
			goto yy174;
		} else {
			if (yych <= '`') goto yy4;
			if (yych <= 'z') goto yy174;
			goto yy4;
		}
	}
yy97:
	yych = *++YYCURSOR;
	if (yych != '/') goto yy81;
yy98:
	yych = *++YYCURSOR;
	if (yych == 'P') goto yy175;
	goto yy100;
yy99:
	yych = *++YYCURSOR;
yy100:
	if (yybm[768+yych] & 8) {
		goto yy99;
	}
	goto yy24;
yy101:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 16) {
		goto yy101;
	}
	if (yych <= ',') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy3;
				goto yy24;
			} else {
				if (yych == '\r') goto yy24;
				if (yych <= 0x1A) goto yy3;
				goto yy24;
			}
		} else {
			if (yych <= '"') {
				if (yych <= 0x1F) goto yy3;
				if (yych == '!') goto yy99;
				goto yy24;
			} else {
				if (yych == '\'') goto yy24;
				if (yych <= '*') goto yy99;
				goto yy24;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= ']') {
				if (yych <= ':') goto yy62;
				if (yych == '\\') goto yy99;
				goto yy24;
			} else {
				if (yych <= '^') goto yy99;
				if (yych <= '}') goto yy24;
				if (yych <= '~') goto yy99;
				goto yy3;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy24;
				if (yych <= 0xDF) goto yy56;
				if (yych <= 0xE0) goto yy57;
				goto yy58;
			} else {
				if (yych <= 0xF0) goto yy59;
				if (yych <= 0xF3) goto yy60;
				if (yych <= 0xF4) goto yy61;
				goto yy24;
			}
		}
	}
yy102:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy24;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy24;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy24;
					if (yych <= 0x1F) goto yy4;
					goto yy24;
				}
			}
		} else {
			if (yych <= '$') {
				if (yych == '"') goto yy24;
				if (yych <= '#') goto yy99;
				goto yy101;
			} else {
				if (yych <= '\'') {
					if (yych <= '&') goto yy99;
					goto yy24;
				} else {
					if (yych <= '*') goto yy99;
					if (yych <= ',') goto yy24;
					goto yy101;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= '[') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy24;
				if (yych <= 'Z') goto yy101;
				goto yy24;
			} else {
				if (yych <= ']') {
					if (yych <= '\\') goto yy99;
					goto yy24;
				} else {
					if (yych <= '^') goto yy99;
					if (yych <= '_') goto yy101;
					goto yy24;
				}
			}
		} else {
			if (yych <= '}') {
				if (yych == 'r') goto yy176;
				if (yych <= 'z') goto yy101;
				goto yy24;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy99;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy24;
					if (yych <= 0xF4) goto yy4;
					goto yy24;
				}
			}
		}
	}
yy103:
	yych = *++YYCURSOR;
yy104:
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych == '%') goto yy177;
	goto yy1;
yy105:
	yyaccept = 10;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy81;
		} else {
			if (yych == ',') goto yy106;
			if (yych <= '.') goto yy81;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy81;
			if (yych <= '?') goto yy106;
			if (yych <= 'Z') goto yy81;
		} else {
			if (yych == '`') goto yy106;
			if (yych <= 'z') goto yy81;
		}
	}
yy106:
#line 256 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_PERCENTAGE); }
#line 2548 "../../lnav/src/data_scanner_re.cc"
yy107:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '0') goto yy178;
				goto yy179;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') {
				if (yych <= '2') goto yy180;
				goto yy178;
			} else {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy46;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy108:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych >= '\v') goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych >= 0x0E) goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy181;
					goto yy182;
				} else {
					if (yych <= '/') goto yy183;
					if (yych <= '7') goto yy184;
					goto yy185;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy186;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy187;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy188;
					goto yy187;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych >= '_') goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych >= 'a') goto yy187;
				} else {
					if (yych <= 'e') goto yy188;
					if (yych <= 'f') goto yy187;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych >= 0x7F) goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
				}
			}
		}
	}
yy109:
#line 255 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_OCTAL_NUMBER); }
#line 2665 "../../lnav/src/data_scanner_re.cc"
yy110:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy181;
				} else {
					if (yych <= '.') goto yy182;
					if (yych <= '/') goto yy183;
					goto yy185;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy186;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy187;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy188;
					goto yy187;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy187;
				} else {
					if (yych <= 'e') goto yy188;
					if (yych <= 'f') goto yy187;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy111:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1;
		if (yych <= '/') goto yy189;
		if (yych <= '9') goto yy190;
		goto yy191;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy192;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy192;
			goto yy1;
		}
	}
yy112:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych >= '\v') goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych != 0x1B) goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy181;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy187;
				if (yych <= ':') goto yy193;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy187;
				if (yych <= 'Z') goto yy115;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy187;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych >= 0x7F) goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
				}
			}
		}
	}
yy113:
#line 258 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_HEX_NUMBER); }
#line 2817 "../../lnav/src/data_scanner_re.cc"
yy114:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy195;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy187;
				if (yych <= ':') goto yy193;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy187;
				if (yych >= '[') goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy187;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy115:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 64) {
		goto yy115;
	}
	if (yych <= '.') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy5;
				if (yych <= 0x08) goto yy3;
				goto yy5;
			} else {
				if (yych == '\r') goto yy5;
				if (yych <= 0x1A) goto yy3;
				goto yy5;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy3;
				if (yych <= '#') goto yy5;
				if (yych <= '$') goto yy3;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy5;
				goto yy46;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '@') {
				if (yych <= '/') goto yy3;
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy5;
				goto yy96;
			} else {
				if (yych == '_') goto yy46;
				if (yych <= '~') goto yy5;
				goto yy3;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) goto yy56;
				if (yych <= 0xE0) goto yy57;
				goto yy58;
			} else {
				if (yych <= 0xF0) goto yy59;
				if (yych <= 0xF3) goto yy60;
				if (yych <= 0xF4) goto yy61;
				goto yy5;
			}
		}
	}
yy116:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy196;
				goto yy148;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy196;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy196;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy117:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy181;
				} else {
					if (yych <= '.') goto yy182;
					if (yych <= '/') goto yy183;
					goto yy197;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy186;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy187;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy188;
					goto yy187;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy187;
				} else {
					if (yych <= 'e') goto yy188;
					if (yych <= 'f') goto yy187;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy118:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy26;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy181;
					goto yy182;
				} else {
					if (yych <= '/') goto yy183;
					if (yych <= '5') goto yy197;
					goto yy185;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy186;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy187;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy188;
					goto yy187;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy187;
				} else {
					if (yych <= 'e') goto yy188;
					if (yych <= 'f') goto yy187;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy119:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy200;
			} else {
				if (yych <= '1') goto yy201;
				if (yych <= '2') goto yy202;
				goto yy200;
			}
		} else {
			if (yych <= 'E') {
				if (yych <= ':') goto yy1;
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy203;
			} else {
				if (yych <= 'F') goto yy204;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'e') goto yy203;
				if (yych <= 'f') goto yy204;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy120:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1;
			goto yy211;
		} else {
			if (yych <= '/') goto yy1;
			if (yych <= ':') goto yy211;
			goto yy1;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy211;
			if (yych <= '^') goto yy1;
			goto yy211;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'z') goto yy211;
			goto yy1;
		}
	}
yy121:
	yych = *++YYCURSOR;
	if (yybm[512+yych] & 16) {
		goto yy214;
	}
	goto yy213;
yy122:
	yych = *++YYCURSOR;
	if (yych <= '\r') {
		if (yych == '\t') goto yy1;
		if (yych <= '\f') goto yy219;
		goto yy1;
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy219;
			goto yy1;
		} else {
			if (yych == '>') goto yy1;
			goto yy219;
		}
	}
yy123:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1;
			goto yy220;
		} else {
			if (yych <= '/') goto yy1;
			if (yych <= ':') goto yy220;
			goto yy1;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy220;
			if (yych <= '^') goto yy1;
			goto yy220;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'z') goto yy220;
			goto yy1;
		}
	}
yy124:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 128) {
		goto yy137;
	}
	if (yych <= '*') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy225;
			if (yych <= 0x08) goto yy125;
			if (yych <= '\n') goto yy225;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy225;
			} else {
				if (yych <= '"') goto yy225;
				if (yych >= '\'') goto yy225;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych == ',') goto yy225;
			if (yych >= '.') {
				yyt4 = YYCURSOR;
				goto yy226;
			}
		} else {
			if (yych <= ';') {
				if (yych >= ':') goto yy225;
			} else {
				if (yych == '?') goto yy225;
			}
		}
	}
yy125:
#line 155 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_WORD);
       }
#line 3304 "../../lnav/src/data_scanner_re.cc"
yy126:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy181;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy227;
				goto yy193;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy227;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy227;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy127:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych <= '.') goto yy1;
			goto yy189;
		} else {
			if (yych <= '9') goto yy192;
			if (yych <= ':') goto yy191;
			goto yy1;
		}
	} else {
		if (yych <= '\\') {
			if (yych <= 'F') goto yy192;
			if (yych <= '[') goto yy1;
			goto yy98;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy192;
			goto yy1;
		}
	}
yy128:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy181;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy227;
				goto yy193;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy228;
				if (yych <= 'Z') goto yy229;
				goto yy4;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy228;
				if (yych <= 'z') goto yy229;
				goto yy4;
			}
		}
	}
yy129:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy115;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy148;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy229;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy229;
				goto yy4;
			}
		}
	}
yy130:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy181;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy227;
			} else {
				if (yych <= ':') {
					yyt4 = YYCURSOR;
					goto yy231;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy228;
				if (yych <= 'Z') goto yy229;
				goto yy4;
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy4;
				} else {
					if (yych <= 'f') goto yy232;
					if (yych <= 'z') goto yy233;
					goto yy4;
				}
			}
		}
	}
yy131:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= ' ') {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				yyt4 = YYCURSOR;
				goto yy221;
			} else {
				if (yych == '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				yyt4 = YYCURSOR;
				goto yy221;
			}
		} else {
			if (yych <= '%') {
				if (yych <= '!') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych <= '&') goto yy4;
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy80;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				goto yy4;
			} else {
				if (yych <= '9') goto yy115;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy229;
				goto yy4;
			} else {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy233;
				goto yy4;
			}
		}
	}
yy132:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= '/') {
				if (yych <= '-') goto yy181;
				if (yych <= '.') goto yy46;
				goto yy4;
			} else {
				if (yych <= '9') goto yy227;
				if (yych <= ':') goto yy193;
				goto yy4;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= 'F') {
				if (yych <= '@') goto yy96;
				goto yy228;
			} else {
				if (yych == 'L') goto yy235;
				goto yy229;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy228;
				if (yych <= 'z') goto yy229;
				goto yy4;
			}
		}
	}
yy133:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '*') {
					if (yych <= '\'') {
						yyt4 = YYCURSOR;
						goto yy224;
					}
					yyt4 = YYCURSOR;
					goto yy221;
				} else {
					if (yych <= '+') goto yy80;
					if (yych <= ',') {
						yyt4 = YYCURSOR;
						goto yy223;
					}
					goto yy181;
				}
			}
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') {
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy227;
				yyt4 = YYCURSOR;
				goto yy231;
			} else {
				if (yych <= '>') {
					if (yych <= ';') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy4;
				} else {
					if (yych <= '?') {
						yyt4 = YYCURSOR;
						goto yy223;
					}
					if (yych <= '@') goto yy96;
					goto yy228;
				}
			}
		} else {
			if (yych <= '`') {
				if (yych <= 'Z') goto yy229;
				if (yych == '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'k') {
					if (yych <= 'f') goto yy232;
					goto yy233;
				} else {
					if (yych <= 'l') goto yy236;
					if (yych <= 'z') goto yy233;
					goto yy4;
				}
			}
		}
	}
yy134:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == '/') goto yy189;
		goto yy1;
	} else {
		if (yych <= ':') goto yy152;
		if (yych == '\\') goto yy98;
		goto yy1;
	}
yy135:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy115;
				goto yy148;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych == 'L') goto yy237;
				goto yy229;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy229;
				goto yy4;
			}
		}
	}
yy136:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy115;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy229;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= 'm') {
					if (yych <= '`') goto yy4;
					goto yy233;
				} else {
					if (yych <= 'n') goto yy238;
					if (yych <= 'z') goto yy233;
					goto yy4;
				}
			}
		}
	}
yy137:
	yych = *++YYCURSOR;
	if (yybm[768+yych] & 128) {
		goto yy137;
	}
	if (yych <= '&') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy1;
			} else {
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '\f') goto yy1;
				yyt4 = YYCURSOR;
				goto yy221;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1;
				yyt4 = YYCURSOR;
				goto yy221;
			} else {
				if (yych <= '!') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy1;
			}
		}
	} else {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				yyt4 = YYCURSOR;
				goto yy221;
			} else {
				if (yych == ',') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				goto yy1;
			}
		} else {
			if (yych <= ';') {
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy226;
				}
				if (yych <= '9') goto yy1;
				yyt4 = YYCURSOR;
				goto yy221;
			} else {
				if (yych == '?') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				goto yy1;
			}
		}
	}
yy138:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy115;
				goto yy148;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych == 'U') goto yy239;
				goto yy229;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy229;
				goto yy4;
			}
		}
	}
yy139:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy115;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy229;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= 't') {
					if (yych <= '`') goto yy4;
					goto yy233;
				} else {
					if (yych <= 'u') goto yy238;
					if (yych <= 'z') goto yy233;
					goto yy4;
				}
			}
		}
	}
yy140:
	++YYCURSOR;
#line 273 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_ESCAPED_CHAR); }
#line 4114 "../../lnav/src/data_scanner_re.cc"
yy141:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy140;
	goto yy1;
yy142:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy141;
	goto yy1;
yy143:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy141;
	goto yy1;
yy144:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy143;
	goto yy1;
yy145:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy143;
	goto yy1;
yy146:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy143;
	goto yy1;
yy147:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1;
		if (yych <= '/') goto yy189;
		if (yych <= '9') goto yy192;
		goto yy191;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy192;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy192;
			goto yy1;
		}
	}
yy148:
	yych = *++YYCURSOR;
	if (yych == '/') goto yy189;
	if (yych == ':') goto yy152;
	goto yy1;
yy149:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy115;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy229;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= 'k') {
					if (yych <= '`') goto yy4;
					goto yy233;
				} else {
					if (yych <= 'l') goto yy240;
					if (yych <= 'z') goto yy233;
					goto yy4;
				}
			}
		}
	}
yy150:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy83;
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy241;
				}
				goto yy83;
			} else {
				if (yych <= '\r') {
					yyt3 = YYCURSOR;
					goto yy241;
				}
				if (yych == 0x1B) goto yy1;
				goto yy83;
			}
		} else {
			if (yych <= '&') {
				if (yych == '!') {
					yyt3 = YYCURSOR;
					goto yy242;
				}
				if (yych <= '"') {
					yyt3 = YYCURSOR;
					goto yy241;
				}
				goto yy83;
			} else {
				if (yych <= '*') {
					if (yych <= '\'') {
						yyt3 = YYCURSOR;
						goto yy243;
					}
					yyt3 = YYCURSOR;
					goto yy241;
				} else {
					if (yych == ',') {
						yyt3 = YYCURSOR;
						goto yy242;
					}
					goto yy83;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '>') {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy244;
				}
				if (yych <= '9') goto yy83;
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy241;
				}
				goto yy83;
			} else {
				if (yych <= '[') {
					if (yych <= '?') {
						yyt3 = YYCURSOR;
						goto yy242;
					}
					goto yy83;
				} else {
					if (yych <= '\\') goto yy86;
					if (yych <= '`') goto yy83;
					goto yy150;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy83;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy87;
				goto yy88;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy89;
					goto yy90;
				} else {
					if (yych <= 0xF3) goto yy91;
					if (yych <= 0xF4) goto yy92;
					goto yy1;
				}
			}
		}
	}
yy151:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= ' ') {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				yyt4 = YYCURSOR;
				goto yy221;
			} else {
				if (yych == '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				yyt4 = YYCURSOR;
				goto yy221;
			}
		} else {
			if (yych <= '%') {
				if (yych <= '!') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych <= '&') goto yy4;
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy80;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '-') goto yy245;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				goto yy4;
			} else {
				if (yych <= '9') goto yy115;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy229;
				goto yy4;
			} else {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy233;
				goto yy4;
			}
		}
	}
yy152:
	yych = *++YYCURSOR;
	if (yybm[1024+yych] & 8) {
		goto yy3;
	}
	if (yych <= 0xEF) {
		if (yych <= 0xC1) goto yy1;
		if (yych <= 0xDF) goto yy56;
		if (yych <= 0xE0) goto yy57;
		goto yy58;
	} else {
		if (yych <= 0xF0) goto yy59;
		if (yych <= 0xF3) goto yy60;
		if (yych <= 0xF4) goto yy61;
		goto yy1;
	}
yy153:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy246;
	goto yy1;
yy154:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy247;
	goto yy1;
yy155:
	yych = *++YYCURSOR;
yy156:
	if (yybm[512+yych] & 1) {
		goto yy155;
	}
	if (yych == '\n') goto yy248;
	goto yy1;
yy157:
	++YYCURSOR;
#line 212 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_CSI);
       }
#line 4483 "../../lnav/src/data_scanner_re.cc"
yy158:
	yych = *++YYCURSOR;
yy159:
	if (yybm[512+yych] & 2) {
		goto yy158;
	}
	if (yych <= ',') goto yy1;
	if (yych <= '.') goto yy249;
	goto yy1;
yy160:
	++YYCURSOR;
	yyt1 = yyt2;
yy161:
	YYCURSOR = yyt1;
#line 158 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 1;
           cap_inner.c_end -= 1;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 4511 "../../lnav/src/data_scanner_re.cc"
yy162:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	yyt1 = yyt2;
	if (yybm[768+yych] & 2) {
		goto yy83;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x1B) goto yy161;
			goto yy85;
		} else {
			if (yych <= '\\') goto yy86;
			if (yych <= 0xC1) goto yy161;
			goto yy87;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy88;
			if (yych <= 0xEF) goto yy89;
			goto yy90;
		} else {
			if (yych <= 0xF3) goto yy91;
			if (yych <= 0xF4) goto yy92;
			goto yy161;
		}
	}
yy163:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy160;
	goto yy1;
yy164:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy163;
	goto yy1;
yy165:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy163;
	goto yy1;
yy166:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy165;
	goto yy1;
yy167:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy165;
	goto yy1;
yy168:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy165;
	goto yy1;
yy169:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy250;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy46;
			goto yy4;
		}
	}
yy170:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '?') {
		if (yych <= '$') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) goto yy109;
					goto yy3;
				} else {
					if (yych <= '\n') goto yy109;
					if (yych <= '\f') goto yy3;
					goto yy109;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy3;
					goto yy109;
				} else {
					if (yych <= 0x1F) goto yy3;
					if (yych <= '#') goto yy109;
					goto yy3;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '*') {
					if (yych <= '%') goto yy105;
					goto yy109;
				} else {
					if (yych <= '+') goto yy80;
					if (yych <= ',') goto yy109;
					goto yy46;
				}
			} else {
				if (yych <= '7') {
					if (yych <= '.') goto yy169;
					if (yych <= '/') goto yy3;
					goto yy170;
				} else {
					if (yych <= '9') goto yy95;
					if (yych <= ':') goto yy62;
					goto yy109;
				}
			}
		}
	} else {
		if (yych <= 'f') {
			if (yych <= 'Z') {
				if (yych <= 'D') {
					if (yych <= '@') goto yy96;
				} else {
					if (yych <= 'E') goto yy172;
					if (yych >= 'G') goto yy46;
				}
			} else {
				if (yych <= '`') {
					if (yych == '_') goto yy46;
					goto yy109;
				} else {
					if (yych == 'e') goto yy172;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= '~') {
					if (yych <= 'z') goto yy46;
					goto yy109;
				} else {
					if (yych <= 0x7F) goto yy3;
					if (yych <= 0xC1) goto yy109;
					goto yy56;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) goto yy57;
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy109;
				}
			}
		}
	}
yy171:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[512+yych] & 8) {
		goto yy171;
	}
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy113;
				goto yy3;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy3;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy113;
				} else {
					if (yych <= '.') goto yy46;
					if (yych <= '/') goto yy3;
					goto yy62;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '?') goto yy113;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy46;
				goto yy113;
			} else {
				if (yych == '`') goto yy113;
				if (yych <= 'z') goto yy46;
				goto yy113;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy113;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy113;
				}
			}
		}
	}
yy172:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[512+yych] & 8) {
		goto yy171;
	}
	if (yych <= ',') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				goto yy113;
			} else {
				if (yych == '\r') goto yy113;
				if (yych <= 0x1A) goto yy4;
				goto yy113;
			}
		} else {
			if (yych <= '$') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy113;
				goto yy4;
			} else {
				if (yych <= '%') goto yy80;
				if (yych == '+') goto yy194;
				goto yy113;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '-') goto yy251;
				if (yych <= '.') goto yy46;
				goto yy4;
			} else {
				if (yych <= '?') goto yy113;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy46;
				goto yy113;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy113;
				if (yych <= 'z') goto yy46;
				goto yy113;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy113;
				if (yych <= 0xF4) goto yy4;
				goto yy113;
			}
		}
	}
yy173:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[512+yych] & 8) {
		goto yy171;
	}
	if (yych <= '?') {
		if (yych <= '*') {
			if (yych == '%') goto yy80;
			goto yy4;
		} else {
			if (yych <= '+') goto yy80;
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			goto yy4;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy46;
			goto yy4;
		}
	}
yy174:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy5;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy5;
				goto yy3;
			} else {
				if (yych <= '\r') goto yy5;
				if (yych == 0x1B) goto yy5;
				goto yy3;
			}
		} else {
			if (yych <= '-') {
				if (yych == '$') goto yy3;
				if (yych <= ',') goto yy5;
				goto yy174;
			} else {
				if (yych <= '.') goto yy252;
				if (yych <= '/') goto yy3;
				if (yych <= '9') goto yy174;
				goto yy62;
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '?') goto yy5;
				if (yych <= '@') goto yy3;
				if (yych <= 'Z') goto yy174;
				goto yy5;
			} else {
				if (yych <= '_') goto yy3;
				if (yych <= '`') goto yy5;
				if (yych <= 'z') goto yy174;
				goto yy5;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy5;
				}
			}
		}
	}
yy175:
	yych = *++YYCURSOR;
	if (yych == 'r') goto yy253;
	goto yy100;
yy176:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy24;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy24;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy24;
					if (yych <= 0x1F) goto yy4;
					goto yy24;
				}
			}
		} else {
			if (yych <= '$') {
				if (yych == '"') goto yy24;
				if (yych <= '#') goto yy99;
				goto yy101;
			} else {
				if (yych <= '\'') {
					if (yych <= '&') goto yy99;
					goto yy24;
				} else {
					if (yych <= '*') goto yy99;
					if (yych <= ',') goto yy24;
					goto yy101;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= '[') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy24;
				if (yych <= 'Z') goto yy101;
				goto yy24;
			} else {
				if (yych <= ']') {
					if (yych <= '\\') goto yy99;
					goto yy24;
				} else {
					if (yych <= '^') goto yy99;
					if (yych <= '_') goto yy101;
					goto yy24;
				}
			}
		} else {
			if (yych <= '}') {
				if (yych == 'o') goto yy254;
				if (yych <= 'z') goto yy101;
				goto yy24;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy99;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy24;
					if (yych <= 0xF4) goto yy4;
					goto yy24;
				}
			}
		}
	}
yy177:
	++YYCURSOR;
	goto yy106;
yy178:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				goto yy26;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy26;
				} else {
					if (yych <= '-') goto yy255;
					if (yych <= '.') goto yy256;
					goto yy4;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy257;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy258;
					goto yy259;
				} else {
					if (yych <= 'Z') goto yy258;
					if (yych <= '^') goto yy26;
					goto yy258;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') goto yy26;
				if (yych == 'e') goto yy259;
				goto yy258;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy179:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				goto yy26;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy26;
				} else {
					if (yych <= '-') goto yy255;
					if (yych <= '.') goto yy256;
					goto yy4;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy178;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy258;
					goto yy259;
				} else {
					if (yych <= 'Z') goto yy258;
					if (yych <= '^') goto yy26;
					goto yy258;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') goto yy26;
				if (yych == 'e') goto yy259;
				goto yy258;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy180:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '4') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy255;
				} else {
					if (yych <= '.') goto yy256;
					if (yych <= '/') goto yy4;
					goto yy178;
				}
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '5') goto yy260;
				if (yych <= '9') goto yy257;
				if (yych <= ':') goto yy4;
				goto yy26;
			} else {
				if (yych <= 'D') {
					if (yych <= '@') goto yy96;
					goto yy258;
				} else {
					if (yych <= 'E') goto yy259;
					if (yych <= 'Z') goto yy258;
					goto yy26;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') {
					if (yych <= '_') goto yy258;
					goto yy26;
				} else {
					if (yych == 'e') goto yy259;
					goto yy258;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy181:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy5;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy5;
				goto yy3;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy5;
					goto yy3;
				} else {
					if (yych <= 0x1B) goto yy5;
					if (yych <= 0x1F) goto yy3;
					goto yy5;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				if (yych <= '*') goto yy5;
				goto yy80;
			} else {
				if (yych <= '.') {
					if (yych <= ',') goto yy5;
					goto yy46;
				} else {
					if (yych <= '/') goto yy3;
					if (yych <= '9') goto yy261;
					goto yy62;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '?') goto yy5;
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy261;
				goto yy46;
			} else {
				if (yych <= '_') {
					if (yych <= '^') goto yy5;
					goto yy46;
				} else {
					if (yych <= '`') goto yy5;
					if (yych <= 'f') goto yy261;
					goto yy46;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy5;
					goto yy3;
				} else {
					if (yych <= 0xC1) goto yy5;
					if (yych <= 0xDF) goto yy56;
					goto yy57;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy5;
				}
			}
		}
	}
yy182:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '0') goto yy262;
				goto yy263;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') {
				if (yych <= '2') goto yy264;
				goto yy262;
			} else {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy46;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy183:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy4;
	if (yych <= 'Z') goto yy265;
	if (yych <= '`') goto yy4;
	if (yych <= 'z') goto yy265;
	goto yy4;
yy184:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy267;
					goto yy268;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy147;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy269;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy270;
					goto yy269;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy269;
				} else {
					if (yych <= 'e') goto yy270;
					if (yych <= 'f') goto yy269;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy185:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy268;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy147;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy269;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy270;
					goto yy269;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy269;
				} else {
					if (yych <= 'e') goto yy270;
					if (yych <= 'f') goto yy269;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy186:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1;
		if (yych <= '/') goto yy189;
		if (yych <= '9') goto yy271;
		goto yy191;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy272;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy272;
			goto yy1;
		}
	}
yy187:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy113;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy269;
				if (yych <= ':') goto yy147;
				if (yych <= '?') goto yy113;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy269;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy113;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy269;
				if (yych <= 'z') goto yy115;
				goto yy113;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy113;
				if (yych <= 0xF4) goto yy4;
				goto yy113;
			}
		}
	}
yy188:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy251;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy269;
				if (yych <= ':') goto yy147;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy269;
				if (yych <= 'Z') goto yy115;
				goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy269;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy189:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= ',') {
			if (yych <= '$') goto yy274;
			if (yych <= '&') goto yy273;
			goto yy274;
		} else {
			if (yych == '.') goto yy274;
			if (yych <= '9') goto yy273;
			goto yy274;
		}
	} else {
		if (yych <= '@') {
			if (yych == '>') goto yy274;
			if (yych <= '?') goto yy273;
			goto yy274;
		} else {
			if (yych <= 'Z') goto yy273;
			if (yych <= '`') goto yy274;
			if (yych <= 'z') goto yy273;
			goto yy274;
		}
	}
yy190:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy281;
		if (yych <= ':') goto yy282;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy283;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy283;
		goto yy1;
	}
yy191:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '0') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				goto yy285;
			} else {
				if (yych <= '2') {
					if (yych <= '1') goto yy286;
					goto yy287;
				} else {
					if (yych <= '9') goto yy285;
					if (yych <= ':') goto yy1;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy288;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy288;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy1;
				}
			}
		}
	}
yy192:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy283;
		if (yych <= ':') goto yy282;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy283;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy283;
		goto yy1;
	}
yy193:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1;
		if (yych <= '/') goto yy189;
		if (yych <= '9') goto yy272;
		goto yy191;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy272;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy272;
			goto yy1;
		}
	}
yy194:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy81;
	if (yych <= '9') goto yy295;
	goto yy81;
yy195:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy296;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy261;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy261;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy196:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy113;
				goto yy3;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy113;
					goto yy3;
				} else {
					if (yych <= 0x1B) goto yy113;
					if (yych <= 0x1F) goto yy3;
					goto yy113;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				if (yych <= '*') goto yy113;
				goto yy80;
			} else {
				if (yych <= '.') {
					if (yych <= ',') goto yy113;
					goto yy46;
				} else {
					if (yych <= '/') goto yy3;
					if (yych <= '9') goto yy196;
					goto yy148;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '?') goto yy113;
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy196;
				goto yy115;
			} else {
				if (yych <= '_') {
					if (yych <= '^') goto yy113;
					goto yy46;
				} else {
					if (yych <= '`') goto yy113;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy3;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xDF) goto yy56;
					goto yy57;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy113;
				}
			}
		}
	}
yy197:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy182;
					if (yych <= '/') goto yy4;
					goto yy268;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy147;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy269;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy270;
					goto yy269;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy269;
				} else {
					if (yych <= 'e') goto yy270;
					if (yych <= 'f') goto yy269;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy198:
	++YYCURSOR;
yy199:
	YYCURSOR = yyt2;
#line 190 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_IPV6_ADDRESS); }
#line 5991 "../../lnav/src/data_scanner_re.cc"
yy200:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy298;
				goto yy299;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy300;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy300;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy201:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy301;
				goto yy299;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy300;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy300;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy202:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '4') goto yy301;
				goto yy302;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy298;
				if (yych <= ':') goto yy299;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy300;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy300;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy203:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy300;
			if (yych <= ':') goto yy299;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy300;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy300;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy204:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy300;
			} else {
				if (yych <= ':') goto yy299;
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy300;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy303;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'e') goto yy300;
				goto yy303;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych >= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy205:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy198;
	goto yy1;
yy206:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy205;
	goto yy1;
yy207:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy205;
	goto yy1;
yy208:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy207;
	goto yy1;
yy209:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy207;
	goto yy1;
yy210:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy207;
	goto yy1;
yy211:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy305;
			goto yy308;
		} else {
			if (yych <= '/') goto yy305;
			if (yych <= ':') goto yy308;
			goto yy305;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy308;
			if (yych <= '^') goto yy305;
			goto yy308;
		} else {
			if (yych <= '`') goto yy305;
			if (yych <= 'z') goto yy308;
			goto yy305;
		}
	}
yy212:
	yych = *++YYCURSOR;
yy213:
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy212;
			} else {
				if (yych == '\r') goto yy212;
				goto yy1;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy212;
				goto yy1;
			} else {
				if (yych <= '-') goto yy311;
				if (yych <= '.') goto yy1;
				goto yy215;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= ':') goto yy311;
				goto yy1;
			} else {
				if (yych <= '>') goto yy216;
				if (yych <= '?') goto yy215;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy311;
				if (yych <= '^') goto yy1;
				goto yy311;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy311;
				goto yy1;
			}
		}
	}
yy214:
	yych = *++YYCURSOR;
	if (yybm[512+yych] & 16) {
		goto yy214;
	}
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy312;
			goto yy1;
		} else {
			if (yych <= '\r') goto yy312;
			if (yych <= 0x1F) goto yy1;
			goto yy312;
		}
	} else {
		if (yych <= '<') {
			if (yych != '/') goto yy1;
		} else {
			if (yych <= '=') goto yy313;
			if (yych <= '>') goto yy216;
			if (yych >= '@') goto yy1;
		}
	}
yy215:
	yych = *++YYCURSOR;
	if (yych == '>') goto yy314;
	goto yy1;
yy216:
	++YYCURSOR;
yy217:
#line 200 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_OPEN_TAG);
       }
#line 6460 "../../lnav/src/data_scanner_re.cc"
yy218:
	yych = *++YYCURSOR;
yy219:
	if (yych <= '/') {
		if (yych <= '\r') {
			if (yych == '\t') goto yy316;
			if (yych <= '\f') goto yy1;
			goto yy316;
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1;
				goto yy316;
			} else {
				if (yych == '-') goto yy218;
				goto yy1;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '=') {
				if (yych <= ':') goto yy218;
				goto yy1;
			} else {
				if (yych <= '>') goto yy317;
				if (yych <= '@') goto yy1;
				goto yy218;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1;
				goto yy218;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy218;
				goto yy1;
			}
		}
	}
yy220:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy319;
			goto yy320;
		} else {
			if (yych <= '/') goto yy319;
			if (yych <= ':') goto yy320;
			goto yy319;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy320;
			if (yych <= '^') goto yy319;
			goto yy320;
		} else {
			if (yych <= '`') goto yy319;
			if (yych <= 'z') goto yy320;
			goto yy319;
		}
	}
yy221:
	++YYCURSOR;
	yyt3 = yyt4;
yy222:
	YYCURSOR = yyt3;
#line 264 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_WORD); }
#line 6528 "../../lnav/src/data_scanner_re.cc"
yy223:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy221;
		yyt3 = yyt4;
		goto yy222;
	} else {
		if (yych <= '\r') goto yy221;
		if (yych == ' ') goto yy221;
		yyt3 = yyt4;
		goto yy222;
	}
yy224:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
yy225:
	if (yych <= '\'') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt3 = yyt4;
				goto yy222;
			} else {
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '\f') {
					yyt3 = yyt4;
					goto yy222;
				}
				yyt4 = YYCURSOR;
				goto yy221;
			}
		} else {
			if (yych <= '!') {
				if (yych <= 0x1F) {
					yyt3 = yyt4;
					goto yy222;
				}
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			} else {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '&') {
					yyt3 = yyt4;
					goto yy222;
				}
				yyt4 = YYCURSOR;
				goto yy224;
			}
		}
	} else {
		if (yych <= '9') {
			if (yych <= ',') {
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') {
					yyt3 = yyt4;
					goto yy222;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			} else {
				if (yych != '.') {
					yyt3 = yyt4;
					goto yy222;
				}
				yyt3 = yyt4;
				yyt4 = YYCURSOR;
			}
		} else {
			if (yych <= '?') {
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') {
					yyt3 = yyt4;
					goto yy222;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			} else {
				if (yych <= '`') {
					yyt3 = yyt4;
					goto yy222;
				}
				if (yych <= 'z') {
					yyt3 = yyt4;
					goto yy137;
				}
				yyt3 = yyt4;
				goto yy222;
			}
		}
	}
yy226:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy221;
		goto yy1;
	} else {
		if (yych <= '\r') goto yy221;
		if (yych == ' ') goto yy221;
		goto yy1;
	}
yy227:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy321;
				goto yy147;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy321;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy321;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy228:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy4;
				goto yy322;
			} else {
				if (yych == '%') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy80;
				goto yy4;
			} else {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				goto yy321;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy147;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy321;
				goto yy115;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy321;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy229:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 64) {
		goto yy115;
	}
	if (yych <= ',') {
		if (yych <= '$') {
			if (yych == ' ') goto yy322;
			goto yy4;
		} else {
			if (yych <= '%') goto yy80;
			if (yych == '+') goto yy80;
			goto yy4;
		}
	} else {
		if (yych <= '?') {
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			if (yych <= ':') goto yy148;
			goto yy4;
		} else {
			if (yych <= '@') goto yy96;
			if (yych == '_') goto yy46;
			goto yy4;
		}
	}
yy230:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[1024+yych] & 16) {
		goto yy46;
	}
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy221;
			goto yy4;
		} else {
			if (yych <= '\r') goto yy221;
			if (yych <= 0x1F) goto yy4;
			goto yy221;
		}
	} else {
		if (yych <= '*') {
			if (yych == '%') goto yy80;
			goto yy4;
		} else {
			if (yych <= '+') goto yy80;
			if (yych == '@') goto yy96;
			goto yy4;
		}
	}
yy231:
	yyaccept = 16;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '.') {
			yyt3 = yyt4;
			goto yy222;
		}
		if (yych <= '/') goto yy189;
		if (yych <= '9') goto yy272;
		goto yy191;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') {
				yyt3 = yyt4;
				goto yy222;
			}
			goto yy272;
		} else {
			if (yych <= '`') {
				yyt3 = yyt4;
				goto yy222;
			}
			if (yych <= 'f') goto yy272;
			yyt3 = yyt4;
			goto yy222;
		}
	}
yy232:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy323;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy321;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy324;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy321;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy4;
				} else {
					if (yych <= 'f') goto yy325;
					if (yych <= 'z') goto yy326;
					goto yy4;
				}
			}
		}
	}
yy233:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= ' ') {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				yyt4 = YYCURSOR;
				goto yy221;
			} else {
				if (yych == '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				yyt3 = YYCURSOR;
				goto yy323;
			}
		} else {
			if (yych <= '%') {
				if (yych <= '!') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych <= '&') goto yy4;
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy80;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				goto yy4;
			} else {
				if (yych <= '9') goto yy115;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			} else {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy326;
				goto yy4;
			}
		}
	}
yy234:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy189;
	if (yych == ':') goto yy152;
	goto yy222;
yy235:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy4;
				goto yy322;
			} else {
				if (yych == '%') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy80;
				goto yy4;
			} else {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				goto yy115;
			}
		}
	} else {
		if (yych <= 'S') {
			if (yych <= '?') {
				if (yych <= ':') goto yy148;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'R') goto yy115;
				goto yy327;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy236:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy323;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy115;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= 'r') {
					if (yych <= '`') goto yy4;
					goto yy326;
				} else {
					if (yych <= 's') goto yy328;
					if (yych <= 'z') goto yy326;
					goto yy4;
				}
			}
		}
	}
yy237:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy4;
				goto yy322;
			} else {
				if (yych == '%') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy80;
				goto yy4;
			} else {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				goto yy115;
			}
		}
	} else {
		if (yych <= 'L') {
			if (yych <= '?') {
				if (yych <= ':') goto yy148;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'K') goto yy115;
				goto yy329;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy238:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy323;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy115;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= 'd') {
					if (yych <= '`') goto yy4;
					goto yy326;
				} else {
					if (yych <= 'e') goto yy330;
					if (yych <= 'z') goto yy326;
					goto yy4;
				}
			}
		}
	}
yy239:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy4;
				goto yy322;
			} else {
				if (yych == '%') goto yy80;
				goto yy4;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy80;
				goto yy4;
			} else {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				goto yy115;
			}
		}
	} else {
		if (yych <= 'E') {
			if (yych <= '?') {
				if (yych <= ':') goto yy148;
				goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'D') goto yy115;
				goto yy329;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy240:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy323;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy115;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= 'k') {
					if (yych <= '`') goto yy4;
					goto yy326;
				} else {
					if (yych <= 'l') goto yy330;
					if (yych <= 'z') goto yy326;
					goto yy4;
				}
			}
		}
	}
yy241:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x1B) {
		if (yych <= 0x00) goto yy222;
		if (yych <= 0x1A) goto yy84;
		goto yy222;
	} else {
		if (yych <= 0x7F) goto yy84;
		if (yych <= 0xC1) goto yy222;
		if (yych <= 0xF4) goto yy84;
		goto yy222;
	}
yy242:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x1A) {
		if (yych <= '\t') {
			if (yych <= 0x00) goto yy222;
			if (yych <= 0x08) goto yy84;
			goto yy241;
		} else {
			if (yych == '\r') goto yy241;
			goto yy84;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1B) goto yy222;
			if (yych <= 0x1F) goto yy84;
			goto yy241;
		} else {
			if (yych <= 0x7F) goto yy84;
			if (yych <= 0xC1) goto yy222;
			if (yych <= 0xF4) goto yy84;
			goto yy222;
		}
	}
yy243:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '&') {
				if (yych == '!') {
					yyt1 = yyt4 = YYCURSOR;
					goto yy331;
				}
				yyt2 = YYCURSOR;
				goto yy160;
			} else {
				if (yych <= '\'') {
					yyt1 = YYCURSOR;
					goto yy332;
				}
				if (yych <= '+') {
					yyt2 = YYCURSOR;
					goto yy160;
				}
				yyt1 = yyt4 = YYCURSOR;
				goto yy331;
			}
		} else {
			if (yych <= '>') {
				if (yych == '.') {
					yyt1 = yyt4 = YYCURSOR;
					goto yy331;
				}
				yyt2 = YYCURSOR;
				goto yy160;
			} else {
				if (yych <= '?') {
					yyt1 = yyt4 = YYCURSOR;
					goto yy331;
				}
				if (yych <= 'R') {
					yyt2 = YYCURSOR;
					goto yy160;
				}
				goto yy222;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 's') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy160;
				}
				if (yych <= 'r') {
					yyt1 = YYCURSOR;
					goto yy333;
				}
				goto yy137;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy333;
				}
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy160;
				}
				goto yy222;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy163;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy164;
				}
				yyt2 = YYCURSOR;
				goto yy165;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy166;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy167;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy168;
				}
				goto yy222;
			}
		}
	}
yy244:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy241;
		goto yy84;
	} else {
		if (yych <= '\r') goto yy241;
		if (yych == ' ') goto yy241;
		goto yy84;
	}
yy245:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych == '/') goto yy4;
			goto yy46;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy334;
			goto yy4;
		} else {
			if (yych <= '_') goto yy46;
			if (yych <= '`') goto yy4;
			if (yych <= 'z') goto yy334;
			goto yy4;
		}
	}
yy246:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy335;
	goto yy1;
yy247:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy336;
	goto yy1;
yy248:
	++YYCURSOR;
#line 208 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_H1);
       }
#line 7589 "../../lnav/src/data_scanner_re.cc"
yy249:
	yych = *++YYCURSOR;
	if (yybm[512+yych] & 64) {
		goto yy249;
	}
	if (yych <= '9') {
		if (yych == '-') goto yy158;
		if (yych <= '/') goto yy1;
		goto yy158;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy1;
			goto yy337;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'z') goto yy337;
			goto yy1;
		}
	}
yy250:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy26;
				goto yy3;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy3;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy3;
					goto yy26;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy3;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy26;
				}
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy46;
					goto yy3;
				} else {
					if (yych <= '9') goto yy250;
					if (yych <= ':') goto yy62;
					goto yy26;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '^') {
				if (yych <= 'D') {
					if (yych <= '@') goto yy96;
					goto yy46;
				} else {
					if (yych <= 'E') goto yy339;
					if (yych <= 'Z') goto yy46;
					goto yy26;
				}
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy26;
				} else {
					if (yych == 'e') goto yy339;
					goto yy46;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy3;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xDF) goto yy56;
					goto yy57;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy26;
				}
			}
		}
	}
yy251:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy340;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy46;
			goto yy4;
		}
	}
yy252:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy5;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy5;
				goto yy3;
			} else {
				if (yych <= '\r') goto yy5;
				if (yych == 0x1B) goto yy5;
				goto yy3;
			}
		} else {
			if (yych <= '-') {
				if (yych == '$') goto yy3;
				if (yych <= ',') goto yy5;
				goto yy174;
			} else {
				if (yych <= '.') goto yy252;
				if (yych <= '/') goto yy3;
				if (yych <= '9') goto yy174;
				goto yy62;
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '?') goto yy5;
				if (yych <= '@') goto yy3;
				if (yych <= 'Z') goto yy341;
				goto yy5;
			} else {
				if (yych <= '_') goto yy3;
				if (yych <= '`') goto yy5;
				if (yych <= 'z') goto yy341;
				goto yy5;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy5;
				}
			}
		}
	}
yy253:
	yych = *++YYCURSOR;
	if (yych == 'o') goto yy342;
	goto yy100;
yy254:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy24;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy24;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy24;
					if (yych <= 0x1F) goto yy4;
					goto yy24;
				}
			}
		} else {
			if (yych <= '$') {
				if (yych == '"') goto yy24;
				if (yych <= '#') goto yy99;
				goto yy101;
			} else {
				if (yych <= '\'') {
					if (yych <= '&') goto yy99;
					goto yy24;
				} else {
					if (yych <= '*') goto yy99;
					if (yych <= ',') goto yy24;
					goto yy101;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= '[') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy24;
				if (yych <= 'Z') goto yy101;
				goto yy24;
			} else {
				if (yych <= ']') {
					if (yych <= '\\') goto yy99;
					goto yy24;
				} else {
					if (yych <= '^') goto yy99;
					if (yych <= '_') goto yy101;
					goto yy24;
				}
			}
		} else {
			if (yych <= '}') {
				if (yych == 'g') goto yy343;
				if (yych <= 'z') goto yy101;
				goto yy24;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy99;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy24;
					if (yych <= 0xF4) goto yy4;
					goto yy24;
				}
			}
		}
	}
yy255:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy344;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy344;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy344;
			goto yy4;
		}
	}
yy256:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '0') goto yy346;
				goto yy347;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') {
				if (yych <= '2') goto yy348;
				goto yy346;
			} else {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy46;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy257:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				goto yy26;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy26;
				} else {
					if (yych <= '-') goto yy255;
					if (yych <= '.') goto yy256;
					goto yy4;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy349;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'E') {
					if (yych >= 'E') goto yy259;
				} else {
					if (yych <= 'Z') goto yy258;
					if (yych <= '^') goto yy26;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') goto yy26;
				if (yych == 'e') goto yy259;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy258:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[512+yych] & 128) {
		goto yy258;
	}
	if (yych <= '-') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy5;
				if (yych <= 0x08) goto yy3;
				goto yy5;
			} else {
				if (yych == '\r') goto yy5;
				if (yych <= 0x1A) goto yy3;
				goto yy5;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy3;
				if (yych <= '#') goto yy5;
				if (yych <= '$') goto yy3;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy5;
				goto yy255;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= ':') {
				if (yych <= '.') goto yy350;
				if (yych <= '/') goto yy3;
				goto yy62;
			} else {
				if (yych == '@') goto yy96;
				if (yych <= '~') goto yy5;
				goto yy3;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) goto yy56;
				if (yych <= 0xE0) goto yy57;
				goto yy58;
			} else {
				if (yych <= 0xF0) goto yy59;
				if (yych <= 0xF3) goto yy60;
				if (yych <= 0xF4) goto yy61;
				goto yy5;
			}
		}
	}
yy259:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[512+yych] & 128) {
		goto yy258;
	}
	if (yych <= ',') {
		if (yych <= '%') {
			if (yych <= '$') goto yy4;
			goto yy80;
		} else {
			if (yych == '+') goto yy351;
			goto yy4;
		}
	} else {
		if (yych <= '.') {
			if (yych <= '-') goto yy352;
			goto yy350;
		} else {
			if (yych == '@') goto yy96;
			goto yy4;
		}
	}
yy260:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '5') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy255;
				} else {
					if (yych <= '.') goto yy256;
					if (yych <= '/') goto yy4;
					goto yy257;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy349;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy258;
					goto yy259;
				} else {
					if (yych <= 'Z') goto yy258;
					if (yych <= '^') goto yy26;
					goto yy258;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') goto yy26;
				if (yych == 'e') goto yy259;
				goto yy258;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy261:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy5;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy5;
				goto yy3;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy5;
					goto yy3;
				} else {
					if (yych <= 0x1B) goto yy5;
					if (yych <= 0x1F) goto yy3;
					goto yy5;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				if (yych <= '*') goto yy5;
				goto yy80;
			} else {
				if (yych <= '.') {
					if (yych <= ',') goto yy5;
					goto yy46;
				} else {
					if (yych <= '/') goto yy3;
					if (yych <= '9') goto yy353;
					goto yy62;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '?') goto yy5;
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy353;
				goto yy46;
			} else {
				if (yych <= '_') {
					if (yych <= '^') goto yy5;
					goto yy46;
				} else {
					if (yych <= '`') goto yy5;
					if (yych <= 'f') goto yy353;
					goto yy46;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy5;
					goto yy3;
				} else {
					if (yych <= 0xC1) goto yy5;
					if (yych <= 0xDF) goto yy56;
					goto yy57;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy5;
				}
			}
		}
	}
yy262:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				goto yy26;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy26;
				} else {
					if (yych <= '-') goto yy255;
					if (yych <= '.') goto yy256;
					goto yy4;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy355;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy258;
					goto yy356;
				} else {
					if (yych <= 'Z') goto yy258;
					if (yych <= '^') goto yy26;
					goto yy258;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') goto yy26;
				if (yych == 'e') goto yy356;
				goto yy258;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy263:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				goto yy26;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy26;
				} else {
					if (yych <= '-') goto yy255;
					if (yych <= '.') goto yy256;
					goto yy4;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy262;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy258;
					goto yy356;
				} else {
					if (yych <= 'Z') goto yy258;
					if (yych <= '^') goto yy26;
					goto yy258;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') goto yy26;
				if (yych == 'e') goto yy356;
				goto yy258;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy264:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '4') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy255;
				} else {
					if (yych <= '.') goto yy256;
					if (yych <= '/') goto yy4;
					goto yy262;
				}
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '5') goto yy357;
				if (yych <= '9') goto yy355;
				if (yych <= ':') goto yy4;
				goto yy26;
			} else {
				if (yych <= 'D') {
					if (yych <= '@') goto yy96;
					goto yy258;
				} else {
					if (yych <= 'E') goto yy356;
					if (yych <= 'Z') goto yy258;
					goto yy26;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') {
					if (yych <= '_') goto yy258;
					goto yy26;
				} else {
					if (yych == 'e') goto yy356;
					goto yy258;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy265:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy4;
	if (yych <= 'Z') goto yy358;
	if (yych <= '`') goto yy4;
	if (yych <= 'z') goto yy358;
	goto yy4;
yy266:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy359;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy46;
			goto yy4;
		}
	}
yy267:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '7') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy360;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') {
					if (yych <= '#') goto yy109;
					goto yy4;
				} else {
					if (yych <= '%') goto yy105;
					if (yych <= '*') goto yy109;
					goto yy80;
				}
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy109;
					goto yy361;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy362;
					goto yy363;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') {
					if (yych <= '9') goto yy364;
					goto yy365;
				} else {
					if (yych <= '?') goto yy109;
					if (yych <= '@') goto yy96;
					goto yy366;
				}
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy367;
					goto yy366;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy366;
				} else {
					if (yych <= 'e') goto yy367;
					if (yych <= 'f') goto yy366;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy268:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy360;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') {
					if (yych <= '#') goto yy26;
					goto yy4;
				} else {
					if (yych <= '%') goto yy105;
					if (yych <= '*') goto yy26;
					goto yy80;
				}
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy361;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy362;
					goto yy364;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy365;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy366;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy367;
					goto yy366;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy366;
				} else {
					if (yych <= 'e') goto yy367;
					if (yych <= 'f') goto yy366;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy269:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy113;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy366;
				if (yych <= ':') goto yy365;
				if (yych <= '?') goto yy113;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy366;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy113;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy366;
				if (yych <= 'z') goto yy115;
				goto yy113;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy113;
				if (yych <= 0xF4) goto yy4;
				goto yy113;
			}
		}
	}
yy270:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy251;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy366;
				if (yych <= ':') goto yy365;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy366;
				if (yych <= 'Z') goto yy115;
				goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy366;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy271:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy368;
		if (yych <= ':') goto yy282;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy369;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy369;
		goto yy1;
	}
yy272:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy369;
		if (yych <= ':') goto yy282;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy369;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy369;
		goto yy1;
	}
yy273:
	yych = *++YYCURSOR;
yy274:
	if (yybm[256+yych] & 1) {
		goto yy273;
	}
	if (yych <= 0xC1) {
		if (yych <= ')') {
			if (yych <= '"') goto yy1;
			if (yych <= '&') goto yy370;
			goto yy1;
		} else {
			if (yych <= 'Z') goto yy370;
			if (yych <= ']') goto yy1;
			if (yych <= 'z') goto yy370;
			goto yy1;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy275;
			if (yych <= 0xE0) goto yy276;
			goto yy277;
		} else {
			if (yych <= 0xF0) goto yy278;
			if (yych <= 0xF3) goto yy279;
			if (yych <= 0xF4) goto yy280;
			goto yy1;
		}
	}
yy275:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy273;
	goto yy1;
yy276:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy275;
	goto yy1;
yy277:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy275;
	goto yy1;
yy278:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy277;
	goto yy1;
yy279:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy277;
	goto yy1;
yy280:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy277;
	goto yy1;
yy281:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy372;
		if (yych <= ':') goto yy373;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy372;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy372;
		goto yy1;
	}
yy282:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy374;
		if (yych <= ':') goto yy375;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy374;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy374;
		goto yy1;
	}
yy283:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy372;
		if (yych <= ':') goto yy282;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy372;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy372;
		goto yy1;
	}
yy284:
	yyaccept = 17;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy199;
				goto yy4;
			} else {
				if (yych <= '\n') goto yy199;
				if (yych <= '\f') goto yy4;
				goto yy199;
			}
		} else {
			if (yych <= 0x1F) {
				if (yych == 0x1B) goto yy199;
				goto yy4;
			} else {
				if (yych == '$') goto yy4;
				goto yy199;
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy199;
				goto yy4;
			} else {
				if (yych == '_') goto yy4;
				goto yy199;
			}
		} else {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy4;
				if (yych <= '~') goto yy199;
				goto yy4;
			} else {
				if (yych <= 0xC1) goto yy199;
				if (yych <= 0xF4) goto yy4;
				goto yy199;
			}
		}
	}
yy285:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '-') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy377;
					if (yych <= ':') goto yy378;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy379;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy379;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy286:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '-') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy380;
					if (yych <= ':') goto yy378;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy379;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy379;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy287:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '.') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy376;
			} else {
				if (yych <= '4') {
					if (yych <= '/') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					goto yy380;
				} else {
					if (yych <= '5') goto yy381;
					if (yych <= '9') goto yy377;
					goto yy378;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy379;
				goto yy3;
			} else {
				if (yych <= '_') {
					if (yych <= '^') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '`') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 'f') goto yy379;
					goto yy3;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0xC1) goto yy5;
					if (yych <= 0xDF) {
						yyt2 = YYCURSOR;
						goto yy289;
					}
					yyt2 = YYCURSOR;
					goto yy290;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy288:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '/') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '9') goto yy379;
				if (yych <= ':') goto yy378;
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '_') {
				if (yych <= 'F') goto yy379;
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy379;
				if (yych <= 'z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych >= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy290;
				}
				yyt2 = YYCURSOR;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy289:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy284;
	goto yy1;
yy290:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy289;
	goto yy1;
yy291:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy289;
	goto yy1;
yy292:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy291;
	goto yy1;
yy293:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy291;
	goto yy1;
yy294:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy291;
	goto yy1;
yy295:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy26;
			goto yy80;
		} else {
			if (yych <= ',') goto yy26;
			if (yych <= '.') goto yy80;
			if (yych <= '/') goto yy26;
			goto yy295;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy26;
			if (yych <= '@') goto yy82;
			if (yych <= 'Z') goto yy80;
			goto yy26;
		} else {
			if (yych == '`') goto yy26;
			if (yych <= 'z') goto yy80;
			goto yy26;
		}
	}
yy296:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy26;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy26;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy382;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy353;
				if (yych <= 'Z') goto yy46;
				if (yych <= '^') goto yy26;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy26;
				if (yych <= 'f') goto yy353;
				if (yych <= 'z') goto yy46;
				goto yy26;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy26;
				if (yych <= 0xF4) goto yy4;
				goto yy26;
			}
		}
	}
yy297:
	yyaccept = 18;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '/') {
			yyt2 = yyt1;
			goto yy199;
		}
		if (yych <= '0') goto yy383;
		goto yy384;
	} else {
		if (yych <= '2') goto yy385;
		if (yych <= '9') goto yy383;
		yyt2 = yyt1;
		goto yy199;
	}
yy298:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy386;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy386;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy386;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy299:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy387;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy387;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy387;
		goto yy1;
	}
yy300:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy386;
			if (yych <= ':') goto yy299;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy386;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy386;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy301:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy388;
				goto yy299;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy386;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy386;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy302:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '5') goto yy388;
				goto yy386;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy299;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy386;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy386;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy303:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy386;
			} else {
				if (yych <= ':') goto yy299;
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy386;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy389;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'e') goto yy386;
				goto yy389;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy304:
	yych = *++YYCURSOR;
yy305:
	if (yych <= ',') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy304;
			} else {
				if (yych == '\r') goto yy304;
				goto yy1;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy304;
				if (yych <= '!') goto yy1;
			} else {
				if (yych == '\'') goto yy307;
				goto yy1;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= ':') {
				if (yych <= '-') goto yy390;
				if (yych <= '/') goto yy1;
				goto yy390;
			} else {
				if (yych == '>') goto yy309;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy390;
				if (yych <= '^') goto yy1;
				goto yy390;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy390;
				goto yy1;
			}
		}
	}
yy306:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 8) {
		goto yy391;
	}
	if (yych <= 0xE0) {
		if (yych <= '\\') {
			if (yych <= '"') goto yy1;
			goto yy392;
		} else {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy393;
			goto yy394;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy395;
			goto yy396;
		} else {
			if (yych <= 0xF3) goto yy397;
			if (yych <= 0xF4) goto yy398;
			goto yy1;
		}
	}
yy307:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 16) {
		goto yy399;
	}
	if (yych <= 0xE0) {
		if (yych <= '\\') {
			if (yych <= '\'') goto yy1;
			goto yy400;
		} else {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy401;
			goto yy402;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy403;
			goto yy404;
		} else {
			if (yych <= 0xF3) goto yy405;
			if (yych <= 0xF4) goto yy406;
			goto yy1;
		}
	}
yy308:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy407;
				goto yy1;
			} else {
				if (yych <= '\r') goto yy407;
				if (yych <= 0x1F) goto yy1;
				goto yy407;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy306;
				goto yy1;
			} else {
				if (yych <= '\'') goto yy307;
				if (yych <= ',') goto yy1;
				goto yy308;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy1;
				if (yych <= ':') goto yy308;
				goto yy1;
			} else {
				if (yych <= '=') goto yy408;
				if (yych >= '?') goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy308;
				if (yych <= '^') goto yy1;
				goto yy308;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy308;
				goto yy1;
			}
		}
	}
yy309:
	++YYCURSOR;
yy310:
#line 192 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_DECL_TAG);
       }
#line 10017 "../../lnav/src/data_scanner_re.cc"
yy311:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy409;
			} else {
				if (yych == '\r') goto yy409;
				goto yy1;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy409;
				if (yych <= ',') goto yy1;
				goto yy311;
			} else {
				if (yych <= '.') goto yy1;
				if (yych <= '/') goto yy215;
				goto yy311;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1;
				goto yy313;
			} else {
				if (yych <= '>') goto yy216;
				if (yych <= '?') goto yy215;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy311;
				if (yych <= '^') goto yy1;
				goto yy311;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy311;
				goto yy1;
			}
		}
	}
yy312:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy312;
			} else {
				if (yych == '\r') goto yy312;
				goto yy1;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy312;
				if (yych <= ',') goto yy1;
				goto yy311;
			} else {
				if (yych <= '.') goto yy1;
				if (yych <= '/') goto yy215;
				goto yy311;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1;
			} else {
				if (yych <= '>') goto yy216;
				if (yych <= '?') goto yy215;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy311;
				if (yych <= '^') goto yy1;
				goto yy311;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy311;
				goto yy1;
			}
		}
	}
yy313:
	yych = *++YYCURSOR;
	if (yych <= '\'') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1;
				goto yy410;
			} else {
				if (yych <= '\t') goto yy411;
				if (yych <= '\f') goto yy410;
				goto yy411;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy411;
				goto yy410;
			} else {
				if (yych <= '"') goto yy412;
				if (yych <= '&') goto yy410;
				goto yy413;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '>') {
				if (yych <= '=') goto yy410;
				goto yy1;
			} else {
				if (yych <= 0x7F) goto yy410;
				if (yych <= 0xC1) goto yy1;
				goto yy414;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy415;
				if (yych <= 0xEF) goto yy416;
				goto yy417;
			} else {
				if (yych <= 0xF3) goto yy418;
				if (yych <= 0xF4) goto yy419;
				goto yy1;
			}
		}
	}
yy314:
	++YYCURSOR;
yy315:
#line 196 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_EMPTY_TAG);
       }
#line 10156 "../../lnav/src/data_scanner_re.cc"
yy316:
	yych = *++YYCURSOR;
	if (yych <= '\r') {
		if (yych == '\t') goto yy316;
		if (yych <= '\f') goto yy1;
		goto yy316;
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1;
			goto yy316;
		} else {
			if (yych != '>') goto yy1;
		}
	}
yy317:
	++YYCURSOR;
#line 204 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_CLOSE_TAG);
       }
#line 10177 "../../lnav/src/data_scanner_re.cc"
yy318:
	yych = *++YYCURSOR;
yy319:
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy318;
			} else {
				if (yych == '\r') goto yy318;
				goto yy1;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy318;
				goto yy1;
			} else {
				if (yych <= '-') goto yy420;
				if (yych <= '.') goto yy1;
				goto yy215;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '>') {
				if (yych <= ':') goto yy420;
				goto yy1;
			} else {
				if (yych <= '?') goto yy215;
				if (yych <= '@') goto yy1;
				goto yy420;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1;
				goto yy420;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy420;
				goto yy1;
			}
		}
	}
yy320:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy421;
			} else {
				if (yych == '\r') goto yy421;
				goto yy1;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy421;
				if (yych <= ',') goto yy1;
				goto yy320;
			} else {
				if (yych <= '.') goto yy1;
				if (yych <= '/') goto yy215;
				goto yy320;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1;
				goto yy422;
			} else {
				if (yych == '?') goto yy215;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy320;
				if (yych <= '^') goto yy1;
				goto yy320;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy320;
				goto yy1;
			}
		}
	}
yy321:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy423;
				goto yy365;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy423;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy423;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy322:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy424;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy424;
	goto yy1;
yy323:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy424;
	if (yych <= '/') goto yy222;
	if (yych <= '9') goto yy424;
	goto yy222;
yy324:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '.') goto yy222;
		if (yych <= '/') goto yy189;
		if (yych <= '9') goto yy192;
		goto yy191;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy222;
			goto yy192;
		} else {
			if (yych <= '`') goto yy222;
			if (yych <= 'f') goto yy192;
			goto yy222;
		}
	}
yy325:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy423;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy425;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy423;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy4;
				} else {
					if (yych <= 'f') goto yy426;
					if (yych >= '{') goto yy4;
				}
			}
		}
	}
yy326:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= '!') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy3;
				} else {
					if (yych <= '\n') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					if (yych <= '\f') goto yy3;
					yyt4 = YYCURSOR;
					goto yy221;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy3;
					goto yy5;
				} else {
					if (yych <= 0x1F) goto yy3;
					if (yych <= ' ') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					yyt4 = YYCURSOR;
					goto yy223;
				}
			}
		} else {
			if (yych <= '&') {
				if (yych <= '#') {
					if (yych <= '"') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy5;
				} else {
					if (yych <= '$') goto yy3;
					if (yych <= '%') goto yy80;
					goto yy5;
				}
			} else {
				if (yych <= '+') {
					if (yych <= '\'') {
						yyt4 = YYCURSOR;
						goto yy224;
					}
					if (yych <= '*') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy80;
				} else {
					if (yych <= ',') {
						yyt4 = YYCURSOR;
						goto yy223;
					}
					if (yych <= '-') goto yy46;
					yyt4 = YYCURSOR;
					goto yy230;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= '>') {
				if (yych <= '9') {
					if (yych <= '/') goto yy3;
					goto yy115;
				} else {
					if (yych <= ':') {
						yyt3 = YYCURSOR;
						goto yy234;
					}
					if (yych <= ';') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy5;
				}
			} else {
				if (yych <= 'Z') {
					if (yych <= '?') {
						yyt4 = YYCURSOR;
						goto yy223;
					}
					if (yych <= '@') goto yy96;
					goto yy115;
				} else {
					if (yych == '_') goto yy46;
					goto yy5;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= '~') {
					if (yych <= 'z') goto yy326;
					goto yy5;
				} else {
					if (yych <= 0x7F) goto yy3;
					if (yych <= 0xC1) goto yy5;
					goto yy56;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) goto yy57;
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy5;
				}
			}
		}
	}
yy327:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy115;
				goto yy148;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych == 'E') goto yy329;
				goto yy115;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy328:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy115;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= 'd') {
					if (yych <= '`') goto yy4;
					goto yy326;
				} else {
					if (yych <= 'e') goto yy330;
					if (yych <= 'z') goto yy326;
					goto yy4;
				}
			}
		}
	}
yy329:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 64) {
		goto yy115;
	}
	if (yych <= '*') {
		if (yych <= 0x1F) {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt1 = YYCURSOR;
					goto yy427;
				}
				if (yych <= 0x08) goto yy4;
				yyt1 = YYCURSOR;
				goto yy427;
			} else {
				if (yych == '\r') {
					yyt1 = YYCURSOR;
					goto yy427;
				}
				goto yy4;
			}
		} else {
			if (yych <= '"') {
				if (yych == '!') {
					yyt1 = YYCURSOR;
					goto yy429;
				}
				yyt1 = YYCURSOR;
				goto yy427;
			} else {
				if (yych == '%') goto yy80;
				if (yych <= '&') goto yy4;
				yyt1 = YYCURSOR;
				goto yy427;
			}
		}
	} else {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= '+') goto yy80;
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy429;
				}
				goto yy46;
			} else {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy430;
				}
				if (yych <= '/') goto yy4;
				yyt1 = YYCURSOR;
				goto yy431;
			}
		} else {
			if (yych <= '?') {
				if (yych <= ';') {
					yyt1 = YYCURSOR;
					goto yy427;
				}
				if (yych <= '>') goto yy4;
				yyt1 = YYCURSOR;
				goto yy429;
			} else {
				if (yych <= '@') goto yy96;
				if (yych == '_') goto yy46;
				goto yy4;
			}
		}
	}
yy330:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= ' ') {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt1 = YYCURSOR;
					goto yy427;
				}
				if (yych <= 0x08) goto yy4;
				yyt1 = YYCURSOR;
				goto yy427;
			} else {
				if (yych == '\r') {
					yyt1 = YYCURSOR;
					goto yy427;
				}
				if (yych <= 0x1F) goto yy4;
				yyt1 = YYCURSOR;
				goto yy427;
			}
		} else {
			if (yych <= '%') {
				if (yych <= '!') {
					yyt1 = YYCURSOR;
					goto yy429;
				}
				if (yych <= '"') {
					yyt1 = YYCURSOR;
					goto yy427;
				}
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych <= '&') goto yy4;
				if (yych <= '\'') {
					yyt1 = YYCURSOR;
					goto yy432;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy427;
				}
				goto yy80;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy429;
				}
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy430;
				}
				goto yy4;
			} else {
				if (yych <= '9') goto yy115;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy431;
				}
				if (yych <= ';') {
					yyt1 = YYCURSOR;
					goto yy427;
				}
				goto yy4;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy429;
				}
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			} else {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy326;
				goto yy4;
			}
		}
	}
yy331:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy221;
		goto yy161;
	} else {
		if (yych <= '\r') goto yy221;
		if (yych == ' ') goto yy221;
		goto yy161;
	}
yy332:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy83;
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy241;
				}
				goto yy83;
			} else {
				if (yych <= '\r') {
					yyt3 = YYCURSOR;
					goto yy241;
				}
				if (yych == 0x1B) goto yy161;
				goto yy83;
			}
		} else {
			if (yych <= '&') {
				if (yych == '!') {
					yyt3 = YYCURSOR;
					goto yy242;
				}
				if (yych <= '"') {
					yyt3 = YYCURSOR;
					goto yy241;
				}
				goto yy83;
			} else {
				if (yych <= '*') {
					if (yych <= '\'') {
						yyt3 = YYCURSOR;
						goto yy243;
					}
					yyt3 = YYCURSOR;
					goto yy241;
				} else {
					if (yych == ',') {
						yyt3 = YYCURSOR;
						goto yy242;
					}
					goto yy83;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '>') {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy244;
				}
				if (yych <= '9') goto yy83;
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy241;
				}
				goto yy83;
			} else {
				if (yych <= '[') {
					if (yych <= '?') {
						yyt3 = YYCURSOR;
						goto yy242;
					}
					goto yy83;
				} else {
					if (yych <= '\\') goto yy86;
					if (yych <= '`') goto yy83;
					goto yy150;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy83;
				if (yych <= 0xC1) goto yy161;
				if (yych <= 0xDF) goto yy87;
				goto yy88;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy89;
					goto yy90;
				} else {
					if (yych <= 0xF3) goto yy91;
					if (yych <= 0xF4) goto yy92;
					goto yy161;
				}
			}
		}
	}
yy333:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 128) {
		goto yy137;
	}
	if (yych <= '*') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy225;
			if (yych <= 0x08) goto yy161;
			if (yych <= '\n') goto yy225;
			goto yy161;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy225;
				goto yy161;
			} else {
				if (yych <= '"') goto yy225;
				if (yych <= '&') goto yy161;
				goto yy225;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych == ',') goto yy225;
			if (yych <= '-') goto yy161;
			yyt4 = YYCURSOR;
			goto yy226;
		} else {
			if (yych <= ';') {
				if (yych <= '9') goto yy161;
				goto yy225;
			} else {
				if (yych == '?') goto yy225;
				goto yy161;
			}
		}
	}
yy334:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '\'') {
			if (yych == '%') goto yy80;
			if (yych <= '&') goto yy4;
			goto yy137;
		} else {
			if (yych <= '+') {
				if (yych <= '*') goto yy4;
				goto yy80;
			} else {
				if (yych <= ',') goto yy4;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy46;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy46;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy433;
				goto yy4;
			}
		}
	}
yy335:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy434;
			}
		} else {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy434;
			}
			if (yych <= 0xC1) goto yy1;
			yyt1 = YYCURSOR;
			goto yy436;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt1 = YYCURSOR;
				goto yy437;
			}
			if (yych <= 0xEF) {
				yyt1 = YYCURSOR;
				goto yy438;
			}
			yyt1 = YYCURSOR;
			goto yy439;
		} else {
			if (yych <= 0xF3) {
				yyt1 = YYCURSOR;
				goto yy440;
			}
			if (yych <= 0xF4) {
				yyt1 = YYCURSOR;
				goto yy441;
			}
			goto yy1;
		}
	}
yy336:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy442;
	goto yy1;
yy337:
	yyaccept = 19;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[512+yych] & 64) {
		goto yy249;
	}
	if (yych <= '9') {
		if (yych == '-') goto yy158;
		if (yych >= '0') goto yy158;
	} else {
		if (yych <= 'Z') {
			if (yych >= 'A') goto yy337;
		} else {
			if (yych <= '`') goto yy338;
			if (yych <= 'z') goto yy337;
		}
	}
yy338:
#line 260 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EMAIL); }
#line 11084 "../../lnav/src/data_scanner_re.cc"
yy339:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy194;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '-') goto yy251;
			if (yych <= '.') goto yy46;
			goto yy4;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy46;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy46;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy340:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy26;
				goto yy3;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy3;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				goto yy26;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy26;
				} else {
					if (yych <= '.') goto yy46;
					if (yych <= '/') goto yy3;
					goto yy340;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy62;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy46;
			} else {
				if (yych <= '_') {
					if (yych <= '^') goto yy26;
					goto yy46;
				} else {
					if (yych <= '`') goto yy26;
					if (yych <= 'z') goto yy46;
					goto yy26;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy26;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy26;
				}
			}
		}
	}
yy341:
	yyaccept = 19;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy338;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy338;
				goto yy3;
			} else {
				if (yych <= '\r') goto yy338;
				if (yych == 0x1B) goto yy338;
				goto yy3;
			}
		} else {
			if (yych <= '-') {
				if (yych == '$') goto yy3;
				if (yych <= ',') goto yy338;
				goto yy174;
			} else {
				if (yych <= '.') goto yy252;
				if (yych <= '/') goto yy3;
				if (yych <= '9') goto yy174;
				goto yy62;
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '?') goto yy338;
				if (yych <= '@') goto yy3;
				if (yych <= 'Z') goto yy341;
				goto yy338;
			} else {
				if (yych <= '_') goto yy3;
				if (yych <= '`') goto yy338;
				if (yych <= 'z') goto yy341;
				goto yy338;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy338;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy338;
				}
			}
		}
	}
yy342:
	yych = *++YYCURSOR;
	if (yych == 'g') goto yy443;
	goto yy100;
yy343:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy24;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy24;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy24;
					if (yych <= 0x1F) goto yy4;
					goto yy24;
				}
			}
		} else {
			if (yych <= '$') {
				if (yych == '"') goto yy24;
				if (yych <= '#') goto yy99;
				goto yy101;
			} else {
				if (yych <= '\'') {
					if (yych <= '&') goto yy99;
					goto yy24;
				} else {
					if (yych <= '*') goto yy99;
					if (yych <= ',') goto yy24;
					goto yy101;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= '[') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy24;
				if (yych <= 'Z') goto yy101;
				goto yy24;
			} else {
				if (yych <= ']') {
					if (yych <= '\\') goto yy99;
					goto yy24;
				} else {
					if (yych <= '^') goto yy99;
					if (yych <= '_') goto yy101;
					goto yy24;
				}
			}
		} else {
			if (yych <= '}') {
				if (yych == 'r') goto yy444;
				if (yych <= 'z') goto yy101;
				goto yy24;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy99;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy24;
					if (yych <= 0xF4) goto yy4;
					goto yy24;
				}
			}
		}
	}
yy344:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy345;
				if (yych <= 0x08) goto yy3;
				if (yych >= '\v') goto yy3;
			} else {
				if (yych <= '\r') goto yy345;
				if (yych != 0x1B) goto yy3;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy345;
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
				} else {
					if (yych <= '.') goto yy46;
					if (yych <= '/') goto yy3;
					goto yy344;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy62;
				if (yych <= '?') goto yy345;
				if (yych <= '@') goto yy96;
				goto yy344;
			} else {
				if (yych <= '_') {
					if (yych >= '_') goto yy344;
				} else {
					if (yych <= '`') goto yy345;
					if (yych <= 'z') goto yy344;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
				}
			}
		}
	}
yy345:
#line 251 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_VERSION_NUMBER);
       }
#line 11380 "../../lnav/src/data_scanner_re.cc"
yy346:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy345;
				if (yych <= 0x08) goto yy4;
				goto yy345;
			} else {
				if (yych == '\r') goto yy345;
				if (yych <= 0x1A) goto yy4;
				goto yy345;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy345;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy345;
				goto yy255;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '.') goto yy445;
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy446;
				goto yy4;
			} else {
				if (yych <= '?') goto yy345;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy447;
				goto yy345;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy345;
				if (yych <= 'z') goto yy447;
				goto yy345;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xF4) goto yy4;
				goto yy345;
			}
		}
	}
yy347:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy345;
				if (yych <= 0x08) goto yy4;
				goto yy345;
			} else {
				if (yych == '\r') goto yy345;
				if (yych <= 0x1A) goto yy4;
				goto yy345;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy345;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy345;
				goto yy255;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '.') goto yy445;
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy346;
				goto yy4;
			} else {
				if (yych <= '?') goto yy345;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy447;
				goto yy345;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy345;
				if (yych <= 'z') goto yy447;
				goto yy345;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xF4) goto yy4;
				goto yy345;
			}
		}
	}
yy348:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy345;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy345;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy345;
				if (yych == 0x1B) goto yy345;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy345;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy345;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy345;
				if (yych <= '-') goto yy255;
				goto yy445;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') {
				if (yych <= '/') goto yy4;
				if (yych <= '4') goto yy346;
				if (yych <= '5') goto yy448;
				goto yy446;
			} else {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy345;
				if (yych <= '@') goto yy96;
				goto yy447;
			}
		} else {
			if (yych <= 'z') {
				if (yych == '_') goto yy447;
				if (yych <= '`') goto yy345;
				goto yy447;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy345;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy345;
					if (yych <= 0xF4) goto yy4;
					goto yy345;
				}
			}
		}
	}
yy349:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '?') {
		if (yych <= '$') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) goto yy26;
					goto yy3;
				} else {
					if (yych <= '\n') goto yy26;
					if (yych <= '\f') goto yy3;
					goto yy26;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy3;
					goto yy26;
				} else {
					if (yych <= 0x1F) goto yy3;
					if (yych <= '#') goto yy26;
					goto yy3;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '*') {
					if (yych <= '%') goto yy105;
					goto yy26;
				} else {
					if (yych <= '+') goto yy80;
					if (yych <= ',') goto yy26;
					goto yy255;
				}
			} else {
				if (yych <= '/') {
					if (yych >= '/') goto yy3;
				} else {
					if (yych <= '9') goto yy349;
					if (yych <= ':') goto yy62;
					goto yy26;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '^') {
				if (yych <= 'D') {
					if (yych <= '@') goto yy96;
					goto yy258;
				} else {
					if (yych <= 'E') goto yy259;
					if (yych <= 'Z') goto yy258;
					goto yy26;
				}
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy258;
					goto yy26;
				} else {
					if (yych == 'e') goto yy259;
					goto yy258;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy3;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xDF) goto yy56;
					goto yy57;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy26;
				}
			}
		}
	}
yy350:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy5;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy5;
				goto yy3;
			} else {
				if (yych <= '\r') goto yy5;
				if (yych == 0x1B) goto yy5;
				goto yy3;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy5;
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				goto yy5;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy5;
				} else {
					if (yych <= '.') goto yy46;
					if (yych <= '/') goto yy3;
					goto yy447;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy62;
				if (yych <= '?') goto yy5;
				if (yych <= '@') goto yy96;
				goto yy46;
			} else {
				if (yych <= '_') {
					if (yych <= '^') goto yy5;
					goto yy46;
				} else {
					if (yych <= '`') goto yy5;
					if (yych <= 'z') goto yy46;
					goto yy5;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy5;
				}
			}
		}
	}
yy351:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy81;
	if (yych <= '9') goto yy449;
	goto yy81;
yy352:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy451;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy344;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy344;
			goto yy4;
		}
	}
yy353:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy354;
				if (yych <= 0x08) goto yy3;
				if (yych >= '\v') goto yy3;
			} else {
				if (yych <= 0x1A) {
					if (yych >= 0x0E) goto yy3;
				} else {
					if (yych <= 0x1B) goto yy354;
					if (yych <= 0x1F) goto yy3;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				if (yych >= '+') goto yy80;
			} else {
				if (yych <= '-') {
					if (yych >= '-') goto yy181;
				} else {
					if (yych == '/') goto yy3;
					goto yy46;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy452;
				if (yych <= '?') goto yy354;
				if (yych <= '@') goto yy96;
				goto yy46;
			} else {
				if (yych <= '_') {
					if (yych >= '_') goto yy46;
				} else {
					if (yych <= '`') goto yy354;
					if (yych <= 'z') goto yy46;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy354;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
				}
			}
		}
	}
yy354:
#line 174 "../../lnav/src/data_scanner_re.re"
	{
           if ((YYCURSOR.val - (this->ds_input.udata() + this->ds_next_offset)) == 17) {
               RET(DT_MAC_ADDRESS);
           } else {
               RET(DT_HEX_DUMP);
           }
       }
#line 11802 "../../lnav/src/data_scanner_re.cc"
yy355:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				goto yy26;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy26;
				} else {
					if (yych <= '-') goto yy255;
					if (yych <= '.') goto yy256;
					goto yy4;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy359;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy258;
				} else {
					if (yych <= 'Z') goto yy258;
					if (yych <= '^') goto yy26;
					goto yy258;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') goto yy26;
				if (yych != 'e') goto yy258;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy356:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[512+yych] & 128) {
		goto yy258;
	}
	if (yych <= ',') {
		if (yych <= '%') {
			if (yych <= '$') goto yy4;
			goto yy80;
		} else {
			if (yych == '+') goto yy194;
			goto yy4;
		}
	} else {
		if (yych <= '.') {
			if (yych <= '-') goto yy255;
			goto yy350;
		} else {
			if (yych == '@') goto yy96;
			goto yy4;
		}
	}
yy357:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '5') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy255;
				} else {
					if (yych <= '.') goto yy256;
					if (yych <= '/') goto yy4;
					goto yy355;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy359;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy258;
					goto yy356;
				} else {
					if (yych <= 'Z') goto yy258;
					if (yych <= '^') goto yy26;
					goto yy258;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '`') goto yy26;
				if (yych == 'e') goto yy356;
				goto yy258;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy358:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy4;
	if (yych <= 'Z') goto yy453;
	if (yych <= '`') goto yy4;
	if (yych <= 'z') goto yy453;
	goto yy4;
yy359:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '?') {
		if (yych <= '$') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) goto yy26;
					goto yy3;
				} else {
					if (yych <= '\n') goto yy26;
					if (yych <= '\f') goto yy3;
					goto yy26;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy3;
					goto yy26;
				} else {
					if (yych <= 0x1F) goto yy3;
					if (yych <= '#') goto yy26;
					goto yy3;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '*') {
					if (yych <= '%') goto yy105;
					goto yy26;
				} else {
					if (yych <= '+') goto yy80;
					if (yych <= ',') goto yy26;
					goto yy255;
				}
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy350;
					goto yy3;
				} else {
					if (yych <= '9') goto yy359;
					if (yych <= ':') goto yy62;
					goto yy26;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '^') {
				if (yych <= 'D') {
					if (yych <= '@') goto yy96;
					goto yy258;
				} else {
					if (yych <= 'E') goto yy356;
					if (yych <= 'Z') goto yy258;
					goto yy26;
				}
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy258;
					goto yy26;
				} else {
					if (yych == 'e') goto yy356;
					goto yy258;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy3;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xDF) goto yy56;
					goto yy57;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy26;
				}
			}
		}
	}
yy360:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy104;
	if (yych <= '9') goto yy454;
	goto yy104;
yy361:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy455;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy46;
			goto yy4;
		}
	}
yy362:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy456;
	goto yy4;
yy363:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy457;
					goto yy458;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy459;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy460;
					goto yy459;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy459;
				} else {
					if (yych <= 'e') goto yy460;
					if (yych <= 'f') goto yy459;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy364:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy458;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy459;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy460;
					goto yy459;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy459;
				} else {
					if (yych <= 'e') goto yy460;
					if (yych <= 'f') goto yy459;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy365:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych <= '$') goto yy1;
			goto yy461;
		} else {
			if (yych <= '.') goto yy1;
			if (yych <= '/') goto yy189;
			goto yy192;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') goto yy462;
			if (yych <= '@') goto yy1;
			goto yy192;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy192;
			goto yy1;
		}
	}
yy366:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy113;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy459;
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy113;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy459;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy113;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy459;
				if (yych <= 'z') goto yy115;
				goto yy113;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy113;
				if (yych <= 0xF4) goto yy4;
				goto yy113;
			}
		}
	}
yy367:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy251;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy459;
				if (yych <= ':') goto yy148;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy459;
				if (yych <= 'Z') goto yy115;
				goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy459;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy368:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'f') {
		if (yych <= '9') {
			if (yych == '-') {
				yyt1 = YYCURSOR;
				goto yy463;
			}
			if (yych <= '/') {
				yyt1 = YYCURSOR;
				goto yy434;
			}
			yyt1 = YYCURSOR;
			goto yy464;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy465;
				yyt1 = YYCURSOR;
				goto yy434;
			} else {
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy464;
				}
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy434;
				}
				yyt1 = YYCURSOR;
				goto yy464;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy434;
			}
			if (yych <= 0xC1) goto yy354;
			if (yych <= 0xDF) {
				yyt1 = YYCURSOR;
				goto yy436;
			}
			yyt1 = YYCURSOR;
			goto yy437;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt1 = YYCURSOR;
					goto yy438;
				}
				yyt1 = YYCURSOR;
				goto yy439;
			} else {
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy440;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy441;
				}
				goto yy354;
			}
		}
	}
yy369:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy354;
			goto yy466;
		} else {
			if (yych <= '/') goto yy354;
			if (yych <= '9') goto yy372;
			goto yy467;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy354;
			goto yy372;
		} else {
			if (yych <= '`') goto yy354;
			if (yych <= 'f') goto yy372;
			goto yy354;
		}
	}
yy370:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[256+yych] & 1) {
		goto yy273;
	}
	if (yych <= 0xC1) {
		if (yych <= ')') {
			if (yych <= '"') goto yy371;
			if (yych <= '&') goto yy370;
		} else {
			if (yych <= 'Z') goto yy370;
			if (yych <= ']') goto yy371;
			if (yych <= 'z') goto yy370;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy275;
			if (yych <= 0xE0) goto yy276;
			goto yy277;
		} else {
			if (yych <= 0xF0) goto yy278;
			if (yych <= 0xF3) goto yy279;
			if (yych <= 0xF4) goto yy280;
		}
	}
yy371:
#line 170 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_URL); }
#line 12517 "../../lnav/src/data_scanner_re.cc"
yy372:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy468;
		if (yych <= ':') goto yy282;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy468;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy468;
		goto yy1;
	}
yy373:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy469;
		if (yych <= ':') goto yy375;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy374;
		if (yych <= '`') goto yy1;
		if (yych >= 'g') goto yy1;
	}
yy374:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy470;
		if (yych <= ':') goto yy471;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy470;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy470;
		goto yy1;
	}
yy375:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy472;
			} else {
				if (yych <= '1') goto yy473;
				if (yych <= '2') goto yy474;
				goto yy472;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy475;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy475;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy376:
	yyaccept = 17;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy199;
				if (yych <= 0x08) goto yy4;
				goto yy199;
			} else {
				if (yych == '\r') goto yy199;
				if (yych <= 0x1A) goto yy4;
				goto yy199;
			}
		} else {
			if (yych <= '$') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy199;
				goto yy4;
			} else {
				if (yych <= ',') goto yy199;
				if (yych <= '/') goto yy4;
				if (yych <= '0') goto yy476;
				goto yy477;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= ':') {
				if (yych <= '2') goto yy478;
				if (yych <= '9') goto yy476;
				goto yy4;
			} else {
				if (yych <= '?') goto yy199;
				if (yych <= 'Z') goto yy4;
				if (yych <= '^') goto yy199;
				goto yy4;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy199;
				if (yych <= 'z') goto yy4;
				goto yy199;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy199;
				if (yych <= 0xF4) goto yy4;
				goto yy199;
			}
		}
	}
yy377:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '-') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy479;
					if (yych >= ';') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy479;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy479;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy378:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy475;
		if (yych <= ':') goto yy152;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy475;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy475;
		goto yy1;
	}
yy379:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '/') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '9') goto yy479;
				if (yych <= ':') goto yy378;
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '_') {
				if (yych <= 'F') goto yy479;
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy479;
				if (yych <= 'z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy380:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '-') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy480;
					if (yych <= ':') goto yy378;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy479;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy479;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy381:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '.') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy376;
			} else {
				if (yych <= '5') {
					if (yych <= '/') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					goto yy480;
				} else {
					if (yych <= '9') goto yy479;
					if (yych <= ':') goto yy378;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy479;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy479;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy382:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy354;
				if (yych <= 0x08) goto yy4;
				goto yy354;
			} else {
				if (yych == '\r') goto yy354;
				if (yych <= 0x1A) goto yy4;
				goto yy354;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy354;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy354;
				goto yy181;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy340;
				goto yy452;
			} else {
				if (yych <= '?') goto yy354;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy46;
				goto yy354;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy354;
				if (yych <= 'z') goto yy46;
				goto yy354;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy354;
				if (yych <= 0xF4) goto yy4;
				goto yy354;
			}
		}
	}
yy383:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy481;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy482;
	goto yy1;
yy384:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy481;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy383;
	goto yy1;
yy385:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych == '.') goto yy481;
		goto yy1;
	} else {
		if (yych <= '4') goto yy383;
		if (yych <= '5') goto yy483;
		if (yych <= '9') goto yy482;
		goto yy1;
	}
yy386:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy484;
			if (yych <= ':') goto yy299;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy484;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy484;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy387:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy485;
			if (yych <= ':') goto yy486;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy485;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy485;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy388:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy484;
				goto yy299;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy484;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy484;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy389:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy484;
			} else {
				if (yych <= ':') goto yy299;
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy484;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy487;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'e') goto yy484;
				goto yy487;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy390:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy488;
				goto yy1;
			} else {
				if (yych <= '\r') goto yy488;
				if (yych <= 0x1F) goto yy1;
				goto yy488;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy306;
				goto yy1;
			} else {
				if (yych <= '\'') goto yy307;
				if (yych <= ',') goto yy1;
				goto yy390;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy1;
				if (yych <= ':') goto yy390;
				goto yy1;
			} else {
				if (yych <= '=') goto yy408;
				if (yych <= '>') goto yy309;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy390;
				if (yych <= '^') goto yy1;
				goto yy390;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy390;
				goto yy1;
			}
		}
	}
yy391:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 8) {
		goto yy391;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1;
			goto yy489;
		} else {
			if (yych <= '\\') goto yy392;
			if (yych <= 0xC1) goto yy1;
			goto yy393;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy394;
			if (yych <= 0xEF) goto yy395;
			goto yy396;
		} else {
			if (yych <= 0xF3) goto yy397;
			if (yych <= 0xF4) goto yy398;
			goto yy1;
		}
	}
yy392:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy391;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy391;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy394;
			if (yych <= 0xEF) goto yy395;
			goto yy396;
		} else {
			if (yych <= 0xF3) goto yy397;
			if (yych <= 0xF4) goto yy398;
			goto yy1;
		}
	}
yy393:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy391;
	goto yy1;
yy394:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy393;
	goto yy1;
yy395:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy393;
	goto yy1;
yy396:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy395;
	goto yy1;
yy397:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy395;
	goto yy1;
yy398:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy395;
	goto yy1;
yy399:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 16) {
		goto yy399;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1;
			goto yy489;
		} else {
			if (yych <= '\\') goto yy400;
			if (yych <= 0xC1) goto yy1;
			goto yy401;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy402;
			if (yych <= 0xEF) goto yy403;
			goto yy404;
		} else {
			if (yych <= 0xF3) goto yy405;
			if (yych <= 0xF4) goto yy406;
			goto yy1;
		}
	}
yy400:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy399;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy399;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy402;
			if (yych <= 0xEF) goto yy403;
			goto yy404;
		} else {
			if (yych <= 0xF3) goto yy405;
			if (yych <= 0xF4) goto yy406;
			goto yy1;
		}
	}
yy401:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy399;
	goto yy1;
yy402:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy401;
	goto yy1;
yy403:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy401;
	goto yy1;
yy404:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy403;
	goto yy1;
yy405:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy403;
	goto yy1;
yy406:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy403;
	goto yy1;
yy407:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy407;
				goto yy1;
			} else {
				if (yych <= '\r') goto yy407;
				if (yych <= 0x1F) goto yy1;
				goto yy407;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy306;
				goto yy1;
			} else {
				if (yych <= '\'') goto yy307;
				if (yych <= ',') goto yy1;
				goto yy390;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy1;
				if (yych <= ':') goto yy390;
				goto yy1;
			} else {
				if (yych <= '=') goto yy408;
				if (yych <= '>') goto yy309;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy390;
				if (yych <= '^') goto yy1;
				goto yy390;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy390;
				goto yy1;
			}
		}
	}
yy408:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 1) {
		goto yy490;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1;
			goto yy491;
		} else {
			if (yych <= '\'') goto yy492;
			if (yych <= 0xC1) goto yy1;
			goto yy493;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy494;
			if (yych <= 0xEF) goto yy495;
			goto yy496;
		} else {
			if (yych <= 0xF3) goto yy497;
			if (yych <= 0xF4) goto yy498;
			goto yy1;
		}
	}
yy409:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy409;
			goto yy1;
		} else {
			if (yych <= '\r') goto yy409;
			if (yych <= 0x1F) goto yy1;
			goto yy409;
		}
	} else {
		if (yych <= '<') {
			if (yych == '/') goto yy215;
			goto yy1;
		} else {
			if (yych <= '=') goto yy313;
			if (yych <= '>') goto yy216;
			if (yych <= '?') goto yy215;
			goto yy1;
		}
	}
yy410:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 32) {
		goto yy410;
	}
	if (yych <= 'z') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '-') goto yy499;
			goto yy500;
		} else {
			if (yych <= ':') goto yy499;
			if (yych <= '>') goto yy216;
			if (yych <= '?') goto yy500;
			goto yy499;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy414;
			if (yych <= 0xE0) goto yy415;
			goto yy416;
		} else {
			if (yych <= 0xF0) goto yy417;
			if (yych <= 0xF3) goto yy418;
			if (yych <= 0xF4) goto yy419;
			goto yy1;
		}
	}
yy411:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 16) {
		goto yy500;
	}
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy411;
				goto yy410;
			} else {
				if (yych <= '\r') goto yy411;
				if (yych == ' ') goto yy411;
				goto yy410;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy412;
				if (yych == '\'') goto yy413;
				goto yy410;
			} else {
				if (yych == '.') goto yy410;
				if (yych <= ':') goto yy499;
				goto yy410;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '>') goto yy216;
				if (yych <= '@') goto yy410;
				if (yych <= 'Z') goto yy499;
				goto yy410;
			} else {
				if (yych == '`') goto yy410;
				if (yych <= 'z') goto yy499;
				goto yy410;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy414;
				if (yych <= 0xE0) goto yy415;
				goto yy416;
			} else {
				if (yych <= 0xF0) goto yy417;
				if (yych <= 0xF3) goto yy418;
				if (yych <= 0xF4) goto yy419;
				goto yy1;
			}
		}
	}
yy412:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 64) {
		goto yy412;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '"') goto yy410;
			if (yych <= '-') goto yy501;
			goto yy502;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy501;
				goto yy503;
			} else {
				if (yych <= '?') goto yy502;
				if (yych <= 'Z') goto yy501;
				goto yy504;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy501;
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy505;
			goto yy506;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy507;
				goto yy508;
			} else {
				if (yych <= 0xF3) goto yy509;
				if (yych <= 0xF4) goto yy510;
				goto yy1;
			}
		}
	}
yy413:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 128) {
		goto yy413;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '\'') goto yy410;
			if (yych <= '-') goto yy511;
			goto yy512;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy511;
				goto yy513;
			} else {
				if (yych <= '?') goto yy512;
				if (yych <= 'Z') goto yy511;
				goto yy514;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy511;
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy515;
			goto yy516;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy517;
				goto yy518;
			} else {
				if (yych <= 0xF3) goto yy519;
				if (yych <= 0xF4) goto yy520;
				goto yy1;
			}
		}
	}
yy414:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy410;
	goto yy1;
yy415:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy414;
	goto yy1;
yy416:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy414;
	goto yy1;
yy417:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy416;
	goto yy1;
yy418:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy416;
	goto yy1;
yy419:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy416;
	goto yy1;
yy420:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy521;
			} else {
				if (yych == '\r') goto yy521;
				goto yy1;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy521;
				if (yych <= ',') goto yy1;
				goto yy420;
			} else {
				if (yych <= '.') goto yy1;
				if (yych <= '/') goto yy215;
				goto yy420;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1;
				goto yy422;
			} else {
				if (yych == '?') goto yy215;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy420;
				if (yych <= '^') goto yy1;
				goto yy420;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy420;
				goto yy1;
			}
		}
	}
yy421:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy421;
			} else {
				if (yych == '\r') goto yy421;
				goto yy1;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy421;
				if (yych <= ',') goto yy1;
				goto yy420;
			} else {
				if (yych <= '.') goto yy1;
				if (yych <= '/') goto yy215;
				goto yy420;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1;
			} else {
				if (yych == '?') goto yy215;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy420;
				if (yych <= '^') goto yy1;
				goto yy420;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy420;
				goto yy1;
			}
		}
	}
yy422:
	yych = *++YYCURSOR;
	if (yych <= '\'') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1;
				goto yy522;
			} else {
				if (yych <= '\t') goto yy523;
				if (yych <= '\f') goto yy522;
				goto yy523;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy523;
				goto yy522;
			} else {
				if (yych <= '"') goto yy524;
				if (yych <= '&') goto yy522;
				goto yy525;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '>') {
				if (yych <= '=') goto yy522;
				goto yy1;
			} else {
				if (yych <= 0x7F) goto yy522;
				if (yych <= 0xC1) goto yy1;
				goto yy526;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy527;
				if (yych <= 0xEF) goto yy528;
				goto yy529;
			} else {
				if (yych <= 0xF3) goto yy530;
				if (yych <= 0xF4) goto yy531;
				goto yy1;
			}
		}
	}
yy423:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy532;
				goto yy148;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy532;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy532;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy424:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy533;
	goto yy1;
yy425:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych <= '$') goto yy222;
			goto yy461;
		} else {
			if (yych <= '.') goto yy222;
			if (yych <= '/') goto yy189;
			goto yy192;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') goto yy462;
			if (yych <= '@') goto yy222;
			goto yy192;
		} else {
			if (yych <= '`') goto yy222;
			if (yych <= 'f') goto yy192;
			goto yy222;
		}
	}
yy426:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy532;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy532;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy4;
				} else {
					if (yych <= 'f') goto yy534;
					if (yych <= 'z') goto yy326;
					goto yy4;
				}
			}
		}
	}
yy427:
	++YYCURSOR;
yy428:
	YYCURSOR = yyt1;
#line 262 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_CONSTANT); }
#line 14275 "../../lnav/src/data_scanner_re.cc"
yy429:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy427;
		goto yy428;
	} else {
		if (yych <= '\r') goto yy427;
		if (yych == ' ') goto yy427;
		goto yy428;
	}
yy430:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[1024+yych] & 16) {
		goto yy46;
	}
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy427;
			goto yy4;
		} else {
			if (yych <= '\r') goto yy427;
			if (yych <= 0x1F) goto yy4;
			goto yy427;
		}
	} else {
		if (yych <= '*') {
			if (yych == '%') goto yy80;
			goto yy4;
		} else {
			if (yych <= '+') goto yy80;
			if (yych == '@') goto yy96;
			goto yy4;
		}
	}
yy431:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy189;
	if (yych == ':') goto yy152;
	goto yy428;
yy432:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 128) {
		goto yy137;
	}
	if (yych <= '*') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy225;
			if (yych <= 0x08) goto yy428;
			if (yych <= '\n') goto yy225;
			goto yy428;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy225;
				goto yy428;
			} else {
				if (yych <= '"') goto yy225;
				if (yych <= '&') goto yy428;
				goto yy225;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych == ',') goto yy225;
			if (yych <= '-') goto yy428;
			yyt4 = YYCURSOR;
			goto yy226;
		} else {
			if (yych <= ';') {
				if (yych <= '9') goto yy428;
				goto yy225;
			} else {
				if (yych == '?') goto yy225;
				goto yy428;
			}
		}
	}
yy433:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= '!') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy3;
				} else {
					if (yych <= '\n') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					if (yych <= '\f') goto yy3;
					yyt4 = YYCURSOR;
					goto yy221;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy3;
					goto yy5;
				} else {
					if (yych <= 0x1F) goto yy3;
					if (yych <= ' ') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					yyt4 = YYCURSOR;
					goto yy223;
				}
			}
		} else {
			if (yych <= '&') {
				if (yych <= '#') {
					if (yych <= '"') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy5;
				} else {
					if (yych <= '$') goto yy3;
					if (yych <= '%') goto yy80;
					goto yy5;
				}
			} else {
				if (yych <= '+') {
					if (yych <= '\'') {
						yyt4 = YYCURSOR;
						goto yy224;
					}
					if (yych <= '*') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy80;
				} else {
					if (yych <= ',') {
						yyt4 = YYCURSOR;
						goto yy223;
					}
					if (yych <= '-') goto yy46;
					yyt4 = YYCURSOR;
					goto yy230;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= '>') {
				if (yych <= '9') {
					if (yych <= '/') goto yy3;
					goto yy46;
				} else {
					if (yych <= ':') {
						yyt3 = YYCURSOR;
						goto yy535;
					}
					if (yych <= ';') {
						yyt4 = YYCURSOR;
						goto yy221;
					}
					goto yy5;
				}
			} else {
				if (yych <= 'Z') {
					if (yych <= '?') {
						yyt4 = YYCURSOR;
						goto yy223;
					}
					if (yych <= '@') goto yy96;
					goto yy46;
				} else {
					if (yych == '_') goto yy46;
					goto yy5;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= '~') {
					if (yych <= 'z') goto yy433;
					goto yy5;
				} else {
					if (yych <= 0x7F) goto yy3;
					if (yych <= 0xC1) goto yy5;
					goto yy56;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) goto yy57;
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy5;
				}
			}
		}
	}
yy434:
	++YYCURSOR;
yy435:
	YYCURSOR = yyt1;
#line 172 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_TIME); }
#line 14484 "../../lnav/src/data_scanner_re.cc"
yy436:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy434;
	goto yy1;
yy437:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy436;
	goto yy1;
yy438:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy436;
	goto yy1;
yy439:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy438;
	goto yy1;
yy440:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy438;
	goto yy1;
yy441:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy438;
	goto yy1;
yy442:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy536;
	goto yy1;
yy443:
	yych = *++YYCURSOR;
	if (yych == 'r') goto yy537;
	goto yy100;
yy444:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy24;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy24;
				if (yych == 0x1B) goto yy24;
				goto yy4;
			}
		} else {
			if (yych <= '#') {
				if (yych == '!') goto yy99;
				if (yych <= '"') goto yy24;
				goto yy99;
			} else {
				if (yych <= '&') {
					if (yych <= '$') goto yy101;
					goto yy99;
				} else {
					if (yych <= '\'') goto yy24;
					if (yych <= '*') goto yy99;
					goto yy24;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') {
				if (yych <= '9') goto yy101;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy24;
				goto yy101;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy24;
					goto yy99;
				} else {
					if (yych <= ']') goto yy24;
					if (yych <= '^') goto yy99;
					goto yy101;
				}
			}
		} else {
			if (yych <= '}') {
				if (yych <= '`') goto yy24;
				if (yych <= 'a') goto yy538;
				if (yych <= 'z') goto yy101;
				goto yy24;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy99;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy24;
					if (yych <= 0xF4) goto yy4;
					goto yy24;
				}
			}
		}
	}
yy445:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '0') goto yy539;
				goto yy540;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') {
				if (yych <= '2') goto yy541;
				goto yy539;
			} else {
				if (yych <= '?') goto yy4;
				if (yych <= '@') goto yy96;
				goto yy46;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy446:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy345;
				if (yych <= 0x08) goto yy4;
				goto yy345;
			} else {
				if (yych == '\r') goto yy345;
				if (yych <= 0x1A) goto yy4;
				goto yy345;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy345;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy345;
				goto yy255;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '.') goto yy445;
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			} else {
				if (yych <= '?') goto yy345;
				if (yych <= '@') goto yy96;
				if (yych >= '[') goto yy345;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy345;
				if (yych >= '{') goto yy345;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xF4) goto yy4;
				goto yy345;
			}
		}
	}
yy447:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy345;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy345;
				goto yy3;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy345;
					goto yy3;
				} else {
					if (yych <= 0x1B) goto yy345;
					if (yych <= 0x1F) goto yy3;
					goto yy345;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				if (yych <= '*') goto yy345;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy345;
					goto yy255;
				} else {
					if (yych <= '.') goto yy350;
					if (yych <= '/') goto yy3;
					goto yy447;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy62;
				if (yych <= '?') goto yy345;
				if (yych <= '@') goto yy96;
				goto yy447;
			} else {
				if (yych <= '_') {
					if (yych <= '^') goto yy345;
					goto yy447;
				} else {
					if (yych <= '`') goto yy345;
					if (yych <= 'z') goto yy447;
					goto yy345;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy345;
				}
			}
		}
	}
yy448:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy345;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy345;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy345;
				if (yych == 0x1B) goto yy345;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy345;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy345;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy345;
				if (yych <= '-') goto yy255;
				goto yy445;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '/') goto yy4;
				if (yych <= '5') goto yy446;
				if (yych <= '9') goto yy447;
				goto yy4;
			} else {
				if (yych <= '?') goto yy345;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy447;
				goto yy345;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy345;
				if (yych <= 'z') goto yy447;
				goto yy345;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xF4) goto yy4;
				goto yy345;
			}
		}
	}
yy449:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych >= '+') goto yy80;
		} else {
			if (yych <= ',') goto yy450;
			if (yych <= '.') goto yy80;
			if (yych >= '0') goto yy449;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy450;
			if (yych <= '@') goto yy82;
			if (yych <= 'Z') goto yy80;
		} else {
			if (yych == '`') goto yy450;
			if (yych <= 'z') goto yy80;
		}
	}
yy450:
#line 249 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_NUMBER); }
#line 14829 "../../lnav/src/data_scanner_re.cc"
yy451:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy450;
				if (yych <= 0x08) goto yy3;
				if (yych <= '\n') goto yy450;
				goto yy3;
			} else {
				if (yych <= '\r') goto yy450;
				if (yych == 0x1B) goto yy450;
				goto yy3;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy450;
				if (yych <= '$') goto yy3;
				if (yych <= '%') goto yy80;
				goto yy450;
			} else {
				if (yych <= ',') {
					if (yych <= '+') goto yy80;
					goto yy450;
				} else {
					if (yych <= '.') goto yy46;
					if (yych <= '/') goto yy3;
					goto yy451;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy62;
				if (yych <= '?') goto yy450;
				if (yych <= '@') goto yy96;
				goto yy344;
			} else {
				if (yych <= '_') {
					if (yych <= '^') goto yy450;
					goto yy344;
				} else {
					if (yych <= '`') goto yy450;
					if (yych <= 'z') goto yy344;
					goto yy450;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy3;
				if (yych <= 0xC1) goto yy450;
				if (yych <= 0xDF) goto yy56;
				goto yy57;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy450;
				}
			}
		}
	}
yy452:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy542;
		if (yych <= ':') goto yy152;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy542;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy542;
		goto yy1;
	}
yy453:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy543;
	goto yy4;
yy454:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy544;
	goto yy1;
yy455:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '-') goto yy545;
			if (yych <= '.') goto yy46;
			goto yy4;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy546;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy46;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy456:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') goto yy4;
	if (yych <= '/') goto yy547;
	if (yych <= '9') goto yy548;
	goto yy4;
yy457:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy549;
					goto yy550;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy551;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy552;
					goto yy551;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy551;
				} else {
					if (yych <= 'e') goto yy552;
					if (yych <= 'f') goto yy551;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy458:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy550;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy551;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy552;
					goto yy551;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy551;
				} else {
					if (yych <= 'e') goto yy552;
					if (yych <= 'f') goto yy551;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy459:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy113;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy551;
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy113;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy551;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy113;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy551;
				if (yych <= 'z') goto yy115;
				goto yy113;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy113;
				if (yych <= 0xF4) goto yy4;
				goto yy113;
			}
		}
	}
yy460:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy251;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy551;
				if (yych <= ':') goto yy148;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy551;
				if (yych <= 'Z') goto yy115;
				goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy551;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy461:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy553;
		goto yy1;
	} else {
		if (yych <= 'Z') goto yy553;
		if (yych <= '`') goto yy1;
		if (yych <= 'z') goto yy553;
		goto yy1;
	}
yy462:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '0') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				goto yy554;
			} else {
				if (yych <= '2') {
					if (yych <= '1') goto yy555;
					goto yy556;
				} else {
					if (yych <= '9') goto yy554;
					if (yych <= ':') goto yy1;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy557;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy557;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy1;
				}
			}
		}
	}
yy463:
	yyaccept = 25;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy435;
		if (yych <= '9') goto yy542;
		goto yy435;
	} else {
		if (yych <= 'F') goto yy542;
		if (yych <= '`') goto yy435;
		if (yych <= 'f') goto yy542;
		goto yy435;
	}
yy464:
	yyaccept = 25;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy435;
		if (yych <= '9') goto yy468;
		if (yych <= ':') goto yy282;
		goto yy435;
	} else {
		if (yych <= 'F') goto yy468;
		if (yych <= '`') goto yy435;
		if (yych <= 'f') goto yy468;
		goto yy435;
	}
yy465:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy558;
		if (yych <= ':') goto yy375;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy559;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy559;
		goto yy1;
	}
yy466:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy542;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy542;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy542;
		goto yy1;
	}
yy467:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy559;
		if (yych <= ':') goto yy375;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy559;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy559;
		goto yy1;
	}
yy468:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy282;
	goto yy1;
yy469:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy560;
		if (yych <= ':') goto yy471;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy470;
		if (yych <= '`') goto yy1;
		if (yych >= 'g') goto yy1;
	}
yy470:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy561;
		if (yych >= ';') goto yy1;
	} else {
		if (yych <= 'F') goto yy561;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy561;
		goto yy1;
	}
yy471:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy562;
		if (yych <= ':') goto yy563;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy562;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy562;
		goto yy1;
	}
yy472:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy564;
				goto yy565;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy566;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy566;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy473:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy567;
				goto yy565;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy566;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy566;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy474:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '4') goto yy567;
				goto yy568;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy564;
				if (yych <= ':') goto yy565;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy566;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy566;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy475:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy566;
			if (yych <= ':') goto yy565;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy566;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy566;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy476:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy569;
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy570;
	goto yy4;
yy477:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy569;
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy476;
	goto yy4;
yy478:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy569;
		goto yy4;
	} else {
		if (yych <= '4') goto yy476;
		if (yych <= '5') goto yy571;
		if (yych <= '9') goto yy570;
		goto yy4;
	}
yy479:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '/') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '9') goto yy572;
				if (yych <= ':') goto yy378;
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '_') {
				if (yych <= 'F') goto yy572;
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy572;
				if (yych <= 'z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy480:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '-') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy572;
					if (yych <= ':') goto yy378;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy572;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy572;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy481:
	yych = *++YYCURSOR;
	if (yych <= '1') {
		if (yych <= '/') goto yy1;
		if (yych <= '0') goto yy573;
		goto yy574;
	} else {
		if (yych <= '2') goto yy575;
		if (yych <= '9') goto yy573;
		goto yy1;
	}
yy482:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy481;
	goto yy1;
yy483:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy481;
	if (yych <= '/') goto yy1;
	if (yych <= '5') goto yy482;
	goto yy1;
yy484:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1;
			if (yych <= ':') goto yy299;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy485:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy576;
			if (yych >= ';') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy576;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy576;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy486:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy475;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy475;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy475;
		goto yy1;
	}
yy487:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1;
			if (yych <= ':') goto yy577;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy488:
	yych = *++YYCURSOR;
	if (yych <= '!') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy488;
			goto yy1;
		} else {
			if (yych <= '\r') goto yy488;
			if (yych == ' ') goto yy488;
			goto yy1;
		}
	} else {
		if (yych <= '\'') {
			if (yych <= '"') goto yy306;
			if (yych <= '&') goto yy1;
			goto yy307;
		} else {
			if (yych <= '<') goto yy1;
			if (yych <= '=') goto yy408;
			if (yych <= '>') goto yy309;
			goto yy1;
		}
	}
yy489:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy578;
			} else {
				if (yych == '\r') goto yy578;
				goto yy1;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy578;
				if (yych <= '!') goto yy1;
				goto yy306;
			} else {
				if (yych == '\'') goto yy307;
				goto yy1;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= ':') {
				if (yych <= '-') goto yy390;
				if (yych <= '/') goto yy1;
				goto yy390;
			} else {
				if (yych == '>') goto yy309;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy390;
				if (yych <= '^') goto yy1;
				goto yy390;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy390;
				goto yy1;
			}
		}
	}
yy490:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 1) {
		goto yy490;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1;
			if (yych >= '#') goto yy492;
		} else {
			if (yych <= '>') goto yy309;
			if (yych <= 0xC1) goto yy1;
			goto yy493;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy494;
			if (yych <= 0xEF) goto yy495;
			goto yy496;
		} else {
			if (yych <= 0xF3) goto yy497;
			if (yych <= 0xF4) goto yy498;
			goto yy1;
		}
	}
yy491:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 2) {
		goto yy491;
	}
	if (yych <= 0xDF) {
		if (yych <= '>') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '\'') goto yy579;
			goto yy580;
		} else {
			if (yych <= '\\') goto yy581;
			if (yych <= 0xC1) goto yy1;
			goto yy582;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy583;
			if (yych <= 0xEF) goto yy584;
			goto yy585;
		} else {
			if (yych <= 0xF3) goto yy586;
			if (yych <= 0xF4) goto yy587;
			goto yy1;
		}
	}
yy492:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 4) {
		goto yy492;
	}
	if (yych <= 0xDF) {
		if (yych <= '>') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '"') goto yy579;
			goto yy588;
		} else {
			if (yych <= '\\') goto yy589;
			if (yych <= 0xC1) goto yy1;
			goto yy590;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy591;
			if (yych <= 0xEF) goto yy592;
			goto yy593;
		} else {
			if (yych <= 0xF3) goto yy594;
			if (yych <= 0xF4) goto yy595;
			goto yy1;
		}
	}
yy493:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy490;
	goto yy1;
yy494:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy493;
	goto yy1;
yy495:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy493;
	goto yy1;
yy496:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy495;
	goto yy1;
yy497:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy495;
	goto yy1;
yy498:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy495;
	goto yy1;
yy499:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 8) {
		goto yy499;
	}
	if (yych <= 0x7F) {
		if (yych <= '<') {
			if (yych <= 0x00) goto yy1;
			if (yych != '/') goto yy410;
		} else {
			if (yych <= '=') goto yy411;
			if (yych <= '>') goto yy216;
			if (yych >= '@') goto yy410;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy414;
			if (yych <= 0xE0) goto yy415;
			goto yy416;
		} else {
			if (yych <= 0xF0) goto yy417;
			if (yych <= 0xF3) goto yy418;
			if (yych <= 0xF4) goto yy419;
			goto yy1;
		}
	}
yy500:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 32) {
		goto yy410;
	}
	if (yych <= 'z') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '-') goto yy499;
			goto yy500;
		} else {
			if (yych <= ':') goto yy499;
			if (yych <= '>') goto yy314;
			if (yych <= '?') goto yy500;
			goto yy499;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy414;
			if (yych <= 0xE0) goto yy415;
			goto yy416;
		} else {
			if (yych <= 0xF0) goto yy417;
			if (yych <= 0xF3) goto yy418;
			if (yych <= 0xF4) goto yy419;
			goto yy1;
		}
	}
yy501:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy501;
				goto yy412;
			} else {
				if (yych <= '\r') goto yy501;
				if (yych == ' ') goto yy501;
				goto yy412;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy410;
				if (yych == '-') goto yy501;
				goto yy412;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy501;
				} else {
					if (yych <= '<') goto yy412;
					if (yych <= '=') goto yy596;
					goto yy503;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy502;
				if (yych <= '@') goto yy412;
				if (yych <= 'Z') goto yy501;
				goto yy412;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy504;
					goto yy412;
				} else {
					if (yych == '`') goto yy412;
					goto yy501;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy412;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy505;
				goto yy506;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy507;
					goto yy508;
				} else {
					if (yych <= 0xF3) goto yy509;
					if (yych <= 0xF4) goto yy510;
					goto yy1;
				}
			}
		}
	}
yy502:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 64) {
		goto yy412;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '"') goto yy410;
			if (yych <= '-') goto yy501;
			goto yy502;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy501;
				goto yy597;
			} else {
				if (yych <= '?') goto yy502;
				if (yych <= 'Z') goto yy501;
				goto yy504;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy501;
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy505;
			goto yy506;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy507;
				goto yy508;
			} else {
				if (yych <= 0xF3) goto yy509;
				if (yych <= 0xF4) goto yy510;
				goto yy1;
			}
		}
	}
yy503:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy217;
			if (yych <= '!') goto yy598;
			goto yy599;
		} else {
			if (yych == '\\') goto yy600;
			if (yych <= 0x7F) goto yy598;
			goto yy217;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy601;
			if (yych <= 0xE0) goto yy602;
			goto yy603;
		} else {
			if (yych <= 0xF0) goto yy604;
			if (yych <= 0xF3) goto yy605;
			if (yych <= 0xF4) goto yy606;
			goto yy217;
		}
	}
yy504:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '.') {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy598;
				if (yych <= '\t') goto yy412;
				goto yy410;
			} else {
				if (yych == '-') goto yy501;
				goto yy412;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '/') goto yy502;
				if (yych <= ':') goto yy501;
				goto yy412;
			} else {
				if (yych <= '>') goto yy503;
				if (yych <= '?') goto yy502;
				if (yych <= '@') goto yy412;
				goto yy501;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') goto yy501;
				goto yy412;
			} else {
				if (yych <= 'z') goto yy501;
				if (yych <= 0x7F) goto yy412;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy505;
				if (yych <= 0xE0) goto yy506;
				goto yy507;
			} else {
				if (yych <= 0xF0) goto yy508;
				if (yych <= 0xF3) goto yy509;
				if (yych <= 0xF4) goto yy510;
				goto yy1;
			}
		}
	}
yy505:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy412;
	goto yy1;
yy506:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy505;
	goto yy1;
yy507:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy505;
	goto yy1;
yy508:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy507;
	goto yy1;
yy509:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy507;
	goto yy1;
yy510:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy507;
	goto yy1;
yy511:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy511;
				goto yy413;
			} else {
				if (yych <= '\r') goto yy511;
				if (yych == ' ') goto yy511;
				goto yy413;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy410;
				if (yych == '-') goto yy511;
				goto yy413;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy511;
				} else {
					if (yych <= '<') goto yy413;
					if (yych <= '=') goto yy607;
					goto yy513;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy512;
				if (yych <= '@') goto yy413;
				if (yych <= 'Z') goto yy511;
				goto yy413;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy514;
					goto yy413;
				} else {
					if (yych == '`') goto yy413;
					goto yy511;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy413;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy515;
				goto yy516;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy517;
					goto yy518;
				} else {
					if (yych <= 0xF3) goto yy519;
					if (yych <= 0xF4) goto yy520;
					goto yy1;
				}
			}
		}
	}
yy512:
	yych = *++YYCURSOR;
	if (yybm[256+yych] & 128) {
		goto yy413;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '\'') goto yy410;
			if (yych <= '-') goto yy511;
			goto yy512;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy511;
				goto yy608;
			} else {
				if (yych <= '?') goto yy512;
				if (yych <= 'Z') goto yy511;
				goto yy514;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy511;
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy515;
			goto yy516;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy517;
				goto yy518;
			} else {
				if (yych <= 0xF3) goto yy519;
				if (yych <= 0xF4) goto yy520;
				goto yy1;
			}
		}
	}
yy513:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy217;
			if (yych <= '&') goto yy609;
			goto yy599;
		} else {
			if (yych == '\\') goto yy610;
			if (yych <= 0x7F) goto yy609;
			goto yy217;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy611;
			if (yych <= 0xE0) goto yy612;
			goto yy613;
		} else {
			if (yych <= 0xF0) goto yy614;
			if (yych <= 0xF3) goto yy615;
			if (yych <= 0xF4) goto yy616;
			goto yy217;
		}
	}
yy514:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '.') {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy609;
				if (yych <= '\t') goto yy413;
				goto yy410;
			} else {
				if (yych == '-') goto yy511;
				goto yy413;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '/') goto yy512;
				if (yych <= ':') goto yy511;
				goto yy413;
			} else {
				if (yych <= '>') goto yy513;
				if (yych <= '?') goto yy512;
				if (yych <= '@') goto yy413;
				goto yy511;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') goto yy511;
				goto yy413;
			} else {
				if (yych <= 'z') goto yy511;
				if (yych <= 0x7F) goto yy413;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy515;
				if (yych <= 0xE0) goto yy516;
				goto yy517;
			} else {
				if (yych <= 0xF0) goto yy518;
				if (yych <= 0xF3) goto yy519;
				if (yych <= 0xF4) goto yy520;
				goto yy1;
			}
		}
	}
yy515:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy413;
	goto yy1;
yy516:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy515;
	goto yy1;
yy517:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy515;
	goto yy1;
yy518:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy517;
	goto yy1;
yy519:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy517;
	goto yy1;
yy520:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy517;
	goto yy1;
yy521:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy521;
			goto yy1;
		} else {
			if (yych <= '\r') goto yy521;
			if (yych <= 0x1F) goto yy1;
			goto yy521;
		}
	} else {
		if (yych <= '<') {
			if (yych == '/') goto yy215;
			goto yy1;
		} else {
			if (yych <= '=') goto yy422;
			if (yych == '?') goto yy215;
			goto yy1;
		}
	}
yy522:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1;
				if (yych <= ',') goto yy522;
				goto yy617;
			} else {
				if (yych <= '.') goto yy522;
				if (yych <= '/') goto yy618;
				goto yy617;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy522;
				if (yych <= '>') goto yy1;
				goto yy618;
			} else {
				if (yych <= '@') goto yy522;
				if (yych <= 'Z') goto yy617;
				goto yy522;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy522;
				goto yy617;
			} else {
				if (yych <= 0x7F) goto yy522;
				if (yych <= 0xC1) goto yy1;
				goto yy526;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy527;
				if (yych <= 0xEF) goto yy528;
				goto yy529;
			} else {
				if (yych <= 0xF3) goto yy530;
				if (yych <= 0xF4) goto yy531;
				goto yy1;
			}
		}
	}
yy523:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy523;
				goto yy522;
			} else {
				if (yych <= '\r') goto yy523;
				if (yych == ' ') goto yy523;
				goto yy522;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy524;
				if (yych == '\'') goto yy525;
				goto yy522;
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy617;
					goto yy522;
				} else {
					if (yych <= '/') goto yy618;
					if (yych <= ':') goto yy617;
					goto yy522;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy1;
				if (yych <= '?') goto yy618;
				if (yych <= '@') goto yy522;
				goto yy617;
			} else {
				if (yych == '_') goto yy617;
				if (yych <= '`') goto yy522;
				goto yy617;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy522;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy526;
				goto yy527;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy528;
					goto yy529;
				} else {
					if (yych <= 0xF3) goto yy530;
					if (yych <= 0xF4) goto yy531;
					goto yy1;
				}
			}
		}
	}
yy524:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy1;
				if (yych <= '!') goto yy524;
				goto yy522;
			} else {
				if (yych == '-') goto yy619;
				if (yych <= '.') goto yy524;
				goto yy620;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy619;
				if (yych <= '=') goto yy524;
				goto yy621;
			} else {
				if (yych <= '?') goto yy620;
				if (yych <= '@') goto yy524;
				if (yych <= 'Z') goto yy619;
				goto yy524;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy622;
				if (yych <= '^') goto yy524;
				goto yy619;
			} else {
				if (yych <= '`') goto yy524;
				if (yych <= 'z') goto yy619;
				if (yych <= 0x7F) goto yy524;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy623;
				if (yych <= 0xE0) goto yy624;
				goto yy625;
			} else {
				if (yych <= 0xF0) goto yy626;
				if (yych <= 0xF3) goto yy627;
				if (yych <= 0xF4) goto yy628;
				goto yy1;
			}
		}
	}
yy525:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '\'') {
				if (yych <= 0x00) goto yy1;
				if (yych <= '&') goto yy525;
				goto yy522;
			} else {
				if (yych == '-') goto yy629;
				if (yych <= '.') goto yy525;
				goto yy630;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy629;
				if (yych <= '=') goto yy525;
				goto yy631;
			} else {
				if (yych <= '?') goto yy630;
				if (yych <= '@') goto yy525;
				if (yych <= 'Z') goto yy629;
				goto yy525;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy632;
				if (yych <= '^') goto yy525;
				goto yy629;
			} else {
				if (yych <= '`') goto yy525;
				if (yych <= 'z') goto yy629;
				if (yych <= 0x7F) goto yy525;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy633;
				if (yych <= 0xE0) goto yy634;
				goto yy635;
			} else {
				if (yych <= 0xF0) goto yy636;
				if (yych <= 0xF3) goto yy637;
				if (yych <= 0xF4) goto yy638;
				goto yy1;
			}
		}
	}
yy526:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy522;
	goto yy1;
yy527:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy526;
	goto yy1;
yy528:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy526;
	goto yy1;
yy529:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy528;
	goto yy1;
yy530:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy528;
	goto yy1;
yy531:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy528;
	goto yy1;
yy532:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy639;
				goto yy148;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy639;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy639;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy533:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy640;
	goto yy1;
yy534:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy639;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy639;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy4;
				} else {
					if (yych <= 'f') goto yy641;
					if (yych <= 'z') goto yy326;
					goto yy4;
				}
			}
		}
	}
yy535:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy152;
	goto yy222;
yy536:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '9') {
			if (yych == '.') {
				yyt2 = YYCURSOR;
				goto yy644;
			}
			yyt2 = YYCURSOR;
			goto yy642;
		} else {
			if (yych <= ':') goto yy1;
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			goto yy1;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy645;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy646;
			}
			yyt2 = YYCURSOR;
			goto yy647;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy648;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy649;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy650;
			}
			goto yy1;
		}
	}
yy537:
	yych = *++YYCURSOR;
	if (yych == 'a') goto yy651;
	goto yy100;
yy538:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy24;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy24;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy24;
					if (yych <= 0x1F) goto yy4;
					goto yy24;
				}
			}
		} else {
			if (yych <= '$') {
				if (yych == '"') goto yy24;
				if (yych <= '#') goto yy99;
				goto yy101;
			} else {
				if (yych <= '\'') {
					if (yych <= '&') goto yy99;
					goto yy24;
				} else {
					if (yych <= '*') goto yy99;
					if (yych <= ',') goto yy24;
					goto yy101;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= '[') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy24;
				if (yych <= 'Z') goto yy101;
				goto yy24;
			} else {
				if (yych <= ']') {
					if (yych <= '\\') goto yy99;
					goto yy24;
				} else {
					if (yych <= '^') goto yy99;
					if (yych <= '_') goto yy101;
					goto yy24;
				}
			}
		} else {
			if (yych <= '}') {
				if (yych == 'm') goto yy652;
				if (yych <= 'z') goto yy101;
				goto yy24;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy99;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy24;
					if (yych <= 0xF4) goto yy4;
					goto yy24;
				}
			}
		}
	}
yy539:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '\n') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy655;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy655;
				} else {
					if (yych <= 0x1B) {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					if (yych <= 0x1F) {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy656;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy656;
			} else {
				if (yych <= '-') {
					if (yych <= ',') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy657;
				} else {
					if (yych <= '.') {
						yyt1 = YYCURSOR;
						goto yy658;
					}
					if (yych <= '/') {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					goto yy659;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy661;
				}
				yyt1 = YYCURSOR;
				goto yy662;
			} else {
				if (yych <= '_') {
					if (yych <= '^') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy662;
				} else {
					if (yych <= '`') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					if (yych <= 'z') {
						yyt1 = YYCURSOR;
						goto yy662;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy663;
				}
				yyt1 = YYCURSOR;
				goto yy664;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy665;
					}
					yyt1 = YYCURSOR;
					goto yy666;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy667;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy668;
					}
					goto yy345;
				}
			}
		}
	}
yy540:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '\n') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy655;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy655;
				} else {
					if (yych <= 0x1B) {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					if (yych <= 0x1F) {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy656;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy656;
			} else {
				if (yych <= '-') {
					if (yych <= ',') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy657;
				} else {
					if (yych <= '.') {
						yyt1 = YYCURSOR;
						goto yy658;
					}
					if (yych <= '/') {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					goto yy539;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy661;
				}
				yyt1 = YYCURSOR;
				goto yy662;
			} else {
				if (yych <= '_') {
					if (yych <= '^') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy662;
				} else {
					if (yych <= '`') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					if (yych <= 'z') {
						yyt1 = YYCURSOR;
						goto yy662;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy663;
				}
				yyt1 = YYCURSOR;
				goto yy664;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy665;
					}
					yyt1 = YYCURSOR;
					goto yy666;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy667;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy668;
					}
					goto yy345;
				}
			}
		}
	}
yy541:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '5') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '\n') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy655;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy655;
				} else {
					if (yych <= 0x1B) {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					if (yych <= 0x1F) {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					yyt1 = YYCURSOR;
					goto yy656;
				} else {
					if (yych == '+') {
						yyt1 = YYCURSOR;
						goto yy656;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') {
						yyt1 = YYCURSOR;
						goto yy657;
					}
					yyt1 = YYCURSOR;
					goto yy658;
				} else {
					if (yych <= '/') {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					if (yych <= '4') goto yy539;
					goto yy669;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '@') {
				if (yych <= '9') goto yy659;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '^') {
					if (yych <= 'Z') {
						yyt1 = YYCURSOR;
						goto yy662;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				} else {
					if (yych == '`') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy662;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy655;
				} else {
					if (yych <= 0xC1) goto yy345;
					if (yych <= 0xDF) {
						yyt1 = YYCURSOR;
						goto yy663;
					}
					yyt1 = YYCURSOR;
					goto yy664;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy665;
					}
					yyt1 = YYCURSOR;
					goto yy666;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy667;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy668;
					}
					goto yy345;
				}
			}
		}
	}
yy542:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy670;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy670;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy670;
		goto yy1;
	}
yy543:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy671;
	goto yy4;
yy544:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy672;
	goto yy1;
yy545:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy673;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy46;
			goto yy4;
		}
	}
yy546:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '-') goto yy545;
			if (yych <= '.') goto yy46;
			goto yy4;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy46;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy46;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy547:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy675;
	goto yy4;
yy548:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy547;
	goto yy4;
yy549:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy676;
					goto yy677;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy678;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy679;
					goto yy678;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy678;
				} else {
					if (yych <= 'e') goto yy679;
					if (yych <= 'f') goto yy678;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy550:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy677;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy678;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy679;
					goto yy678;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy678;
				} else {
					if (yych <= 'e') goto yy679;
					if (yych <= 'f') goto yy678;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy551:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy113;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy678;
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy113;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy678;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy113;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy678;
				if (yych <= 'z') goto yy115;
				goto yy113;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy113;
				if (yych <= 0xF4) goto yy4;
				goto yy113;
			}
		}
	}
yy552:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy251;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy678;
				if (yych <= ':') goto yy148;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy678;
				if (yych <= 'Z') goto yy115;
				goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy678;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy553:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy553;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy553;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy553;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy554:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy681;
					if (yych <= ':') goto yy682;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy683;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy683;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy555:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy684;
					if (yych <= ':') goto yy682;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy683;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy683;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy556:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '.') {
				if (yych <= '%') {
					if (yych <= '$') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt1 = YYCURSOR;
					goto yy680;
				} else {
					if (yych <= ',') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= '-') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy376;
				}
			} else {
				if (yych <= '4') {
					if (yych <= '/') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					goto yy684;
				} else {
					if (yych <= '5') goto yy685;
					if (yych <= '9') goto yy681;
					goto yy682;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy683;
				goto yy3;
			} else {
				if (yych <= '_') {
					if (yych <= '^') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '`') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 'f') goto yy683;
					goto yy3;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0xC1) goto yy5;
					if (yych <= 0xDF) {
						yyt2 = YYCURSOR;
						goto yy289;
					}
					yyt2 = YYCURSOR;
					goto yy290;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy557:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '#') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '9') {
					if (yych <= '/') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					goto yy683;
				} else {
					if (yych <= ':') goto yy682;
					if (yych <= '?') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '_') {
				if (yych <= 'F') goto yy683;
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy683;
				if (yych <= 'z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy558:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy686;
		if (yych <= ':') goto yy471;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy687;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy687;
		goto yy1;
	}
yy559:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy687;
		if (yych <= ':') goto yy471;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy687;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy687;
		goto yy1;
	}
yy560:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= '9') {
			if (yych == '.') {
				yyt2 = YYCURSOR;
				goto yy644;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			yyt1 = YYCURSOR;
			goto yy688;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy471;
				yyt2 = YYCURSOR;
				goto yy642;
			} else {
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy688;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy642;
				}
				yyt1 = YYCURSOR;
				goto yy688;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy645;
			}
			yyt2 = YYCURSOR;
			goto yy646;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy647;
				}
				yyt2 = YYCURSOR;
				goto yy648;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy649;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy650;
				}
				goto yy1;
			}
		}
	}
yy561:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy689;
		if (yych <= ':') goto yy471;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy689;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy689;
		goto yy1;
	}
yy562:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy690;
		if (yych <= ':') goto yy691;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy690;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy690;
		goto yy1;
	}
yy563:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy692;
			} else {
				if (yych <= '1') goto yy693;
				if (yych <= '2') goto yy694;
				goto yy692;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy695;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy695;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy564:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy696;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy696;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy696;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy565:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy695;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy695;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy695;
		goto yy1;
	}
yy566:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy696;
			if (yych <= ':') goto yy565;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy696;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy696;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy567:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy697;
				goto yy565;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy696;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy696;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy568:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '5') goto yy697;
				goto yy696;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy565;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy696;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy696;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy569:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '/') goto yy4;
		if (yych <= '0') goto yy698;
		goto yy699;
	} else {
		if (yych <= '2') goto yy700;
		if (yych <= '9') goto yy698;
		goto yy4;
	}
yy570:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy569;
	goto yy4;
yy571:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy569;
	if (yych <= '/') goto yy4;
	if (yych <= '5') goto yy570;
	goto yy4;
yy572:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych == '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x1A) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= ',') {
				if (yych <= 0x1F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '9') goto yy3;
				if (yych <= ':') goto yy378;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'z') goto yy3;
				if (yych <= '~') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy290;
				}
				yyt2 = YYCURSOR;
				goto yy291;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy292;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy293;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy294;
				}
				goto yy5;
			}
		}
	}
yy573:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy701;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy702;
	goto yy1;
yy574:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy701;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy573;
	goto yy1;
yy575:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych == '.') goto yy701;
		goto yy1;
	} else {
		if (yych <= '4') goto yy573;
		if (yych <= '5') goto yy703;
		if (yych <= '9') goto yy702;
		goto yy1;
	}
yy576:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy704;
			if (yych <= ':') goto yy486;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy704;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy704;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy577:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '0') {
			if (yych <= '/') goto yy1;
			goto yy705;
		} else {
			if (yych <= '1') goto yy706;
			if (yych <= '2') goto yy707;
			goto yy708;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy387;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy387;
			goto yy1;
		}
	}
yy578:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy578;
			goto yy1;
		} else {
			if (yych <= '\r') goto yy578;
			if (yych <= 0x1F) goto yy1;
			goto yy578;
		}
	} else {
		if (yych <= '&') {
			if (yych == '"') goto yy306;
			goto yy1;
		} else {
			if (yych <= '\'') goto yy307;
			if (yych == '>') goto yy309;
			goto yy1;
		}
	}
yy579:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 32) {
		goto yy579;
	}
	if (yych <= 0xDF) {
		if (yych <= '>') {
			if (yych <= 0x00) goto yy1;
			goto yy709;
		} else {
			if (yych <= '\\') goto yy710;
			if (yych <= 0xC1) goto yy1;
			goto yy711;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy712;
			if (yych <= 0xEF) goto yy713;
			goto yy714;
		} else {
			if (yych <= 0xF3) goto yy715;
			if (yych <= 0xF4) goto yy716;
			goto yy1;
		}
	}
yy580:
	yyaccept = 27;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[256+yych] & 8) {
		goto yy391;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy310;
			goto yy489;
		} else {
			if (yych <= '\\') goto yy392;
			if (yych <= 0xC1) goto yy310;
			goto yy393;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy394;
			if (yych <= 0xEF) goto yy395;
			goto yy396;
		} else {
			if (yych <= 0xF3) goto yy397;
			if (yych <= 0xF4) goto yy398;
			goto yy310;
		}
	}
yy581:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy391;
			if (yych == '\n') goto yy490;
			goto yy491;
		} else {
			if (yych <= '\'') goto yy579;
			if (yych == '>') goto yy580;
			goto yy491;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy582;
			if (yych <= 0xE0) goto yy583;
			goto yy584;
		} else {
			if (yych <= 0xF0) goto yy585;
			if (yych <= 0xF3) goto yy586;
			if (yych <= 0xF4) goto yy587;
			goto yy1;
		}
	}
yy582:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy491;
	goto yy1;
yy583:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy582;
	goto yy1;
yy584:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy582;
	goto yy1;
yy585:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy584;
	goto yy1;
yy586:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy584;
	goto yy1;
yy587:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy584;
	goto yy1;
yy588:
	yyaccept = 27;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[256+yych] & 16) {
		goto yy399;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy310;
			goto yy489;
		} else {
			if (yych <= '\\') goto yy400;
			if (yych <= 0xC1) goto yy310;
			goto yy401;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy402;
			if (yych <= 0xEF) goto yy403;
			goto yy404;
		} else {
			if (yych <= 0xF3) goto yy405;
			if (yych <= 0xF4) goto yy406;
			goto yy310;
		}
	}
yy589:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '!') {
			if (yych <= 0x00) goto yy399;
			if (yych == '\n') goto yy490;
			goto yy492;
		} else {
			if (yych <= '"') goto yy579;
			if (yych == '>') goto yy588;
			goto yy492;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy590;
			if (yych <= 0xE0) goto yy591;
			goto yy592;
		} else {
			if (yych <= 0xF0) goto yy593;
			if (yych <= 0xF3) goto yy594;
			if (yych <= 0xF4) goto yy595;
			goto yy1;
		}
	}
yy590:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy492;
	goto yy1;
yy591:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy590;
	goto yy1;
yy592:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy590;
	goto yy1;
yy593:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy592;
	goto yy1;
yy594:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy592;
	goto yy1;
yy595:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy592;
	goto yy1;
yy596:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy596;
				goto yy412;
			} else {
				if (yych <= '\r') goto yy596;
				if (yych == ' ') goto yy596;
				goto yy412;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy717;
				if (yych == '-') goto yy501;
				goto yy412;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy502;
					goto yy501;
				} else {
					if (yych <= '=') goto yy412;
					if (yych <= '>') goto yy503;
					goto yy502;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy412;
				if (yych <= 'Z') goto yy501;
				if (yych <= '[') goto yy412;
				goto yy504;
			} else {
				if (yych == '_') goto yy501;
				if (yych <= '`') goto yy412;
				goto yy501;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy412;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy505;
				goto yy506;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy507;
					goto yy508;
				} else {
					if (yych <= 0xF3) goto yy509;
					if (yych <= 0xF4) goto yy510;
					goto yy1;
				}
			}
		}
	}
yy597:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy315;
			if (yych >= '"') goto yy599;
		} else {
			if (yych == '\\') goto yy600;
			if (yych >= 0x80) goto yy315;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy601;
			if (yych <= 0xE0) goto yy602;
			goto yy603;
		} else {
			if (yych <= 0xF0) goto yy604;
			if (yych <= 0xF3) goto yy605;
			if (yych <= 0xF4) goto yy606;
			goto yy315;
		}
	}
yy598:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '!') goto yy598;
		} else {
			if (yych == '\\') goto yy600;
			if (yych <= 0x7F) goto yy598;
			goto yy1;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy601;
			if (yych <= 0xE0) goto yy602;
			goto yy603;
		} else {
			if (yych <= 0xF0) goto yy604;
			if (yych <= 0xF3) goto yy605;
			if (yych <= 0xF4) goto yy606;
			goto yy1;
		}
	}
yy599:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy718;
			} else {
				if (yych == '\r') goto yy718;
				goto yy1;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy718;
				goto yy1;
			} else {
				if (yych <= '-') goto yy311;
				if (yych <= '.') goto yy1;
				goto yy215;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= ':') goto yy311;
				goto yy1;
			} else {
				if (yych <= '>') goto yy216;
				if (yych <= '?') goto yy215;
				goto yy1;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy311;
				if (yych <= '^') goto yy1;
				goto yy311;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy311;
				goto yy1;
			}
		}
	}
yy600:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy598;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy598;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy602;
			if (yych <= 0xEF) goto yy603;
			goto yy604;
		} else {
			if (yych <= 0xF3) goto yy605;
			if (yych <= 0xF4) goto yy606;
			goto yy1;
		}
	}
yy601:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy598;
	goto yy1;
yy602:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy601;
	goto yy1;
yy603:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy601;
	goto yy1;
yy604:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy603;
	goto yy1;
yy605:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy603;
	goto yy1;
yy606:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy603;
	goto yy1;
yy607:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy607;
				goto yy413;
			} else {
				if (yych <= '\r') goto yy607;
				if (yych == ' ') goto yy607;
				goto yy413;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy717;
				if (yych == '-') goto yy511;
				goto yy413;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy512;
					goto yy511;
				} else {
					if (yych <= '=') goto yy413;
					if (yych <= '>') goto yy513;
					goto yy512;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy413;
				if (yych <= 'Z') goto yy511;
				if (yych <= '[') goto yy413;
				goto yy514;
			} else {
				if (yych == '_') goto yy511;
				if (yych <= '`') goto yy413;
				goto yy511;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy413;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy515;
				goto yy516;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy517;
					goto yy518;
				} else {
					if (yych <= 0xF3) goto yy519;
					if (yych <= 0xF4) goto yy520;
					goto yy1;
				}
			}
		}
	}
yy608:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy315;
			if (yych >= '\'') goto yy599;
		} else {
			if (yych == '\\') goto yy610;
			if (yych >= 0x80) goto yy315;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy611;
			if (yych <= 0xE0) goto yy612;
			goto yy613;
		} else {
			if (yych <= 0xF0) goto yy614;
			if (yych <= 0xF3) goto yy615;
			if (yych <= 0xF4) goto yy616;
			goto yy315;
		}
	}
yy609:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '&') goto yy609;
			goto yy599;
		} else {
			if (yych == '\\') goto yy610;
			if (yych <= 0x7F) goto yy609;
			goto yy1;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy611;
			if (yych <= 0xE0) goto yy612;
			goto yy613;
		} else {
			if (yych <= 0xF0) goto yy614;
			if (yych <= 0xF3) goto yy615;
			if (yych <= 0xF4) goto yy616;
			goto yy1;
		}
	}
yy610:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy609;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy609;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy612;
			if (yych <= 0xEF) goto yy613;
			goto yy614;
		} else {
			if (yych <= 0xF3) goto yy615;
			if (yych <= 0xF4) goto yy616;
			goto yy1;
		}
	}
yy611:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy609;
	goto yy1;
yy612:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy611;
	goto yy1;
yy613:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy611;
	goto yy1;
yy614:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy613;
	goto yy1;
yy615:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy613;
	goto yy1;
yy616:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy613;
	goto yy1;
yy617:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy522;
				goto yy617;
			} else {
				if (yych == '\r') goto yy617;
				if (yych <= 0x1F) goto yy522;
				goto yy617;
			}
		} else {
			if (yych <= '/') {
				if (yych == '-') goto yy617;
				if (yych <= '.') goto yy522;
			} else {
				if (yych <= ':') goto yy617;
				if (yych <= '<') goto yy522;
				if (yych <= '=') goto yy523;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '?') goto yy618;
				if (yych <= '@') goto yy522;
				if (yych <= 'Z') goto yy617;
				goto yy522;
			} else {
				if (yych == '`') goto yy522;
				if (yych <= 'z') goto yy617;
				goto yy522;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy526;
				if (yych <= 0xE0) goto yy527;
				goto yy528;
			} else {
				if (yych <= 0xF0) goto yy529;
				if (yych <= 0xF3) goto yy530;
				if (yych <= 0xF4) goto yy531;
				goto yy1;
			}
		}
	}
yy618:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1;
				if (yych <= ',') goto yy522;
				goto yy617;
			} else {
				if (yych <= '.') goto yy522;
				if (yych <= '/') goto yy618;
				goto yy617;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy522;
				if (yych <= '>') goto yy314;
				goto yy618;
			} else {
				if (yych <= '@') goto yy522;
				if (yych <= 'Z') goto yy617;
				goto yy522;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy522;
				goto yy617;
			} else {
				if (yych <= 0x7F) goto yy522;
				if (yych <= 0xC1) goto yy1;
				goto yy526;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy527;
				if (yych <= 0xEF) goto yy528;
				goto yy529;
			} else {
				if (yych <= 0xF3) goto yy530;
				if (yych <= 0xF4) goto yy531;
				goto yy1;
			}
		}
	}
yy619:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy619;
				goto yy524;
			} else {
				if (yych <= '\r') goto yy619;
				if (yych == ' ') goto yy619;
				goto yy524;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy522;
				if (yych == '-') goto yy619;
				goto yy524;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy619;
				} else {
					if (yych <= '<') goto yy524;
					if (yych <= '=') goto yy719;
					goto yy621;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy620;
				if (yych <= '@') goto yy524;
				if (yych <= 'Z') goto yy619;
				goto yy524;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy622;
					goto yy524;
				} else {
					if (yych == '`') goto yy524;
					goto yy619;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy524;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy623;
				goto yy624;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy625;
					goto yy626;
				} else {
					if (yych <= 0xF3) goto yy627;
					if (yych <= 0xF4) goto yy628;
					goto yy1;
				}
			}
		}
	}
yy620:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy1;
				if (yych <= '!') goto yy524;
				goto yy522;
			} else {
				if (yych == '-') goto yy619;
				if (yych <= '.') goto yy524;
				goto yy620;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy619;
				if (yych <= '=') goto yy524;
				goto yy720;
			} else {
				if (yych <= '?') goto yy620;
				if (yych <= '@') goto yy524;
				if (yych <= 'Z') goto yy619;
				goto yy524;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy622;
				if (yych <= '^') goto yy524;
				goto yy619;
			} else {
				if (yych <= '`') goto yy524;
				if (yych <= 'z') goto yy619;
				if (yych <= 0x7F) goto yy524;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy623;
				if (yych <= 0xE0) goto yy624;
				goto yy625;
			} else {
				if (yych <= 0xF0) goto yy626;
				if (yych <= 0xF3) goto yy627;
				if (yych <= 0xF4) goto yy628;
				goto yy1;
			}
		}
	}
yy621:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '!') goto yy621;
			goto yy721;
		} else {
			if (yych == '\\') goto yy722;
			if (yych <= 0x7F) goto yy621;
			goto yy1;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy723;
			if (yych <= 0xE0) goto yy724;
			goto yy725;
		} else {
			if (yych <= 0xF0) goto yy726;
			if (yych <= 0xF3) goto yy727;
			if (yych <= 0xF4) goto yy728;
			goto yy1;
		}
	}
yy622:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '.') {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy621;
				if (yych <= '\t') goto yy524;
				goto yy522;
			} else {
				if (yych == '-') goto yy619;
				goto yy524;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '/') goto yy620;
				if (yych <= ':') goto yy619;
				goto yy524;
			} else {
				if (yych <= '>') goto yy621;
				if (yych <= '?') goto yy620;
				if (yych <= '@') goto yy524;
				goto yy619;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') goto yy619;
				goto yy524;
			} else {
				if (yych <= 'z') goto yy619;
				if (yych <= 0x7F) goto yy524;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy623;
				if (yych <= 0xE0) goto yy624;
				goto yy625;
			} else {
				if (yych <= 0xF0) goto yy626;
				if (yych <= 0xF3) goto yy627;
				if (yych <= 0xF4) goto yy628;
				goto yy1;
			}
		}
	}
yy623:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy524;
	goto yy1;
yy624:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy623;
	goto yy1;
yy625:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy623;
	goto yy1;
yy626:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy625;
	goto yy1;
yy627:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy625;
	goto yy1;
yy628:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy625;
	goto yy1;
yy629:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy629;
				goto yy525;
			} else {
				if (yych <= '\r') goto yy629;
				if (yych == ' ') goto yy629;
				goto yy525;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy522;
				if (yych == '-') goto yy629;
				goto yy525;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy629;
				} else {
					if (yych <= '<') goto yy525;
					if (yych <= '=') goto yy729;
					goto yy631;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy630;
				if (yych <= '@') goto yy525;
				if (yych <= 'Z') goto yy629;
				goto yy525;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy632;
					goto yy525;
				} else {
					if (yych == '`') goto yy525;
					goto yy629;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy525;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy633;
				goto yy634;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy635;
					goto yy636;
				} else {
					if (yych <= 0xF3) goto yy637;
					if (yych <= 0xF4) goto yy638;
					goto yy1;
				}
			}
		}
	}
yy630:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '\'') {
				if (yych <= 0x00) goto yy1;
				if (yych <= '&') goto yy525;
				goto yy522;
			} else {
				if (yych == '-') goto yy629;
				if (yych <= '.') goto yy525;
				goto yy630;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy629;
				if (yych <= '=') goto yy525;
				goto yy730;
			} else {
				if (yych <= '?') goto yy630;
				if (yych <= '@') goto yy525;
				if (yych <= 'Z') goto yy629;
				goto yy525;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy632;
				if (yych <= '^') goto yy525;
				goto yy629;
			} else {
				if (yych <= '`') goto yy525;
				if (yych <= 'z') goto yy629;
				if (yych <= 0x7F) goto yy525;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy633;
				if (yych <= 0xE0) goto yy634;
				goto yy635;
			} else {
				if (yych <= 0xF0) goto yy636;
				if (yych <= 0xF3) goto yy637;
				if (yych <= 0xF4) goto yy638;
				goto yy1;
			}
		}
	}
yy631:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '&') goto yy631;
			goto yy721;
		} else {
			if (yych == '\\') goto yy731;
			if (yych <= 0x7F) goto yy631;
			goto yy1;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy732;
			if (yych <= 0xE0) goto yy733;
			goto yy734;
		} else {
			if (yych <= 0xF0) goto yy735;
			if (yych <= 0xF3) goto yy736;
			if (yych <= 0xF4) goto yy737;
			goto yy1;
		}
	}
yy632:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '.') {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy631;
				if (yych <= '\t') goto yy525;
				goto yy522;
			} else {
				if (yych == '-') goto yy629;
				goto yy525;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '/') goto yy630;
				if (yych <= ':') goto yy629;
				goto yy525;
			} else {
				if (yych <= '>') goto yy631;
				if (yych <= '?') goto yy630;
				if (yych <= '@') goto yy525;
				goto yy629;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') goto yy629;
				goto yy525;
			} else {
				if (yych <= 'z') goto yy629;
				if (yych <= 0x7F) goto yy525;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy633;
				if (yych <= 0xE0) goto yy634;
				goto yy635;
			} else {
				if (yych <= 0xF0) goto yy636;
				if (yych <= 0xF3) goto yy637;
				if (yych <= 0xF4) goto yy638;
				goto yy1;
			}
		}
	}
yy633:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy525;
	goto yy1;
yy634:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy633;
	goto yy1;
yy635:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy633;
	goto yy1;
yy636:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy635;
	goto yy1;
yy637:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy635;
	goto yy1;
yy638:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy635;
	goto yy1;
yy639:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy738;
				goto yy148;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy4;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy738;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy738;
				if (yych <= 'z') goto yy115;
				goto yy4;
			}
		}
	}
yy640:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy739;
	goto yy1;
yy641:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			} else {
				if (yych <= '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych == '%') goto yy80;
				goto yy4;
			} else {
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '+') goto yy80;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '9') {
				if (yych <= '-') goto yy46;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				if (yych <= '/') goto yy4;
				goto yy738;
			} else {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '>') goto yy4;
				yyt4 = YYCURSOR;
				goto yy223;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy738;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy4;
				} else {
					if (yych <= 'f') goto yy740;
					if (yych <= 'z') goto yy326;
					goto yy4;
				}
			}
		}
	}
yy642:
	++YYCURSOR;
	yyt1 = yyt2;
yy643:
	YYCURSOR = yyt1;
#line 173 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_TIME); }
#line 20685 "../../lnav/src/data_scanner_re.cc"
yy644:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	yyt1 = yyt2;
	if (yych <= '/') goto yy643;
	if (yych <= '9') goto yy741;
	goto yy643;
yy645:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy642;
	goto yy1;
yy646:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy645;
	goto yy1;
yy647:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy645;
	goto yy1;
yy648:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy647;
	goto yy1;
yy649:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy647;
	goto yy1;
yy650:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy647;
	goto yy1;
yy651:
	yych = *++YYCURSOR;
	if (yych == 'm') goto yy742;
	goto yy100;
yy652:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 16) {
		goto yy101;
	}
	if (yych <= '\'') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy24;
				if (yych <= 0x08) goto yy4;
				goto yy24;
			} else {
				if (yych == '\r') goto yy24;
				goto yy4;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1B) goto yy24;
				if (yych <= 0x1F) goto yy4;
				goto yy743;
			} else {
				if (yych == '"') goto yy24;
				if (yych <= '&') goto yy99;
				goto yy24;
			}
		}
	} else {
		if (yych <= ']') {
			if (yych <= ':') {
				if (yych <= '*') goto yy99;
				if (yych <= ',') goto yy24;
				goto yy4;
			} else {
				if (yych == '\\') goto yy99;
				goto yy24;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '^') goto yy99;
				if (yych <= '}') goto yy24;
				goto yy99;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy24;
				if (yych <= 0xF4) goto yy4;
				goto yy24;
			}
		}
	}
yy653:
	++YYCURSOR;
yy654:
	YYCURSOR = yyt1;
#line 230 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_IPV4_ADDRESS);
       }
#line 20785 "../../lnav/src/data_scanner_re.cc"
yy655:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy654;
				goto yy4;
			} else {
				if (yych <= '\n') goto yy654;
				if (yych <= '\f') goto yy4;
				goto yy654;
			}
		} else {
			if (yych <= 0x1F) {
				if (yych == 0x1B) goto yy654;
				goto yy4;
			} else {
				if (yych == '$') goto yy4;
				goto yy654;
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy654;
				goto yy4;
			} else {
				if (yych == '_') goto yy4;
				goto yy654;
			}
		} else {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy4;
				if (yych <= '~') goto yy654;
				goto yy4;
			} else {
				if (yych <= 0xC1) goto yy654;
				if (yych <= 0xF4) goto yy4;
				goto yy654;
			}
		}
	}
yy656:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy81;
			goto yy654;
		} else {
			if (yych == ',') goto yy654;
			if (yych <= '.') goto yy81;
			goto yy654;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy81;
			if (yych <= '?') goto yy654;
			if (yych <= 'Z') goto yy81;
			goto yy654;
		} else {
			if (yych == '`') goto yy654;
			if (yych <= 'z') goto yy81;
			goto yy654;
		}
	}
yy657:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy654;
				if (yych <= 0x08) goto yy4;
				goto yy654;
			} else {
				if (yych == '\r') goto yy654;
				if (yych <= 0x1A) goto yy4;
				goto yy654;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy654;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy654;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy344;
				goto yy4;
			} else {
				if (yych <= '?') goto yy654;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy344;
				goto yy654;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy654;
				if (yych <= 'z') goto yy344;
				goto yy654;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy654;
				if (yych <= 0xF4) goto yy4;
				goto yy654;
			}
		}
	}
yy658:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy654;
				if (yych <= 0x08) goto yy4;
				goto yy654;
			} else {
				if (yych == '\r') goto yy654;
				if (yych <= 0x1A) goto yy4;
				goto yy654;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy654;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy654;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy447;
				goto yy4;
			} else {
				if (yych <= '?') goto yy654;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy46;
				goto yy654;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy654;
				if (yych <= 'z') goto yy46;
				goto yy654;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy654;
				if (yych <= 0xF4) goto yy4;
				goto yy654;
			}
		}
	}
yy659:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '\n') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy655;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy655;
				} else {
					if (yych <= 0x1B) {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					if (yych <= 0x1F) {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy656;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy656;
			} else {
				if (yych <= '-') {
					if (yych <= ',') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy657;
				} else {
					if (yych <= '.') {
						yyt1 = YYCURSOR;
						goto yy658;
					}
					if (yych <= '/') {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					goto yy447;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= 'Z') {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy661;
				}
				yyt1 = YYCURSOR;
				goto yy662;
			} else {
				if (yych <= '_') {
					if (yych <= '^') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy662;
				} else {
					if (yych <= '`') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					if (yych <= 'z') {
						yyt1 = YYCURSOR;
						goto yy662;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= 0xC1) goto yy345;
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy663;
				}
				yyt1 = YYCURSOR;
				goto yy664;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy665;
					}
					yyt1 = YYCURSOR;
					goto yy666;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy667;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy668;
					}
					goto yy345;
				}
			}
		}
	}
yy660:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy152;
	goto yy654;
yy661:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy654;
				if (yych <= 0x08) goto yy4;
				goto yy654;
			} else {
				if (yych == '\r') goto yy654;
				goto yy4;
			}
		} else {
			if (yych <= '#') {
				if (yych <= 0x1B) goto yy654;
				if (yych <= 0x1F) goto yy4;
				goto yy654;
			} else {
				if (yych <= '$') goto yy4;
				if (yych <= ',') goto yy654;
				if (yych <= '.') goto yy174;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '?') {
				if (yych <= '9') goto yy174;
				if (yych <= ':') goto yy4;
				goto yy654;
			} else {
				if (yych <= '@') goto yy4;
				if (yych <= 'Z') goto yy174;
				if (yych <= '^') goto yy654;
				goto yy4;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy654;
				if (yych <= 'z') goto yy174;
				goto yy654;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy654;
				if (yych <= 0xF4) goto yy4;
				goto yy654;
			}
		}
	}
yy662:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy654;
				if (yych <= 0x08) goto yy4;
				goto yy654;
			} else {
				if (yych == '\r') goto yy654;
				if (yych <= 0x1A) goto yy4;
				goto yy654;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy654;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy654;
				goto yy255;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '.') goto yy350;
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy447;
				goto yy4;
			} else {
				if (yych <= '?') goto yy654;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy447;
				goto yy654;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy654;
				if (yych <= 'z') goto yy447;
				goto yy654;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy654;
				if (yych <= 0xF4) goto yy4;
				goto yy654;
			}
		}
	}
yy663:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy655;
	goto yy1;
yy664:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy663;
	goto yy1;
yy665:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy663;
	goto yy1;
yy666:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy665;
	goto yy1;
yy667:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy665;
	goto yy1;
yy668:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy665;
	goto yy1;
yy669:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '5') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '\n') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy655;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy655;
				} else {
					if (yych <= 0x1B) {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					if (yych <= 0x1F) {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') {
					yyt1 = YYCURSOR;
					goto yy655;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy656;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy656;
			} else {
				if (yych <= '-') {
					if (yych <= ',') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy657;
				} else {
					if (yych <= '.') {
						yyt1 = YYCURSOR;
						goto yy658;
					}
					if (yych <= '/') {
						yyt1 = YYCURSOR;
						goto yy655;
					}
					goto yy659;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '@') {
				if (yych <= '9') goto yy447;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy653;
				}
				yyt1 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '^') {
					if (yych <= 'Z') {
						yyt1 = YYCURSOR;
						goto yy662;
					}
					yyt1 = YYCURSOR;
					goto yy653;
				} else {
					if (yych == '`') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy662;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') {
						yyt1 = YYCURSOR;
						goto yy653;
					}
					yyt1 = YYCURSOR;
					goto yy655;
				} else {
					if (yych <= 0xC1) goto yy345;
					if (yych <= 0xDF) {
						yyt1 = YYCURSOR;
						goto yy663;
					}
					yyt1 = YYCURSOR;
					goto yy664;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy665;
					}
					yyt1 = YYCURSOR;
					goto yy666;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy667;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy668;
					}
					goto yy345;
				}
			}
		}
	}
yy670:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy466;
	if (yych == ':') goto yy466;
	goto yy354;
yy671:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy744;
	goto yy4;
yy672:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy745;
	goto yy1;
yy673:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy674;
				if (yych <= 0x08) goto yy4;
				if (yych >= '\v') goto yy4;
			} else {
				if (yych <= '\r') goto yy674;
				if (yych != 0x1B) goto yy4;
			}
		} else {
			if (yych <= '%') {
				if (yych <= ' ') goto yy746;
				if (yych <= '#') goto yy674;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych >= '-') goto yy46;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy747;
				if (yych <= ':') goto yy4;
			} else {
				if (yych <= '@') goto yy96;
				if (yych == 'T') goto yy748;
				goto yy46;
			}
		} else {
			if (yych <= 'z') {
				if (yych == '_') goto yy46;
				if (yych >= 'a') goto yy46;
			} else {
				if (yych <= 0x7F) {
					if (yych >= 0x7F) goto yy4;
				} else {
					if (yych <= 0xC1) goto yy674;
					if (yych <= 0xF4) goto yy4;
				}
			}
		}
	}
yy674:
#line 187 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE);
       }
#line 21464 "../../lnav/src/data_scanner_re.cc"
yy675:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy674;
				if (yych <= 0x08) goto yy4;
				goto yy674;
			} else {
				if (yych == '\r') goto yy674;
				if (yych <= 0x1A) goto yy4;
				goto yy674;
			}
		} else {
			if (yych <= '#') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= ' ') goto yy746;
				goto yy674;
			} else {
				if (yych <= '$') goto yy4;
				if (yych <= ',') goto yy674;
				if (yych <= '/') goto yy4;
				goto yy749;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'S') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy674;
				goto yy4;
			} else {
				if (yych <= 'T') goto yy750;
				if (yych <= 'Z') goto yy4;
				if (yych <= '^') goto yy674;
				goto yy4;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy674;
				if (yych <= 'z') goto yy4;
				goto yy674;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy674;
				if (yych <= 0xF4) goto yy4;
				goto yy674;
			}
		}
	}
yy676:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy751;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy752;
					goto yy753;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy677:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy751;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy753;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy678:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy751;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy196;
				if (yych <= ':') goto yy148;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy196;
				if (yych <= 'Z') goto yy115;
				goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy196;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy679:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy755;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy196;
				if (yych <= ':') goto yy148;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy196;
				if (yych <= 'Z') goto yy115;
				goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy196;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy680:
	yyaccept = 18;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			yyt2 = yyt1;
			goto yy199;
		}
		if (yych <= '9') goto yy553;
		yyt2 = yyt1;
		goto yy199;
	} else {
		if (yych <= 'Z') goto yy553;
		if (yych <= '`') {
			yyt2 = yyt1;
			goto yy199;
		}
		if (yych <= 'z') goto yy553;
		yyt2 = yyt1;
		goto yy199;
	}
yy681:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy756;
					if (yych >= ';') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy756;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy756;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy682:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy757;
		if (yych <= ':') goto yy152;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy757;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy757;
		goto yy1;
	}
yy683:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '#') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '9') {
					if (yych <= '/') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					goto yy756;
				} else {
					if (yych <= ':') goto yy682;
					if (yych <= '?') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '_') {
				if (yych <= 'F') goto yy756;
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy756;
				if (yych <= 'z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy684:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy758;
					if (yych <= ':') goto yy682;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy756;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy756;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy685:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '5') goto yy758;
					if (yych <= '9') goto yy756;
					goto yy682;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy756;
				goto yy3;
			} else {
				if (yych <= '_') {
					if (yych <= '^') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '`') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 'f') goto yy756;
					goto yy3;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= '~') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0xC1) goto yy5;
					if (yych <= 0xDF) {
						yyt2 = YYCURSOR;
						goto yy289;
					}
					yyt2 = YYCURSOR;
					goto yy290;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy686:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '/') {
			if (yych <= ',') {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= '-') {
				yyt1 = YYCURSOR;
				goto yy759;
			}
			if (yych <= '.') {
				yyt2 = YYCURSOR;
				goto yy644;
			}
			yyt2 = YYCURSOR;
			goto yy642;
		} else {
			if (yych <= ':') {
				if (yych <= '9') {
					yyt1 = YYCURSOR;
					goto yy688;
				}
				goto yy760;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy642;
				}
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy688;
				}
				yyt2 = YYCURSOR;
				goto yy642;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'f') {
					yyt1 = YYCURSOR;
					goto yy688;
				}
				yyt2 = YYCURSOR;
				goto yy642;
			} else {
				if (yych <= 0xC1) goto yy354;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy645;
				}
				yyt2 = YYCURSOR;
				goto yy646;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy647;
				}
				yyt2 = YYCURSOR;
				goto yy648;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy649;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy650;
				}
				goto yy354;
			}
		}
	}
yy687:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy354;
			goto yy466;
		} else {
			if (yych <= '/') goto yy354;
			if (yych <= '9') goto yy561;
			goto yy760;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy354;
			goto yy561;
		} else {
			if (yych <= '`') goto yy354;
			if (yych <= 'f') goto yy561;
			goto yy354;
		}
	}
yy688:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy643;
		if (yych <= '9') goto yy689;
		if (yych <= ':') goto yy471;
		goto yy643;
	} else {
		if (yych <= 'F') goto yy689;
		if (yych <= '`') goto yy643;
		if (yych >= 'g') goto yy643;
	}
yy689:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy471;
	goto yy1;
yy690:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy761;
		if (yych >= ';') goto yy1;
	} else {
		if (yych <= 'F') goto yy761;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy761;
		goto yy1;
	}
yy691:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy762;
		if (yych <= ':') goto yy763;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy762;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy762;
		goto yy1;
	}
yy692:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy764;
				goto yy765;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy766;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy766;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy693:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy767;
				goto yy765;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy766;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy766;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy694:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '4') goto yy767;
				goto yy768;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy764;
				if (yych <= ':') goto yy765;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy766;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy766;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy695:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy766;
			if (yych <= ':') goto yy765;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy766;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy766;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy696:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy769;
			if (yych <= ':') goto yy565;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy769;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy769;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy697:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy769;
				goto yy565;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy769;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy769;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy698:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy770;
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy771;
	goto yy4;
yy699:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy770;
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy698;
	goto yy4;
yy700:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy770;
		goto yy4;
	} else {
		if (yych <= '4') goto yy698;
		if (yych <= '5') goto yy772;
		if (yych <= '9') goto yy771;
		goto yy4;
	}
yy701:
	yych = *++YYCURSOR;
	if (yych <= '1') {
		if (yych <= '/') goto yy1;
		if (yych <= '0') goto yy773;
		goto yy774;
	} else {
		if (yych <= '2') goto yy775;
		if (yych <= '9') goto yy773;
		goto yy1;
	}
yy702:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy701;
	goto yy1;
yy703:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy701;
	if (yych <= '/') goto yy1;
	if (yych <= '5') goto yy702;
	goto yy1;
yy704:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1;
			if (yych <= ':') goto yy486;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy705:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy776;
				goto yy777;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy485;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy485;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy706:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy778;
				goto yy486;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy485;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy485;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy707:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '4') goto yy778;
				goto yy779;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy776;
				if (yych <= ':') goto yy486;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy485;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy485;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy708:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy776;
				goto yy486;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy485;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy485;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy709:
	yyaccept = 27;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[0+yych] & 128) {
		goto yy780;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy310;
			if (yych <= '"') goto yy781;
			goto yy782;
		} else {
			if (yych <= '\\') goto yy783;
			if (yych <= 0xC1) goto yy310;
			goto yy784;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy785;
			if (yych <= 0xEF) goto yy786;
			goto yy787;
		} else {
			if (yych <= 0xF3) goto yy788;
			if (yych <= 0xF4) goto yy789;
			goto yy310;
		}
	}
yy710:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy780;
			if (yych <= '\t') goto yy579;
			goto yy490;
		} else {
			if (yych == '>') goto yy709;
			if (yych <= 0x7F) goto yy579;
			goto yy1;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy711;
			if (yych <= 0xE0) goto yy712;
			goto yy713;
		} else {
			if (yych <= 0xF0) goto yy714;
			if (yych <= 0xF3) goto yy715;
			if (yych <= 0xF4) goto yy716;
			goto yy1;
		}
	}
yy711:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy579;
	goto yy1;
yy712:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy711;
	goto yy1;
yy713:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy711;
	goto yy1;
yy714:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy713;
	goto yy1;
yy715:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy713;
	goto yy1;
yy716:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy713;
	goto yy1;
yy717:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 64) {
		goto yy717;
	}
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '"') goto yy413;
			if (yych <= '\'') goto yy412;
			goto yy790;
		} else {
			if (yych <= ':') {
				if (yych <= '/') goto yy791;
				goto yy790;
			} else {
				if (yych <= '>') goto yy792;
				if (yych <= '?') goto yy791;
				goto yy790;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') {
				if (yych <= '\\') goto yy793;
				goto yy790;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy794;
				goto yy795;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy796;
				goto yy797;
			} else {
				if (yych <= 0xF3) goto yy798;
				if (yych <= 0xF4) goto yy799;
				goto yy1;
			}
		}
	}
yy718:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy718;
			goto yy1;
		} else {
			if (yych <= '\r') goto yy718;
			if (yych <= 0x1F) goto yy1;
			goto yy718;
		}
	} else {
		if (yych <= '=') {
			if (yych == '/') goto yy215;
			goto yy1;
		} else {
			if (yych <= '>') goto yy216;
			if (yych <= '?') goto yy215;
			goto yy1;
		}
	}
yy719:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy719;
				goto yy524;
			} else {
				if (yych <= '\r') goto yy719;
				if (yych == ' ') goto yy719;
				goto yy524;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy800;
				if (yych == '-') goto yy619;
				goto yy524;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy620;
					goto yy619;
				} else {
					if (yych <= '=') goto yy524;
					if (yych <= '>') goto yy621;
					goto yy620;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy524;
				if (yych <= 'Z') goto yy619;
				if (yych <= '[') goto yy524;
				goto yy622;
			} else {
				if (yych == '_') goto yy619;
				if (yych <= '`') goto yy524;
				goto yy619;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy524;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy623;
				goto yy624;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy625;
					goto yy626;
				} else {
					if (yych <= 0xF3) goto yy627;
					if (yych <= 0xF4) goto yy628;
					goto yy1;
				}
			}
		}
	}
yy720:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy315;
			if (yych <= '!') goto yy621;
		} else {
			if (yych == '\\') goto yy722;
			if (yych <= 0x7F) goto yy621;
			goto yy315;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy723;
			if (yych <= 0xE0) goto yy724;
			goto yy725;
		} else {
			if (yych <= 0xF0) goto yy726;
			if (yych <= 0xF3) goto yy727;
			if (yych <= 0xF4) goto yy728;
			goto yy315;
		}
	}
yy721:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1;
				goto yy801;
			} else {
				if (yych == '\r') goto yy801;
				goto yy1;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy801;
				goto yy1;
			} else {
				if (yych <= '-') goto yy420;
				if (yych <= '.') goto yy1;
				goto yy215;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '>') {
				if (yych <= ':') goto yy420;
				goto yy1;
			} else {
				if (yych <= '?') goto yy215;
				if (yych <= '@') goto yy1;
				goto yy420;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1;
				goto yy420;
			} else {
				if (yych <= '`') goto yy1;
				if (yych <= 'z') goto yy420;
				goto yy1;
			}
		}
	}
yy722:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy621;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy621;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy724;
			if (yych <= 0xEF) goto yy725;
			goto yy726;
		} else {
			if (yych <= 0xF3) goto yy727;
			if (yych <= 0xF4) goto yy728;
			goto yy1;
		}
	}
yy723:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy621;
	goto yy1;
yy724:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy723;
	goto yy1;
yy725:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy723;
	goto yy1;
yy726:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy725;
	goto yy1;
yy727:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy725;
	goto yy1;
yy728:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy725;
	goto yy1;
yy729:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy729;
				goto yy525;
			} else {
				if (yych <= '\r') goto yy729;
				if (yych == ' ') goto yy729;
				goto yy525;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy800;
				if (yych == '-') goto yy629;
				goto yy525;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy630;
					goto yy629;
				} else {
					if (yych <= '=') goto yy525;
					if (yych <= '>') goto yy631;
					goto yy630;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy525;
				if (yych <= 'Z') goto yy629;
				if (yych <= '[') goto yy525;
				goto yy632;
			} else {
				if (yych == '_') goto yy629;
				if (yych <= '`') goto yy525;
				goto yy629;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy525;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy633;
				goto yy634;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy635;
					goto yy636;
				} else {
					if (yych <= 0xF3) goto yy637;
					if (yych <= 0xF4) goto yy638;
					goto yy1;
				}
			}
		}
	}
yy730:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy315;
			if (yych <= '&') goto yy631;
			goto yy721;
		} else {
			if (yych == '\\') goto yy731;
			if (yych <= 0x7F) goto yy631;
			goto yy315;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy732;
			if (yych <= 0xE0) goto yy733;
			goto yy734;
		} else {
			if (yych <= 0xF0) goto yy735;
			if (yych <= 0xF3) goto yy736;
			if (yych <= 0xF4) goto yy737;
			goto yy315;
		}
	}
yy731:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy631;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy631;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy733;
			if (yych <= 0xEF) goto yy734;
			goto yy735;
		} else {
			if (yych <= 0xF3) goto yy736;
			if (yych <= 0xF4) goto yy737;
			goto yy1;
		}
	}
yy732:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy631;
	goto yy1;
yy733:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy732;
	goto yy1;
yy734:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy732;
	goto yy1;
yy735:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy734;
	goto yy1;
yy736:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy734;
	goto yy1;
yy737:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy734;
	goto yy1;
yy738:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 64) {
		goto yy115;
	}
	if (yych <= '.') {
		if (yych <= '*') {
			if (yych == '%') goto yy80;
			goto yy4;
		} else {
			if (yych <= '+') goto yy80;
			if (yych <= ',') goto yy4;
			if (yych <= '-') goto yy751;
			goto yy46;
		}
	} else {
		if (yych <= '?') {
			if (yych <= '/') goto yy4;
			if (yych <= ':') goto yy148;
			goto yy4;
		} else {
			if (yych <= '@') goto yy96;
			if (yych == '_') goto yy46;
			goto yy4;
		}
	}
yy739:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy802;
	goto yy1;
yy740:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= ' ') {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x08) goto yy4;
				yyt4 = YYCURSOR;
				goto yy221;
			} else {
				if (yych == '\r') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= 0x1F) goto yy4;
				yyt4 = YYCURSOR;
				goto yy221;
			}
		} else {
			if (yych <= '%') {
				if (yych <= '!') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '"') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych <= '&') goto yy4;
				if (yych <= '\'') {
					yyt4 = YYCURSOR;
					goto yy224;
				}
				if (yych <= '*') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy80;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '-') goto yy751;
				if (yych <= '.') {
					yyt4 = YYCURSOR;
					goto yy230;
				}
				goto yy4;
			} else {
				if (yych <= '9') goto yy115;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy234;
				}
				if (yych <= ';') {
					yyt4 = YYCURSOR;
					goto yy221;
				}
				goto yy4;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt4 = YYCURSOR;
					goto yy223;
				}
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy115;
				goto yy4;
			} else {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy326;
				goto yy4;
			}
		}
	}
yy741:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy803;
	goto yy1;
yy742:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych != ' ') goto yy100;
yy743:
	yych = *++YYCURSOR;
	if (yych == 'F') goto yy804;
	goto yy1;
yy744:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy805;
	goto yy4;
yy745:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy806;
	goto yy1;
yy746:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy807;
	goto yy1;
yy747:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy674;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy674;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy674;
				if (yych == 0x1B) goto yy674;
				goto yy4;
			}
		} else {
			if (yych <= '%') {
				if (yych <= ' ') goto yy746;
				if (yych <= '#') goto yy674;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy674;
				goto yy46;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy46;
				if (yych <= ':') goto yy4;
				goto yy674;
			} else {
				if (yych <= '@') goto yy96;
				if (yych != 'T') goto yy46;
			}
		} else {
			if (yych <= 'z') {
				if (yych == '_') goto yy46;
				if (yych <= '`') goto yy674;
				goto yy46;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy674;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy674;
					if (yych <= 0xF4) goto yy4;
					goto yy674;
				}
			}
		}
	}
yy748:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy808;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy46;
			goto yy4;
		}
	}
yy749:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy674;
				if (yych <= 0x08) goto yy4;
				goto yy674;
			} else {
				if (yych == '\r') goto yy674;
				goto yy4;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1B) goto yy674;
				if (yych <= 0x1F) goto yy4;
				goto yy746;
			} else {
				if (yych == '$') goto yy4;
				if (yych <= ',') goto yy674;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'T') {
				if (yych <= '?') goto yy674;
				if (yych <= 'S') goto yy4;
			} else {
				if (yych <= 'Z') goto yy4;
				if (yych <= '^') goto yy674;
				goto yy4;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy674;
				if (yych <= 'z') goto yy4;
				goto yy674;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy674;
				if (yych <= 0xF4) goto yy4;
				goto yy674;
			}
		}
	}
yy750:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy809;
	goto yy4;
yy751:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy810;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy810;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy810;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy752:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy811;
					goto yy812;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy753:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy812;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych >= 'F') goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy754:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy113;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy113;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy113;
				if (yych == 0x1B) goto yy113;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy113;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy113;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy113;
				if (yych <= '-') goto yy251;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy196;
				if (yych <= ':') goto yy148;
				goto yy113;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy196;
				if (yych <= 'Z') goto yy115;
				goto yy113;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy113;
				if (yych <= 'f') goto yy196;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy113;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy113;
					if (yych <= 0xF4) goto yy4;
					goto yy113;
				}
			}
		}
	}
yy755:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy813;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy810;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy810;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy756:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '#') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '9') {
					if (yych <= '/') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					goto yy814;
				} else {
					if (yych <= ':') goto yy682;
					if (yych <= '?') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '_') {
				if (yych <= 'F') goto yy814;
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy814;
				if (yych <= 'z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy757:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy815;
				goto yy816;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy815;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy815;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy758:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= 0x1B) {
						yyt2 = YYCURSOR;
						goto yy198;
					}
					if (yych <= 0x1F) {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '/') {
					if (yych <= '.') {
						yyt2 = YYCURSOR;
						goto yy376;
					}
					yyt2 = YYCURSOR;
					goto yy284;
				} else {
					if (yych <= '9') goto yy814;
					if (yych <= ':') goto yy682;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'F') goto yy814;
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '`') {
					if (yych <= '_') {
						yyt2 = YYCURSOR;
						goto yy284;
					}
					yyt2 = YYCURSOR;
					goto yy198;
				} else {
					if (yych <= 'f') goto yy814;
					if (yych <= 'z') goto yy3;
					yyt2 = YYCURSOR;
					goto yy198;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy759:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy643;
		if (yych <= '9') goto yy542;
		goto yy643;
	} else {
		if (yych <= 'F') goto yy542;
		if (yych <= '`') goto yy643;
		if (yych <= 'f') goto yy542;
		goto yy643;
	}
yy760:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy817;
		if (yych <= ':') goto yy563;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy817;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy817;
		goto yy1;
	}
yy761:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy818;
		if (yych <= ':') goto yy691;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy818;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy818;
		goto yy1;
	}
yy762:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy819;
		if (yych <= ':') goto yy820;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy819;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy819;
		goto yy1;
	}
yy763:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy821;
			} else {
				if (yych <= '1') goto yy822;
				if (yych <= '2') goto yy823;
				goto yy821;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy824;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy824;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy764:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy825;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy825;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy825;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy765:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy824;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy824;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy824;
		goto yy1;
	}
yy766:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy825;
			if (yych <= ':') goto yy765;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy825;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy825;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy767:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy826;
				goto yy765;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy825;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy825;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy768:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '5') goto yy826;
				goto yy825;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy765;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy825;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy825;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy769:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1;
			if (yych <= ':') goto yy565;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy770:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '/') goto yy4;
		if (yych <= '0') goto yy827;
		goto yy828;
	} else {
		if (yych <= '2') goto yy829;
		if (yych <= '9') goto yy827;
		goto yy4;
	}
yy771:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy770;
	goto yy4;
yy772:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy770;
	if (yych <= '/') goto yy4;
	if (yych <= '5') goto yy771;
	goto yy4;
yy773:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy830;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy774:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy773;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy775:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '4') goto yy773;
			if (yych <= '5') goto yy831;
			goto yy830;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy776:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy576;
				goto yy486;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy576;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy576;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy777:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '0') {
			if (yych <= '/') goto yy1;
			goto yy832;
		} else {
			if (yych <= '1') goto yy473;
			if (yych <= '2') goto yy474;
			goto yy472;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy475;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy475;
			goto yy1;
		}
	}
yy778:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy833;
				goto yy486;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy576;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy576;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy779:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '5') goto yy833;
				goto yy576;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy486;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy576;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy576;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy780:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 128) {
		goto yy780;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1;
			if (yych >= '#') goto yy782;
		} else {
			if (yych <= '\\') goto yy783;
			if (yych <= 0xC1) goto yy1;
			goto yy784;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy785;
			if (yych <= 0xEF) goto yy786;
			goto yy787;
		} else {
			if (yych <= 0xF3) goto yy788;
			if (yych <= 0xF4) goto yy789;
			goto yy1;
		}
	}
yy781:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy834;
				goto yy399;
			} else {
				if (yych <= '\r') goto yy834;
				if (yych == ' ') goto yy834;
				goto yy399;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy835;
				if (yych == '\'') goto yy836;
				goto yy399;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy837;
					goto yy399;
				} else {
					if (yych <= ':') goto yy837;
					if (yych <= '=') goto yy399;
					goto yy588;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy399;
				if (yych <= 'Z') goto yy837;
				if (yych <= '[') goto yy399;
				goto yy400;
			} else {
				if (yych == '_') goto yy837;
				if (yych <= '`') goto yy399;
				goto yy837;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy399;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy401;
				goto yy402;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy403;
					goto yy404;
				} else {
					if (yych <= 0xF3) goto yy405;
					if (yych <= 0xF4) goto yy406;
					goto yy1;
				}
			}
		}
	}
yy782:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy838;
				goto yy391;
			} else {
				if (yych <= '\r') goto yy838;
				if (yych == ' ') goto yy838;
				goto yy391;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy839;
				if (yych == '\'') goto yy840;
				goto yy391;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy841;
					goto yy391;
				} else {
					if (yych <= ':') goto yy841;
					if (yych <= '=') goto yy391;
					goto yy580;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy391;
				if (yych <= 'Z') goto yy841;
				if (yych <= '[') goto yy391;
				goto yy392;
			} else {
				if (yych == '_') goto yy841;
				if (yych <= '`') goto yy391;
				goto yy841;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy391;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy393;
				goto yy394;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy395;
					goto yy396;
				} else {
					if (yych <= 0xF3) goto yy397;
					if (yych <= 0xF4) goto yy398;
					goto yy1;
				}
			}
		}
	}
yy783:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy780;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy780;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy785;
			if (yych <= 0xEF) goto yy786;
			goto yy787;
		} else {
			if (yych <= 0xF3) goto yy788;
			if (yych <= 0xF4) goto yy789;
			goto yy1;
		}
	}
yy784:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy780;
	goto yy1;
yy785:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy784;
	goto yy1;
yy786:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy784;
	goto yy1;
yy787:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy786;
	goto yy1;
yy788:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy786;
	goto yy1;
yy789:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy786;
	goto yy1;
yy790:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy790;
				goto yy717;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy790;
					goto yy717;
				} else {
					if (yych <= ' ') goto yy790;
					if (yych <= '!') goto yy717;
					goto yy413;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych == '\'') goto yy412;
				if (yych <= ',') goto yy717;
				goto yy790;
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy717;
				} else {
					if (yych <= ':') goto yy790;
					if (yych <= '<') goto yy717;
					goto yy842;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy792;
				if (yych <= '?') goto yy791;
				if (yych <= '@') goto yy717;
				goto yy790;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy717;
					goto yy793;
				} else {
					if (yych == '_') goto yy790;
					goto yy717;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy790;
					goto yy717;
				} else {
					if (yych <= 0xC1) goto yy1;
					if (yych <= 0xDF) goto yy794;
					goto yy795;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy796;
					goto yy797;
				} else {
					if (yych <= 0xF3) goto yy798;
					if (yych <= 0xF4) goto yy799;
					goto yy1;
				}
			}
		}
	}
yy791:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 64) {
		goto yy717;
	}
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '"') goto yy413;
			if (yych <= '\'') goto yy412;
			goto yy790;
		} else {
			if (yych <= ':') {
				if (yych <= '/') goto yy791;
				goto yy790;
			} else {
				if (yych <= '>') goto yy843;
				if (yych <= '?') goto yy791;
				goto yy790;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') {
				if (yych <= '\\') goto yy793;
				goto yy790;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy794;
				goto yy795;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy796;
				goto yy797;
			} else {
				if (yych <= 0xF3) goto yy798;
				if (yych <= 0xF4) goto yy799;
				goto yy1;
			}
		}
	}
yy792:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy217;
			if (yych == '"') goto yy845;
			goto yy844;
		} else {
			if (yych <= '\'') goto yy846;
			if (yych == '\\') goto yy847;
			goto yy844;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy217;
			if (yych <= 0xDF) goto yy848;
			if (yych <= 0xE0) goto yy849;
			goto yy850;
		} else {
			if (yych <= 0xF0) goto yy851;
			if (yych <= 0xF3) goto yy852;
			if (yych <= 0xF4) goto yy853;
			goto yy217;
		}
	}
yy793:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '.') {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy844;
				if (yych <= '\t') goto yy717;
				goto yy410;
			} else {
				if (yych == '-') goto yy790;
				goto yy717;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '/') goto yy791;
				if (yych <= ':') goto yy790;
				goto yy717;
			} else {
				if (yych <= '>') goto yy792;
				if (yych <= '?') goto yy791;
				if (yych <= '@') goto yy717;
				goto yy790;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') goto yy790;
				goto yy717;
			} else {
				if (yych <= 'z') goto yy790;
				if (yych <= 0x7F) goto yy717;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy794;
				if (yych <= 0xE0) goto yy795;
				goto yy796;
			} else {
				if (yych <= 0xF0) goto yy797;
				if (yych <= 0xF3) goto yy798;
				if (yych <= 0xF4) goto yy799;
				goto yy1;
			}
		}
	}
yy794:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy717;
	goto yy1;
yy795:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy794;
	goto yy1;
yy796:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy794;
	goto yy1;
yy797:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy796;
	goto yy1;
yy798:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy796;
	goto yy1;
yy799:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy796;
	goto yy1;
yy800:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy1;
				if (yych <= '!') goto yy800;
				goto yy525;
			} else {
				if (yych == '\'') goto yy524;
				if (yych <= ',') goto yy800;
				goto yy854;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '.') goto yy800;
				if (yych <= '/') goto yy855;
				if (yych <= ':') goto yy854;
				goto yy800;
			} else {
				if (yych <= '>') goto yy856;
				if (yych <= '?') goto yy855;
				if (yych <= '@') goto yy800;
				goto yy854;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych == '\\') goto yy857;
				goto yy800;
			} else {
				if (yych == '`') goto yy800;
				if (yych <= 'z') goto yy854;
				goto yy800;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy858;
				if (yych <= 0xE0) goto yy859;
				goto yy860;
			} else {
				if (yych <= 0xF0) goto yy861;
				if (yych <= 0xF3) goto yy862;
				if (yych <= 0xF4) goto yy863;
				goto yy1;
			}
		}
	}
yy801:
	yych = *++YYCURSOR;
	if (yych <= 0x1F) {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1;
			goto yy801;
		} else {
			if (yych == '\r') goto yy801;
			goto yy1;
		}
	} else {
		if (yych <= '/') {
			if (yych <= ' ') goto yy801;
			if (yych <= '.') goto yy1;
			goto yy215;
		} else {
			if (yych == '?') goto yy215;
			goto yy1;
		}
	}
yy802:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy864;
	goto yy1;
yy803:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy865;
	goto yy1;
yy804:
	yych = *++YYCURSOR;
	if (yych == 'i') goto yy866;
	goto yy1;
yy805:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy749;
	goto yy4;
yy806:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy867;
	goto yy1;
yy807:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy868;
	goto yy1;
yy808:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '.') goto yy46;
			if (yych <= '/') goto yy4;
			goto yy869;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych == '`') goto yy4;
			if (yych <= 'z') goto yy46;
			goto yy4;
		}
	}
yy809:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy4;
	if (yych <= '9') goto yy870;
	goto yy4;
yy810:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy871;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy871;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy871;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy811:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy872;
					goto yy873;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy812:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy873;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy813:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy26;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy26;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy874;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy871;
				if (yych <= 'Z') goto yy46;
				if (yych <= '^') goto yy26;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy26;
				if (yych <= 'f') goto yy871;
				if (yych <= 'z') goto yy46;
				goto yy26;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy26;
				if (yych <= 0xF4) goto yy4;
				goto yy26;
			}
		}
	}
yy814:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '#') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy680;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '9') goto yy3;
				if (yych <= ':') goto yy682;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'z') goto yy3;
				if (yych <= '~') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy290;
				}
				yyt2 = YYCURSOR;
				goto yy291;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy292;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy293;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy294;
				}
				goto yy5;
			}
		}
	}
yy815:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy875;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy875;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy875;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy816:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy876;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy876;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy876;
		goto yy1;
	}
yy817:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy877;
		if (yych <= ':') goto yy691;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy877;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy877;
		goto yy1;
	}
yy818:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy691;
	goto yy1;
yy819:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy878;
		if (yych >= ';') goto yy1;
	} else {
		if (yych <= 'F') goto yy878;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy878;
		goto yy1;
	}
yy820:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy879;
		if (yych <= ':') goto yy880;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy879;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy879;
		goto yy1;
	}
yy821:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy881;
				goto yy882;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy883;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy883;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy822:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy884;
				goto yy882;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy883;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy883;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy823:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '4') goto yy884;
				goto yy885;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy881;
				if (yych <= ':') goto yy882;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy883;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy883;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy824:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy883;
			if (yych <= ':') goto yy882;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy883;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy883;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy825:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy886;
			if (yych <= ':') goto yy765;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy886;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy886;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy826:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy886;
				goto yy765;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy886;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy886;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy827:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych == '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x1A) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= ',') {
				if (yych <= 0x1F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '9') goto yy887;
				if (yych <= ':') goto yy62;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'z') goto yy3;
				if (yych <= '~') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy290;
				}
				yyt2 = YYCURSOR;
				goto yy291;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy292;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy293;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy294;
				}
				goto yy5;
			}
		}
	}
yy828:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych == '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x1A) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= ',') {
				if (yych <= 0x1F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '9') goto yy827;
				if (yych <= ':') goto yy62;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'Z') goto yy3;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'z') goto yy3;
				if (yych <= '~') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy290;
				}
				yyt2 = YYCURSOR;
				goto yy291;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy292;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy293;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy294;
				}
				goto yy5;
			}
		}
	}
yy829:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych == 0x1B) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= '/') {
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '4') goto yy827;
				if (yych <= '5') goto yy888;
				if (yych <= '9') goto yy887;
				goto yy62;
			}
		}
	} else {
		if (yych <= '~') {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'Z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'z') goto yy3;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 0xC1) goto yy5;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				yyt2 = YYCURSOR;
				goto yy290;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) {
						yyt2 = YYCURSOR;
						goto yy291;
					}
					yyt2 = YYCURSOR;
					goto yy292;
				} else {
					if (yych <= 0xF3) {
						yyt2 = YYCURSOR;
						goto yy293;
					}
					if (yych <= 0xF4) {
						yyt2 = YYCURSOR;
						goto yy294;
					}
					goto yy5;
				}
			}
		}
	}
yy830:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy831:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '5') goto yy830;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy832:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy564;
				goto yy889;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy566;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy566;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy833:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy704;
				goto yy486;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy704;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy704;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy834:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy399;
				goto yy834;
			} else {
				if (yych == '\r') goto yy834;
				goto yy399;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy834;
				if (yych <= '!') goto yy399;
			} else {
				if (yych == '\'') goto yy836;
				goto yy399;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '>') goto yy588;
				if (yych <= '[') goto yy399;
				goto yy400;
			} else {
				if (yych <= 0x7F) goto yy399;
				if (yych <= 0xC1) goto yy1;
				goto yy401;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy402;
				if (yych <= 0xEF) goto yy403;
				goto yy404;
			} else {
				if (yych <= 0xF3) goto yy405;
				if (yych <= 0xF4) goto yy406;
				goto yy1;
			}
		}
	}
yy835:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 128) {
		goto yy780;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '"') goto yy399;
			goto yy782;
		} else {
			if (yych <= '\\') goto yy783;
			if (yych <= 0xC1) goto yy1;
			goto yy784;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy785;
			if (yych <= 0xEF) goto yy786;
			goto yy787;
		} else {
			if (yych <= 0xF3) goto yy788;
			if (yych <= 0xF4) goto yy789;
			goto yy1;
		}
	}
yy836:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy834;
				goto yy399;
			} else {
				if (yych <= '\r') goto yy834;
				if (yych == ' ') goto yy834;
				goto yy399;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy835;
				if (yych == '\'') goto yy307;
				goto yy399;
			} else {
				if (yych <= '/') {
					if (yych >= '.') goto yy399;
				} else {
					if (yych <= ':') goto yy837;
					if (yych <= '=') goto yy399;
					goto yy588;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy399;
				if (yych <= 'Z') goto yy837;
				if (yych <= '[') goto yy399;
				goto yy400;
			} else {
				if (yych == '_') goto yy837;
				if (yych <= '`') goto yy399;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy399;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy401;
				goto yy402;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy403;
					goto yy404;
				} else {
					if (yych <= 0xF3) goto yy405;
					if (yych <= 0xF4) goto yy406;
					goto yy1;
				}
			}
		}
	}
yy837:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy890;
				goto yy399;
			} else {
				if (yych <= '\r') goto yy890;
				if (yych == ' ') goto yy890;
				goto yy399;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy835;
				if (yych == '\'') goto yy836;
				goto yy399;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy837;
					goto yy399;
				} else {
					if (yych <= ':') goto yy837;
					if (yych <= '<') goto yy399;
					goto yy891;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy588;
				if (yych <= '@') goto yy399;
				if (yych <= 'Z') goto yy837;
				goto yy399;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy400;
					goto yy399;
				} else {
					if (yych == '`') goto yy399;
					goto yy837;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy399;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy401;
				goto yy402;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy403;
					goto yy404;
				} else {
					if (yych <= 0xF3) goto yy405;
					if (yych <= 0xF4) goto yy406;
					goto yy1;
				}
			}
		}
	}
yy838:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy391;
				goto yy838;
			} else {
				if (yych == '\r') goto yy838;
				goto yy391;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy838;
				if (yych <= '!') goto yy391;
			} else {
				if (yych == '\'') goto yy840;
				goto yy391;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '>') goto yy580;
				if (yych <= '[') goto yy391;
				goto yy392;
			} else {
				if (yych <= 0x7F) goto yy391;
				if (yych <= 0xC1) goto yy1;
				goto yy393;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy394;
				if (yych <= 0xEF) goto yy395;
				goto yy396;
			} else {
				if (yych <= 0xF3) goto yy397;
				if (yych <= 0xF4) goto yy398;
				goto yy1;
			}
		}
	}
yy839:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy838;
				goto yy391;
			} else {
				if (yych <= '\r') goto yy838;
				if (yych == ' ') goto yy838;
				goto yy391;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy306;
				if (yych != '\'') goto yy391;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy841;
					goto yy391;
				} else {
					if (yych <= ':') goto yy841;
					if (yych <= '=') goto yy391;
					goto yy580;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy391;
				if (yych <= 'Z') goto yy841;
				if (yych <= '[') goto yy391;
				goto yy392;
			} else {
				if (yych == '_') goto yy841;
				if (yych <= '`') goto yy391;
				goto yy841;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy391;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy393;
				goto yy394;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy395;
					goto yy396;
				} else {
					if (yych <= 0xF3) goto yy397;
					if (yych <= 0xF4) goto yy398;
					goto yy1;
				}
			}
		}
	}
yy840:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 128) {
		goto yy780;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '"') goto yy781;
			goto yy391;
		} else {
			if (yych <= '\\') goto yy783;
			if (yych <= 0xC1) goto yy1;
			goto yy784;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy785;
			if (yych <= 0xEF) goto yy786;
			goto yy787;
		} else {
			if (yych <= 0xF3) goto yy788;
			if (yych <= 0xF4) goto yy789;
			goto yy1;
		}
	}
yy841:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy892;
				goto yy391;
			} else {
				if (yych <= '\r') goto yy892;
				if (yych == ' ') goto yy892;
				goto yy391;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy839;
				if (yych == '\'') goto yy840;
				goto yy391;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy841;
					goto yy391;
				} else {
					if (yych <= ':') goto yy841;
					if (yych <= '<') goto yy391;
					goto yy893;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy580;
				if (yych <= '@') goto yy391;
				if (yych <= 'Z') goto yy841;
				goto yy391;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy392;
					goto yy391;
				} else {
					if (yych == '`') goto yy391;
					goto yy841;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy391;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy393;
				goto yy394;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy395;
					goto yy396;
				} else {
					if (yych <= 0xF3) goto yy397;
					if (yych <= 0xF4) goto yy398;
					goto yy1;
				}
			}
		}
	}
yy842:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= ',') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy842;
				goto yy717;
			} else {
				if (yych <= '\r') goto yy842;
				if (yych == ' ') goto yy842;
				goto yy717;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '-') goto yy790;
				if (yych <= '.') goto yy717;
				if (yych <= '/') goto yy791;
				goto yy790;
			} else {
				if (yych <= '=') goto yy717;
				if (yych <= '>') goto yy792;
				if (yych <= '?') goto yy791;
				goto yy717;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= 'Z') goto yy790;
				if (yych == '\\') goto yy793;
				goto yy717;
			} else {
				if (yych == '`') goto yy717;
				if (yych <= 'z') goto yy790;
				goto yy717;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy794;
				if (yych <= 0xE0) goto yy795;
				goto yy796;
			} else {
				if (yych <= 0xF0) goto yy797;
				if (yych <= 0xF3) goto yy798;
				if (yych <= 0xF4) goto yy799;
				goto yy1;
			}
		}
	}
yy843:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy315;
			if (yych == '"') goto yy845;
		} else {
			if (yych <= '\'') goto yy846;
			if (yych == '\\') goto yy847;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy315;
			if (yych <= 0xDF) goto yy848;
			if (yych <= 0xE0) goto yy849;
			goto yy850;
		} else {
			if (yych <= 0xF0) goto yy851;
			if (yych <= 0xF3) goto yy852;
			if (yych <= 0xF4) goto yy853;
			goto yy315;
		}
	}
yy844:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy1;
			if (yych != '"') goto yy844;
		} else {
			if (yych <= '\'') goto yy846;
			if (yych == '\\') goto yy847;
			goto yy844;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy848;
			if (yych <= 0xE0) goto yy849;
			goto yy850;
		} else {
			if (yych <= 0xF0) goto yy851;
			if (yych <= 0xF3) goto yy852;
			if (yych <= 0xF4) goto yy853;
			goto yy1;
		}
	}
yy845:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy894;
				goto yy609;
			} else {
				if (yych <= '\r') goto yy894;
				if (yych == ' ') goto yy894;
				goto yy609;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy599;
				if (yych == '-') goto yy895;
				goto yy609;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy896;
					goto yy895;
				} else {
					if (yych <= '=') goto yy609;
					if (yych <= '>') goto yy513;
					goto yy896;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy609;
				if (yych <= 'Z') goto yy895;
				if (yych <= '[') goto yy609;
				goto yy610;
			} else {
				if (yych == '_') goto yy895;
				if (yych <= '`') goto yy609;
				goto yy895;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy609;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy611;
				goto yy612;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy613;
					goto yy614;
				} else {
					if (yych <= 0xF3) goto yy615;
					if (yych <= 0xF4) goto yy616;
					goto yy1;
				}
			}
		}
	}
yy846:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy897;
				goto yy598;
			} else {
				if (yych <= '\r') goto yy897;
				if (yych == ' ') goto yy897;
				goto yy598;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy599;
				if (yych == '-') goto yy898;
				goto yy598;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy899;
					goto yy898;
				} else {
					if (yych <= '=') goto yy598;
					if (yych <= '>') goto yy503;
					goto yy899;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy598;
				if (yych <= 'Z') goto yy898;
				if (yych <= '[') goto yy598;
				goto yy600;
			} else {
				if (yych == '_') goto yy898;
				if (yych <= '`') goto yy598;
				goto yy898;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy598;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy601;
				goto yy602;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy603;
					goto yy604;
				} else {
					if (yych <= 0xF3) goto yy605;
					if (yych <= 0xF4) goto yy606;
					goto yy1;
				}
			}
		}
	}
yy847:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy844;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy844;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy849;
			if (yych <= 0xEF) goto yy850;
			goto yy851;
		} else {
			if (yych <= 0xF3) goto yy852;
			if (yych <= 0xF4) goto yy853;
			goto yy1;
		}
	}
yy848:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy844;
	goto yy1;
yy849:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy848;
	goto yy1;
yy850:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy848;
	goto yy1;
yy851:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy850;
	goto yy1;
yy852:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy850;
	goto yy1;
yy853:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy850;
	goto yy1;
yy854:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy854;
				goto yy800;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy854;
					goto yy800;
				} else {
					if (yych <= ' ') goto yy854;
					if (yych <= '!') goto yy800;
					goto yy525;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych == '\'') goto yy524;
				if (yych <= ',') goto yy800;
				goto yy854;
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy800;
				} else {
					if (yych <= ':') goto yy854;
					if (yych <= '<') goto yy800;
					goto yy900;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy856;
				if (yych <= '?') goto yy855;
				if (yych <= '@') goto yy800;
				goto yy854;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy800;
					goto yy857;
				} else {
					if (yych == '_') goto yy854;
					goto yy800;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy854;
					goto yy800;
				} else {
					if (yych <= 0xC1) goto yy1;
					if (yych <= 0xDF) goto yy858;
					goto yy859;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy860;
					goto yy861;
				} else {
					if (yych <= 0xF3) goto yy862;
					if (yych <= 0xF4) goto yy863;
					goto yy1;
				}
			}
		}
	}
yy855:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy1;
				if (yych <= '!') goto yy800;
				goto yy525;
			} else {
				if (yych == '\'') goto yy524;
				if (yych <= ',') goto yy800;
				goto yy854;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '.') goto yy800;
				if (yych <= '/') goto yy855;
				if (yych <= ':') goto yy854;
				goto yy800;
			} else {
				if (yych <= '>') goto yy901;
				if (yych <= '?') goto yy855;
				if (yych <= '@') goto yy800;
				goto yy854;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych == '\\') goto yy857;
				goto yy800;
			} else {
				if (yych == '`') goto yy800;
				if (yych <= 'z') goto yy854;
				goto yy800;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy858;
				if (yych <= 0xE0) goto yy859;
				goto yy860;
			} else {
				if (yych <= 0xF0) goto yy861;
				if (yych <= 0xF3) goto yy862;
				if (yych <= 0xF4) goto yy863;
				goto yy1;
			}
		}
	}
yy856:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy1;
			if (yych == '"') goto yy902;
			goto yy856;
		} else {
			if (yych <= '\'') goto yy903;
			if (yych == '\\') goto yy904;
			goto yy856;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy905;
			if (yych <= 0xE0) goto yy906;
			goto yy907;
		} else {
			if (yych <= 0xF0) goto yy908;
			if (yych <= 0xF3) goto yy909;
			if (yych <= 0xF4) goto yy910;
			goto yy1;
		}
	}
yy857:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '.') {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy856;
				if (yych <= '\t') goto yy800;
				goto yy522;
			} else {
				if (yych == '-') goto yy854;
				goto yy800;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '/') goto yy855;
				if (yych <= ':') goto yy854;
				goto yy800;
			} else {
				if (yych <= '>') goto yy856;
				if (yych <= '?') goto yy855;
				if (yych <= '@') goto yy800;
				goto yy854;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') goto yy854;
				goto yy800;
			} else {
				if (yych <= 'z') goto yy854;
				if (yych <= 0x7F) goto yy800;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy858;
				if (yych <= 0xE0) goto yy859;
				goto yy860;
			} else {
				if (yych <= 0xF0) goto yy861;
				if (yych <= 0xF3) goto yy862;
				if (yych <= 0xF4) goto yy863;
				goto yy1;
			}
		}
	}
yy858:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy800;
	goto yy1;
yy859:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy858;
	goto yy1;
yy860:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy858;
	goto yy1;
yy861:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy860;
	goto yy1;
yy862:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy860;
	goto yy1;
yy863:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy860;
	goto yy1;
yy864:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy911;
	goto yy1;
yy865:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy912;
			}
			goto yy1;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= 0xC1) goto yy1;
			yyt2 = YYCURSOR;
			goto yy645;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy646;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy647;
			}
			yyt2 = YYCURSOR;
			goto yy648;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy649;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy650;
			}
			goto yy1;
		}
	}
yy866:
	yych = *++YYCURSOR;
	if (yych == 'l') goto yy913;
	goto yy1;
yy867:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy914;
	goto yy1;
yy868:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy915;
	goto yy1;
yy869:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[1024+yych] & 16) {
		goto yy46;
	}
	if (yych <= '+') {
		if (yych == '%') goto yy80;
		if (yych <= '*') goto yy4;
		goto yy80;
	} else {
		if (yych <= ':') {
			if (yych <= '/') goto yy4;
			goto yy916;
		} else {
			if (yych == '@') goto yy96;
			goto yy4;
		}
	}
yy870:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy916;
	goto yy4;
yy871:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy917;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy917;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy917;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy872:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy918;
					goto yy919;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy873:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy919;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy874:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy26;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy26;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy920;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy917;
				if (yych <= 'Z') goto yy46;
				if (yych <= '^') goto yy26;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy26;
				if (yych <= 'f') goto yy917;
				if (yych <= 'z') goto yy46;
				goto yy26;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy26;
				if (yych <= 0xF4) goto yy4;
				goto yy26;
			}
		}
	}
yy875:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy921;
				goto yy816;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy921;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy921;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy876:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy922;
				goto yy923;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy922;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy922;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy877:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy354;
			goto yy466;
		} else {
			if (yych <= '/') goto yy354;
			if (yych <= '9') goto yy761;
			goto yy924;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy354;
			goto yy761;
		} else {
			if (yych <= '`') goto yy354;
			if (yych <= 'f') goto yy761;
			goto yy354;
		}
	}
yy878:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy925;
		if (yych <= ':') goto yy820;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy925;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy925;
		goto yy1;
	}
yy879:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy926;
		if (yych <= ':') goto yy927;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy926;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy926;
		goto yy1;
	}
yy880:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy928;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy928;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy928;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy881:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy929;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy929;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy929;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy882:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy928;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy928;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy928;
		goto yy1;
	}
yy883:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy929;
			if (yych <= ':') goto yy882;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy929;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy929;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy884:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy930;
				goto yy882;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy929;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy929;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy885:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '5') goto yy930;
				goto yy929;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy882;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy929;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy929;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy886:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1;
			if (yych <= ':') goto yy765;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy887:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych == '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x1A) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= ',') {
				if (yych <= 0x1F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= ':') goto yy4;
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy4;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'z') goto yy4;
				if (yych <= '~') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy4;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy290;
				}
				yyt2 = YYCURSOR;
				goto yy291;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy292;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy293;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy294;
				}
				goto yy4;
			}
		}
	}
yy888:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x08) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych == '\r') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0x1A) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= ',') {
				if (yych <= 0x1F) {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych == '$') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= '5') goto yy887;
				if (yych <= ':') goto yy4;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy284;
				}
				if (yych <= 'Z') goto yy4;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'z') goto yy4;
				if (yych <= '~') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy284;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy4;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy289;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy290;
				}
				yyt2 = YYCURSOR;
				goto yy291;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy292;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy293;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy294;
				}
				goto yy4;
			}
		}
	}
yy889:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '0') {
			if (yych <= '/') goto yy1;
			goto yy931;
		} else {
			if (yych <= '1') goto yy693;
			if (yych <= '2') goto yy694;
			goto yy692;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy695;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy695;
			goto yy1;
		}
	}
yy890:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy399;
				goto yy890;
			} else {
				if (yych == '\r') goto yy890;
				goto yy399;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy890;
				if (yych <= '!') goto yy399;
				goto yy835;
			} else {
				if (yych == '\'') goto yy836;
				goto yy399;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '=') goto yy891;
				if (yych <= '>') goto yy588;
				goto yy399;
			} else {
				if (yych <= '\\') goto yy400;
				if (yych <= 0x7F) goto yy399;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy401;
				if (yych <= 0xE0) goto yy402;
				goto yy403;
			} else {
				if (yych <= 0xF0) goto yy404;
				if (yych <= 0xF3) goto yy405;
				if (yych <= 0xF4) goto yy406;
				goto yy1;
			}
		}
	}
yy891:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 4) {
		goto yy492;
	}
	if (yych <= 0xDF) {
		if (yych <= '>') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '"') goto yy579;
			goto yy399;
		} else {
			if (yych <= '\\') goto yy589;
			if (yych <= 0xC1) goto yy1;
			goto yy590;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy591;
			if (yych <= 0xEF) goto yy592;
			goto yy593;
		} else {
			if (yych <= 0xF3) goto yy594;
			if (yych <= 0xF4) goto yy595;
			goto yy1;
		}
	}
yy892:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy391;
				goto yy892;
			} else {
				if (yych == '\r') goto yy892;
				goto yy391;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy892;
				if (yych <= '!') goto yy391;
				goto yy839;
			} else {
				if (yych == '\'') goto yy840;
				goto yy391;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '=') goto yy893;
				if (yych <= '>') goto yy580;
				goto yy391;
			} else {
				if (yych <= '\\') goto yy392;
				if (yych <= 0x7F) goto yy391;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy393;
				if (yych <= 0xE0) goto yy394;
				goto yy395;
			} else {
				if (yych <= 0xF0) goto yy396;
				if (yych <= 0xF3) goto yy397;
				if (yych <= 0xF4) goto yy398;
				goto yy1;
			}
		}
	}
yy893:
	yych = *++YYCURSOR;
	if (yybm[0+yych] & 2) {
		goto yy491;
	}
	if (yych <= 0xDF) {
		if (yych <= '>') {
			if (yych <= 0x00) goto yy1;
			if (yych <= '\'') goto yy579;
			goto yy391;
		} else {
			if (yych <= '\\') goto yy581;
			if (yych <= 0xC1) goto yy1;
			goto yy582;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy583;
			if (yych <= 0xEF) goto yy584;
			goto yy585;
		} else {
			if (yych <= 0xF3) goto yy586;
			if (yych <= 0xF4) goto yy587;
			goto yy1;
		}
	}
yy894:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy609;
				goto yy894;
			} else {
				if (yych == '\r') goto yy894;
				goto yy609;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy894;
				if (yych <= '&') goto yy609;
				goto yy599;
			} else {
				if (yych == '/') goto yy896;
				goto yy609;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy513;
				if (yych <= '?') goto yy896;
				goto yy609;
			} else {
				if (yych <= '\\') goto yy610;
				if (yych <= 0x7F) goto yy609;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy611;
				if (yych <= 0xE0) goto yy612;
				goto yy613;
			} else {
				if (yych <= 0xF0) goto yy614;
				if (yych <= 0xF3) goto yy615;
				if (yych <= 0xF4) goto yy616;
				goto yy1;
			}
		}
	}
yy895:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy932;
				goto yy609;
			} else {
				if (yych <= '\r') goto yy932;
				if (yych == ' ') goto yy932;
				goto yy609;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy599;
				if (yych == '-') goto yy895;
				goto yy609;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy895;
				} else {
					if (yych <= '<') goto yy609;
					if (yych <= '=') goto yy933;
					goto yy513;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy896;
				if (yych <= '@') goto yy609;
				if (yych <= 'Z') goto yy895;
				goto yy609;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy610;
					goto yy609;
				} else {
					if (yych == '`') goto yy609;
					goto yy895;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy609;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy611;
				goto yy612;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy613;
					goto yy614;
				} else {
					if (yych <= 0xF3) goto yy615;
					if (yych <= 0xF4) goto yy616;
					goto yy1;
				}
			}
		}
	}
yy896:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1;
			if (yych == '\'') goto yy599;
			goto yy609;
		} else {
			if (yych <= '>') goto yy608;
			if (yych == '\\') goto yy610;
			goto yy609;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy611;
			if (yych <= 0xE0) goto yy612;
			goto yy613;
		} else {
			if (yych <= 0xF0) goto yy614;
			if (yych <= 0xF3) goto yy615;
			if (yych <= 0xF4) goto yy616;
			goto yy1;
		}
	}
yy897:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy598;
				goto yy897;
			} else {
				if (yych == '\r') goto yy897;
				goto yy598;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy897;
				if (yych <= '!') goto yy598;
				goto yy599;
			} else {
				if (yych == '/') goto yy899;
				goto yy598;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy503;
				if (yych <= '?') goto yy899;
				goto yy598;
			} else {
				if (yych <= '\\') goto yy600;
				if (yych <= 0x7F) goto yy598;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy601;
				if (yych <= 0xE0) goto yy602;
				goto yy603;
			} else {
				if (yych <= 0xF0) goto yy604;
				if (yych <= 0xF3) goto yy605;
				if (yych <= 0xF4) goto yy606;
				goto yy1;
			}
		}
	}
yy898:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy934;
				goto yy598;
			} else {
				if (yych <= '\r') goto yy934;
				if (yych == ' ') goto yy934;
				goto yy598;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy599;
				if (yych == '-') goto yy898;
				goto yy598;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy898;
				} else {
					if (yych <= '<') goto yy598;
					if (yych <= '=') goto yy935;
					goto yy503;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy899;
				if (yych <= '@') goto yy598;
				if (yych <= 'Z') goto yy898;
				goto yy598;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy600;
					goto yy598;
				} else {
					if (yych == '`') goto yy598;
					goto yy898;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy598;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy601;
				goto yy602;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy603;
					goto yy604;
				} else {
					if (yych <= 0xF3) goto yy605;
					if (yych <= 0xF4) goto yy606;
					goto yy1;
				}
			}
		}
	}
yy899:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1;
			if (yych == '"') goto yy599;
			goto yy598;
		} else {
			if (yych <= '>') goto yy597;
			if (yych == '\\') goto yy600;
			goto yy598;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy601;
			if (yych <= 0xE0) goto yy602;
			goto yy603;
		} else {
			if (yych <= 0xF0) goto yy604;
			if (yych <= 0xF3) goto yy605;
			if (yych <= 0xF4) goto yy606;
			goto yy1;
		}
	}
yy900:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= ',') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy900;
				goto yy800;
			} else {
				if (yych <= '\r') goto yy900;
				if (yych == ' ') goto yy900;
				goto yy800;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '-') goto yy854;
				if (yych <= '.') goto yy800;
				if (yych <= '/') goto yy855;
				goto yy854;
			} else {
				if (yych <= '=') goto yy800;
				if (yych <= '>') goto yy856;
				if (yych <= '?') goto yy855;
				goto yy800;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= 'Z') goto yy854;
				if (yych == '\\') goto yy857;
				goto yy800;
			} else {
				if (yych == '`') goto yy800;
				if (yych <= 'z') goto yy854;
				goto yy800;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy858;
				if (yych <= 0xE0) goto yy859;
				goto yy860;
			} else {
				if (yych <= 0xF0) goto yy861;
				if (yych <= 0xF3) goto yy862;
				if (yych <= 0xF4) goto yy863;
				goto yy1;
			}
		}
	}
yy901:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy315;
			if (yych != '"') goto yy856;
		} else {
			if (yych <= '\'') goto yy903;
			if (yych == '\\') goto yy904;
			goto yy856;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy315;
			if (yych <= 0xDF) goto yy905;
			if (yych <= 0xE0) goto yy906;
			goto yy907;
		} else {
			if (yych <= 0xF0) goto yy908;
			if (yych <= 0xF3) goto yy909;
			if (yych <= 0xF4) goto yy910;
			goto yy315;
		}
	}
yy902:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy936;
				goto yy631;
			} else {
				if (yych <= '\r') goto yy936;
				if (yych == ' ') goto yy936;
				goto yy631;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy721;
				if (yych == '-') goto yy937;
				goto yy631;
			} else {
				if (yych <= '/') goto yy938;
				if (yych <= ':') goto yy937;
				if (yych <= '>') goto yy631;
				goto yy938;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy631;
				if (yych <= 'Z') goto yy937;
				if (yych <= '[') goto yy631;
				goto yy731;
			} else {
				if (yych == '_') goto yy937;
				if (yych <= '`') goto yy631;
				goto yy937;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy631;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy732;
				goto yy733;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy734;
					goto yy735;
				} else {
					if (yych <= 0xF3) goto yy736;
					if (yych <= 0xF4) goto yy737;
					goto yy1;
				}
			}
		}
	}
yy903:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy939;
				goto yy621;
			} else {
				if (yych <= '\r') goto yy939;
				if (yych == ' ') goto yy939;
				goto yy621;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy721;
				if (yych == '-') goto yy940;
				goto yy621;
			} else {
				if (yych <= '/') goto yy941;
				if (yych <= ':') goto yy940;
				if (yych <= '>') goto yy621;
				goto yy941;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy621;
				if (yych <= 'Z') goto yy940;
				if (yych <= '[') goto yy621;
				goto yy722;
			} else {
				if (yych == '_') goto yy940;
				if (yych <= '`') goto yy621;
				goto yy940;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy621;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy723;
				goto yy724;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy725;
					goto yy726;
				} else {
					if (yych <= 0xF3) goto yy727;
					if (yych <= 0xF4) goto yy728;
					goto yy1;
				}
			}
		}
	}
yy904:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy856;
			goto yy1;
		} else {
			if (yych <= 0x7F) goto yy856;
			if (yych <= 0xC1) goto yy1;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy906;
			if (yych <= 0xEF) goto yy907;
			goto yy908;
		} else {
			if (yych <= 0xF3) goto yy909;
			if (yych <= 0xF4) goto yy910;
			goto yy1;
		}
	}
yy905:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy856;
	goto yy1;
yy906:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy905;
	goto yy1;
yy907:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy905;
	goto yy1;
yy908:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy907;
	goto yy1;
yy909:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy907;
	goto yy1;
yy910:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy907;
	goto yy1;
yy911:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy942;
	goto yy1;
yy912:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy944;
			}
			goto yy643;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= 0xC1) goto yy643;
			yyt2 = YYCURSOR;
			goto yy645;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy646;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy647;
			}
			yyt2 = YYCURSOR;
			goto yy648;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy649;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy650;
			}
			goto yy643;
		}
	}
yy913:
	yych = *++YYCURSOR;
	if (yych == 'e') goto yy945;
	goto yy1;
yy914:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy946;
	goto yy1;
yy915:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy947;
	goto yy1;
yy916:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy947;
	if (yych <= ':') goto yy152;
	goto yy1;
yy917:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy948;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy948;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy948;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy918:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy949;
					goto yy950;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy919:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy950;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy920:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy26;
				if (yych == 0x1B) goto yy26;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy26;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy26;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy951;
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy26;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy948;
				if (yych <= 'Z') goto yy46;
				if (yych <= '^') goto yy26;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy26;
				if (yych <= 'f') goto yy948;
				if (yych <= 'z') goto yy46;
				goto yy26;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy26;
				if (yych <= 0xF4) goto yy4;
				goto yy26;
			}
		}
	}
yy921:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy680;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			goto yy1;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy816;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy922:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy952;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy952;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy952;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy923:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy953;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy953;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy953;
		goto yy1;
	}
yy924:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy954;
		if (yych <= ':') goto yy763;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy954;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy954;
		goto yy1;
	}
yy925:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy820;
	goto yy1;
yy926:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy955;
		if (yych >= ';') goto yy1;
	} else {
		if (yych <= 'F') goto yy955;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy955;
		goto yy1;
	}
yy927:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy956;
		if (yych <= ':') goto yy957;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy956;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy956;
		goto yy1;
	}
yy928:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy958;
			if (yych <= ':') goto yy959;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy958;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy958;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy929:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy960;
			if (yych <= ':') goto yy882;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy960;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy960;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy930:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy960;
				goto yy882;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy960;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy960;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy931:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy764;
				goto yy961;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy766;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy766;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy932:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy609;
				goto yy932;
			} else {
				if (yych == '\r') goto yy932;
				goto yy609;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy932;
				if (yych <= '&') goto yy609;
				goto yy599;
			} else {
				if (yych == '/') goto yy896;
				if (yych <= '<') goto yy609;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy513;
				if (yych <= '?') goto yy896;
				goto yy609;
			} else {
				if (yych <= '\\') goto yy610;
				if (yych <= 0x7F) goto yy609;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy611;
				if (yych <= 0xE0) goto yy612;
				goto yy613;
			} else {
				if (yych <= 0xF0) goto yy614;
				if (yych <= 0xF3) goto yy615;
				if (yych <= 0xF4) goto yy616;
				goto yy1;
			}
		}
	}
yy933:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1;
				goto yy413;
			} else {
				if (yych <= '\t') goto yy607;
				if (yych <= '\f') goto yy413;
				goto yy607;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy607;
				goto yy413;
			} else {
				if (yych <= '"') goto yy717;
				if (yych <= '=') goto yy413;
				goto yy609;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy413;
				goto yy514;
			} else {
				if (yych <= 0x7F) goto yy413;
				if (yych <= 0xC1) goto yy1;
				goto yy515;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy516;
				if (yych <= 0xEF) goto yy517;
				goto yy518;
			} else {
				if (yych <= 0xF3) goto yy519;
				if (yych <= 0xF4) goto yy520;
				goto yy1;
			}
		}
	}
yy934:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy598;
				goto yy934;
			} else {
				if (yych == '\r') goto yy934;
				goto yy598;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy934;
				if (yych <= '!') goto yy598;
				goto yy599;
			} else {
				if (yych == '/') goto yy899;
				if (yych <= '<') goto yy598;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy503;
				if (yych <= '?') goto yy899;
				goto yy598;
			} else {
				if (yych <= '\\') goto yy600;
				if (yych <= 0x7F) goto yy598;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy601;
				if (yych <= 0xE0) goto yy602;
				goto yy603;
			} else {
				if (yych <= 0xF0) goto yy604;
				if (yych <= 0xF3) goto yy605;
				if (yych <= 0xF4) goto yy606;
				goto yy1;
			}
		}
	}
yy935:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1;
				goto yy412;
			} else {
				if (yych <= '\t') goto yy596;
				if (yych <= '\f') goto yy412;
				goto yy596;
			}
		} else {
			if (yych <= '&') {
				if (yych == ' ') goto yy596;
				goto yy412;
			} else {
				if (yych <= '\'') goto yy717;
				if (yych <= '=') goto yy412;
				goto yy598;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy412;
				goto yy504;
			} else {
				if (yych <= 0x7F) goto yy412;
				if (yych <= 0xC1) goto yy1;
				goto yy505;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy506;
				if (yych <= 0xEF) goto yy507;
				goto yy508;
			} else {
				if (yych <= 0xF3) goto yy509;
				if (yych <= 0xF4) goto yy510;
				goto yy1;
			}
		}
	}
yy936:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy631;
				goto yy936;
			} else {
				if (yych == '\r') goto yy936;
				goto yy631;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy936;
				if (yych <= '&') goto yy631;
				goto yy721;
			} else {
				if (yych == '/') goto yy938;
				goto yy631;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '?') goto yy938;
				if (yych <= '[') goto yy631;
				goto yy731;
			} else {
				if (yych <= 0x7F) goto yy631;
				if (yych <= 0xC1) goto yy1;
				goto yy732;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy733;
				if (yych <= 0xEF) goto yy734;
				goto yy735;
			} else {
				if (yych <= 0xF3) goto yy736;
				if (yych <= 0xF4) goto yy737;
				goto yy1;
			}
		}
	}
yy937:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy962;
				goto yy631;
			} else {
				if (yych <= '\r') goto yy962;
				if (yych == ' ') goto yy962;
				goto yy631;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy721;
				if (yych == '-') goto yy937;
				goto yy631;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy937;
				} else {
					if (yych == '=') goto yy963;
					goto yy631;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy938;
				if (yych <= '@') goto yy631;
				if (yych <= 'Z') goto yy937;
				goto yy631;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy731;
					goto yy631;
				} else {
					if (yych == '`') goto yy631;
					goto yy937;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy631;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy732;
				goto yy733;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy734;
					goto yy735;
				} else {
					if (yych <= 0xF3) goto yy736;
					if (yych <= 0xF4) goto yy737;
					goto yy1;
				}
			}
		}
	}
yy938:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1;
			if (yych == '\'') goto yy721;
			goto yy631;
		} else {
			if (yych <= '>') goto yy730;
			if (yych == '\\') goto yy731;
			goto yy631;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy732;
			if (yych <= 0xE0) goto yy733;
			goto yy734;
		} else {
			if (yych <= 0xF0) goto yy735;
			if (yych <= 0xF3) goto yy736;
			if (yych <= 0xF4) goto yy737;
			goto yy1;
		}
	}
yy939:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy621;
				goto yy939;
			} else {
				if (yych == '\r') goto yy939;
				goto yy621;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy939;
				if (yych <= '!') goto yy621;
				goto yy721;
			} else {
				if (yych == '/') goto yy941;
				goto yy621;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '?') goto yy941;
				if (yych <= '[') goto yy621;
				goto yy722;
			} else {
				if (yych <= 0x7F) goto yy621;
				if (yych <= 0xC1) goto yy1;
				goto yy723;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy724;
				if (yych <= 0xEF) goto yy725;
				goto yy726;
			} else {
				if (yych <= 0xF3) goto yy727;
				if (yych <= 0xF4) goto yy728;
				goto yy1;
			}
		}
	}
yy940:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1;
				if (yych == '\t') goto yy964;
				goto yy621;
			} else {
				if (yych <= '\r') goto yy964;
				if (yych == ' ') goto yy964;
				goto yy621;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy721;
				if (yych == '-') goto yy940;
				goto yy621;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy940;
				} else {
					if (yych == '=') goto yy965;
					goto yy621;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy941;
				if (yych <= '@') goto yy621;
				if (yych <= 'Z') goto yy940;
				goto yy621;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy722;
					goto yy621;
				} else {
					if (yych == '`') goto yy621;
					goto yy940;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy621;
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) goto yy723;
				goto yy724;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy725;
					goto yy726;
				} else {
					if (yych <= 0xF3) goto yy727;
					if (yych <= 0xF4) goto yy728;
					goto yy1;
				}
			}
		}
	}
yy941:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1;
			if (yych == '"') goto yy721;
			goto yy621;
		} else {
			if (yych <= '>') goto yy720;
			if (yych == '\\') goto yy722;
			goto yy621;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) goto yy723;
			if (yych <= 0xE0) goto yy724;
			goto yy725;
		} else {
			if (yych <= 0xF0) goto yy726;
			if (yych <= 0xF3) goto yy727;
			if (yych <= 0xF4) goto yy728;
			goto yy1;
		}
	}
yy942:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy966;
yy943:
#line 184 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE_TIME);
       }
#line 31412 "../../lnav/src/data_scanner_re.cc"
yy944:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy967;
			}
			goto yy643;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= 0xC1) goto yy643;
			yyt2 = YYCURSOR;
			goto yy645;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy646;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy647;
			}
			yyt2 = YYCURSOR;
			goto yy648;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy649;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy650;
			}
			goto yy643;
		}
	}
yy945:
	yych = *++YYCURSOR;
	if (yych == 's') goto yy968;
	goto yy1;
yy946:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy969;
	goto yy1;
yy947:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy970;
	goto yy1;
yy948:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '-') goto yy972;
			if (yych <= '.') goto yy46;
			goto yy4;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy46;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy46;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy949:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy973;
					goto yy974;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy950:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy974;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy951:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				goto yy26;
			} else {
				if (yych == '\r') goto yy26;
				if (yych <= 0x1A) goto yy4;
				goto yy26;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy26;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy26;
				goto yy972;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '.') goto yy46;
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy340;
				goto yy4;
			} else {
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy46;
				goto yy26;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy26;
				if (yych <= 'z') goto yy46;
				goto yy26;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy26;
				if (yych <= 0xF4) goto yy4;
				goto yy26;
			}
		}
	}
yy952:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy975;
				goto yy923;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy975;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy975;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy953:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy976;
				goto yy882;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy976;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy976;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy954:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy977;
		if (yych <= ':') goto yy820;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy977;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy977;
		goto yy1;
	}
yy955:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy978;
		if (yych <= ':') goto yy927;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy978;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy978;
		goto yy1;
	}
yy956:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy979;
		if (yych <= ':') goto yy980;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy979;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy979;
		goto yy1;
	}
yy957:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy981;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy981;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy981;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy958:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy982;
			if (yych >= ';') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy982;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy982;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy959:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy981;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy981;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy981;
		goto yy1;
	}
yy960:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1;
			if (yych <= ':') goto yy882;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy961:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '0') {
			if (yych <= '/') goto yy1;
			goto yy983;
		} else {
			if (yych <= '1') goto yy822;
			if (yych <= '2') goto yy823;
			goto yy821;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy824;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy824;
			goto yy1;
		}
	}
yy962:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy631;
				goto yy962;
			} else {
				if (yych == '\r') goto yy962;
				goto yy631;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy962;
				if (yych <= '&') goto yy631;
				goto yy721;
			} else {
				if (yych == '/') goto yy938;
				if (yych <= '<') goto yy631;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych == '?') goto yy938;
				goto yy631;
			} else {
				if (yych <= '\\') goto yy731;
				if (yych <= 0x7F) goto yy631;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy732;
				if (yych <= 0xE0) goto yy733;
				goto yy734;
			} else {
				if (yych <= 0xF0) goto yy735;
				if (yych <= 0xF3) goto yy736;
				if (yych <= 0xF4) goto yy737;
				goto yy1;
			}
		}
	}
yy963:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1;
				goto yy525;
			} else {
				if (yych <= '\t') goto yy729;
				if (yych <= '\f') goto yy525;
				goto yy729;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy729;
				goto yy525;
			} else {
				if (yych <= '"') goto yy800;
				if (yych <= '=') goto yy525;
				goto yy631;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy525;
				goto yy632;
			} else {
				if (yych <= 0x7F) goto yy525;
				if (yych <= 0xC1) goto yy1;
				goto yy633;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy634;
				if (yych <= 0xEF) goto yy635;
				goto yy636;
			} else {
				if (yych <= 0xF3) goto yy637;
				if (yych <= 0xF4) goto yy638;
				goto yy1;
			}
		}
	}
yy964:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1;
				if (yych <= 0x08) goto yy621;
				goto yy964;
			} else {
				if (yych == '\r') goto yy964;
				goto yy621;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy964;
				if (yych <= '!') goto yy621;
				goto yy721;
			} else {
				if (yych == '/') goto yy941;
				if (yych <= '<') goto yy621;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych == '?') goto yy941;
				goto yy621;
			} else {
				if (yych <= '\\') goto yy722;
				if (yych <= 0x7F) goto yy621;
				goto yy1;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy723;
				if (yych <= 0xE0) goto yy724;
				goto yy725;
			} else {
				if (yych <= 0xF0) goto yy726;
				if (yych <= 0xF3) goto yy727;
				if (yych <= 0xF4) goto yy728;
				goto yy1;
			}
		}
	}
yy965:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1;
				goto yy524;
			} else {
				if (yych <= '\t') goto yy719;
				if (yych <= '\f') goto yy524;
				goto yy719;
			}
		} else {
			if (yych <= '&') {
				if (yych == ' ') goto yy719;
				goto yy524;
			} else {
				if (yych <= '\'') goto yy800;
				if (yych <= '=') goto yy524;
				goto yy621;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy524;
				goto yy622;
			} else {
				if (yych <= 0x7F) goto yy524;
				if (yych <= 0xC1) goto yy1;
				goto yy623;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy624;
				if (yych <= 0xEF) goto yy625;
				goto yy626;
			} else {
				if (yych <= 0xF3) goto yy627;
				if (yych <= 0xF4) goto yy628;
				goto yy1;
			}
		}
	}
yy966:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy984;
	goto yy1;
yy967:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '9') {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			goto yy643;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy642;
			}
			if (yych <= 0xC1) goto yy643;
			yyt2 = YYCURSOR;
			goto yy645;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy646;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy647;
			}
			yyt2 = YYCURSOR;
			goto yy648;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy649;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy650;
			}
			goto yy643;
		}
	}
yy968:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy985;
	goto yy100;
yy969:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy986;
	goto yy1;
yy970:
	yyaccept = 33;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy987;
yy971:
#line 181 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE_TIME);
       }
#line 32346 "../../lnav/src/data_scanner_re.cc"
yy972:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy988;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy988;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy988;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy973:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy989;
					goto yy990;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy974:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy990;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy975:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy680;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			goto yy1;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy923;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy976:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy991;
				goto yy882;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy991;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy991;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy977:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy354;
			goto yy466;
		} else {
			if (yych <= '/') goto yy354;
			if (yych <= '9') goto yy878;
			goto yy992;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy354;
			goto yy878;
		} else {
			if (yych <= '`') goto yy354;
			if (yych <= 'f') goto yy878;
			goto yy354;
		}
	}
yy978:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy927;
	goto yy1;
yy979:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy993;
		if (yych >= ';') goto yy1;
	} else {
		if (yych <= 'F') goto yy993;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy993;
		goto yy1;
	}
yy980:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy981;
		if (yych <= ':') goto yy830;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy981;
		if (yych <= '`') goto yy1;
		if (yych >= 'g') goto yy1;
	}
yy981:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy994;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy994;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy994;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy982:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy995;
			if (yych <= ':') goto yy959;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy995;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy995;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy983:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy881;
				goto yy996;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy883;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy883;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy984:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy997;
	goto yy1;
yy985:
	yych = *++YYCURSOR;
	if (yych == '(') goto yy998;
	goto yy1;
yy986:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy999;
	goto yy1;
yy987:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1000;
	goto yy1;
yy988:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy1001;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1001;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1001;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy989:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy109;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy109;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy109;
					if (yych <= 0x1F) goto yy4;
					goto yy109;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') goto yy4;
					goto yy105;
				} else {
					if (yych == '+') goto yy80;
					goto yy109;
				}
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy46;
					goto yy266;
				} else {
					if (yych <= '/') goto yy4;
					if (yych <= '7') goto yy1002;
					goto yy1003;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy109;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy109;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy109;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy109;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy109;
					if (yych <= 0xF4) goto yy4;
					goto yy109;
				}
			}
		}
	}
yy990:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '9') {
		if (yych <= '#') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy26;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy26;
				goto yy4;
			} else {
				if (yych <= 0x1A) {
					if (yych <= '\r') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0x1B) goto yy26;
					if (yych <= 0x1F) goto yy4;
					goto yy26;
				}
			}
		} else {
			if (yych <= '+') {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy105;
				if (yych <= '*') goto yy26;
				goto yy80;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy26;
					goto yy46;
				} else {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy4;
					goto yy1003;
				}
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'D') {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy26;
				if (yych <= '@') goto yy96;
				goto yy196;
			} else {
				if (yych <= 'F') {
					if (yych <= 'E') goto yy754;
					goto yy196;
				} else {
					if (yych <= 'Z') goto yy115;
					if (yych <= '^') goto yy26;
					goto yy46;
				}
			}
		} else {
			if (yych <= 'z') {
				if (yych <= 'd') {
					if (yych <= '`') goto yy26;
					goto yy196;
				} else {
					if (yych <= 'e') goto yy754;
					if (yych <= 'f') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy26;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy26;
					if (yych <= 0xF4) goto yy4;
					goto yy26;
				}
			}
		}
	}
yy991:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy680;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy1004;
				goto yy882;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy1004;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy1004;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy992:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy1005;
		if (yych <= ':') goto yy880;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy1005;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy1005;
		goto yy1;
	}
yy993:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy1006;
		if (yych <= ':') goto yy980;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy1006;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy1006;
		goto yy1;
	}
yy994:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1007;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1007;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy1007;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy995:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1;
			if (yych <= ':') goto yy959;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') goto yy1;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 'z') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy206;
			}
			yyt2 = YYCURSOR;
			goto yy207;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy208;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy209;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy210;
			}
			goto yy1;
		}
	}
yy996:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '0') {
			if (yych <= '/') goto yy1;
			goto yy1008;
		} else {
			if (yych <= '1') goto yy1009;
			if (yych <= '2') goto yy1010;
			goto yy1008;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1;
			goto yy928;
		} else {
			if (yych <= '`') goto yy1;
			if (yych <= 'f') goto yy928;
			goto yy1;
		}
	}
yy997:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1011;
	goto yy943;
yy998:
	yych = *++YYCURSOR;
	if (yych == 'x') goto yy1012;
	goto yy1;
yy999:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1013;
	goto yy1;
yy1000:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1014;
	goto yy1;
yy1001:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy1015;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1015;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1015;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1002:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) {
						yyt1 = YYCURSOR;
						goto yy1016;
					}
					yyt1 = YYCURSOR;
					goto yy1018;
				} else {
					if (yych <= '\n') {
						yyt1 = YYCURSOR;
						goto yy1016;
					}
					if (yych <= '\f') {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				} else {
					if (yych <= 0x1F) {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					if (yych <= ' ') {
						yyt1 = YYCURSOR;
						goto yy1019;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					yyt1 = YYCURSOR;
					goto yy1020;
				} else {
					if (yych == '+') {
						yyt1 = YYCURSOR;
						goto yy1020;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				}
			} else {
				if (yych <= '/') {
					if (yych <= '-') {
						yyt1 = YYCURSOR;
						goto yy1021;
					}
					if (yych <= '.') {
						yyt1 = YYCURSOR;
						goto yy1022;
					}
					yyt1 = YYCURSOR;
					goto yy1018;
				} else {
					if (yych <= '9') goto yy1024;
					if (yych <= ':') {
						yyt1 = YYCURSOR;
						goto yy1026;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				}
			}
		}
	} else {
		if (yych <= 'f') {
			if (yych <= 'Z') {
				if (yych <= 'D') {
					if (yych <= '@') {
						yyt1 = YYCURSOR;
						goto yy1027;
					}
					yyt1 = YYCURSOR;
					goto yy1028;
				} else {
					if (yych <= 'E') {
						yyt1 = YYCURSOR;
						goto yy1029;
					}
					if (yych <= 'F') {
						yyt1 = YYCURSOR;
						goto yy1028;
					}
					yyt1 = YYCURSOR;
					goto yy1030;
				}
			} else {
				if (yych <= '`') {
					if (yych == '_') {
						yyt1 = YYCURSOR;
						goto yy1021;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				} else {
					if (yych == 'e') {
						yyt1 = YYCURSOR;
						goto yy1029;
					}
					yyt1 = YYCURSOR;
					goto yy1028;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= '~') {
					if (yych <= 'z') {
						yyt1 = YYCURSOR;
						goto yy1030;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				} else {
					if (yych <= 0x7F) {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					if (yych <= 0xC1) goto yy1024;
					yyt1 = YYCURSOR;
					goto yy1031;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) {
						yyt1 = YYCURSOR;
						goto yy1032;
					}
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy1033;
					}
					yyt1 = YYCURSOR;
					goto yy1034;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy1035;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy1036;
					}
					goto yy1024;
				}
			}
		}
	}
yy1003:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '#') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) {
						yyt1 = YYCURSOR;
						goto yy1016;
					}
					yyt1 = YYCURSOR;
					goto yy1018;
				} else {
					if (yych <= '\n') {
						yyt1 = YYCURSOR;
						goto yy1016;
					}
					if (yych <= '\f') {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				} else {
					if (yych <= 0x1F) {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					if (yych <= ' ') {
						yyt1 = YYCURSOR;
						goto yy1019;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				}
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') {
					if (yych <= '$') {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					yyt1 = YYCURSOR;
					goto yy1020;
				} else {
					if (yych == '+') {
						yyt1 = YYCURSOR;
						goto yy1020;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				}
			} else {
				if (yych <= '/') {
					if (yych <= '-') {
						yyt1 = YYCURSOR;
						goto yy1021;
					}
					if (yych <= '.') {
						yyt1 = YYCURSOR;
						goto yy1022;
					}
					yyt1 = YYCURSOR;
					goto yy1018;
				} else {
					if (yych <= '9') goto yy1025;
					if (yych <= ':') {
						yyt1 = YYCURSOR;
						goto yy1026;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				}
			}
		}
	} else {
		if (yych <= 'f') {
			if (yych <= 'Z') {
				if (yych <= 'D') {
					if (yych <= '@') {
						yyt1 = YYCURSOR;
						goto yy1027;
					}
					yyt1 = YYCURSOR;
					goto yy1028;
				} else {
					if (yych <= 'E') {
						yyt1 = YYCURSOR;
						goto yy1029;
					}
					if (yych <= 'F') {
						yyt1 = YYCURSOR;
						goto yy1028;
					}
					yyt1 = YYCURSOR;
					goto yy1030;
				}
			} else {
				if (yych <= '`') {
					if (yych == '_') {
						yyt1 = YYCURSOR;
						goto yy1021;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				} else {
					if (yych == 'e') {
						yyt1 = YYCURSOR;
						goto yy1029;
					}
					yyt1 = YYCURSOR;
					goto yy1028;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= '~') {
					if (yych <= 'z') {
						yyt1 = YYCURSOR;
						goto yy1030;
					}
					yyt1 = YYCURSOR;
					goto yy1016;
				} else {
					if (yych <= 0x7F) {
						yyt1 = YYCURSOR;
						goto yy1018;
					}
					if (yych <= 0xC1) goto yy26;
					yyt1 = YYCURSOR;
					goto yy1031;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) {
						yyt1 = YYCURSOR;
						goto yy1032;
					}
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy1033;
					}
					yyt1 = YYCURSOR;
					goto yy1034;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy1035;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy1036;
					}
					goto yy26;
				}
			}
		}
	}
yy1004:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy680;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			goto yy1;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy882;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1005:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy1037;
		if (yych <= ':') goto yy927;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy1037;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy1037;
		goto yy1;
	}
yy1006:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy980;
	goto yy1;
yy1007:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy830;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy830;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy830;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1008:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy1038;
				goto yy959;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy958;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy958;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1009:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy1039;
				goto yy959;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy958;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy958;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1010:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '4') goto yy1039;
				goto yy1040;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy1038;
				if (yych <= ':') goto yy959;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy958;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy958;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1011:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1041;
	goto yy1;
yy1012:
	yych = *++YYCURSOR;
	if (yych == '8') goto yy1042;
	goto yy1;
yy1013:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1043;
	goto yy1;
yy1014:
	yyaccept = 33;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1044;
	goto yy971;
yy1015:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy1045;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1045;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1045;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1016:
	++YYCURSOR;
yy1017:
	YYCURSOR = yyt1;
#line 236 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_CREDIT_CARD_NUMBER);
           if (!this->is_credit_card(this->to_string_fragment(cap_all))) {
               if (cap_all.length() > 16) {
                   cap_all.c_end = cap_all.c_begin + 4;
                   cap_inner.c_end = cap_inner.c_begin + 4;
               }
               this->ds_next_offset = cap_all.c_end;
               token_out = DT_NUMBER;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 34170 "../../lnav/src/data_scanner_re.cc"
yy1018:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1017;
				goto yy4;
			} else {
				if (yych <= '\n') goto yy1017;
				if (yych <= '\f') goto yy4;
				goto yy1017;
			}
		} else {
			if (yych <= 0x1F) {
				if (yych == 0x1B) goto yy1017;
				goto yy4;
			} else {
				if (yych == '$') goto yy4;
				goto yy1017;
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= ':') goto yy4;
				if (yych <= '?') goto yy1017;
				goto yy4;
			} else {
				if (yych == '_') goto yy4;
				goto yy1017;
			}
		} else {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy4;
				if (yych <= '~') goto yy1017;
				goto yy4;
			} else {
				if (yych <= 0xC1) goto yy1017;
				if (yych <= 0xF4) goto yy4;
				goto yy1017;
			}
		}
	}
yy1019:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych == '%') goto yy177;
	goto yy1017;
yy1020:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy81;
			goto yy1017;
		} else {
			if (yych == ',') goto yy1017;
			if (yych <= '.') goto yy81;
			goto yy1017;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy81;
			if (yych <= '?') goto yy1017;
			if (yych <= 'Z') goto yy81;
			goto yy1017;
		} else {
			if (yych == '`') goto yy1017;
			if (yych <= 'z') goto yy81;
			goto yy1017;
		}
	}
yy1021:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[1024+yych] & 16) {
		goto yy46;
	}
	if (yych <= '%') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1017;
				goto yy4;
			} else {
				if (yych <= '\n') goto yy1017;
				if (yych <= '\f') goto yy4;
				goto yy1017;
			}
		} else {
			if (yych <= 0x1F) {
				if (yych == 0x1B) goto yy1017;
				goto yy4;
			} else {
				if (yych <= '#') goto yy1017;
				if (yych <= '$') goto yy4;
				goto yy80;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '+') {
				if (yych <= '*') goto yy1017;
				goto yy80;
			} else {
				if (yych <= ',') goto yy1017;
				if (yych <= ':') goto yy4;
				goto yy1017;
			}
		} else {
			if (yych <= 0x7F) {
				if (yych <= '@') goto yy96;
				if (yych <= '~') goto yy1017;
				goto yy4;
			} else {
				if (yych <= 0xC1) goto yy1017;
				if (yych <= 0xF4) goto yy4;
				goto yy1017;
			}
		}
	}
yy1022:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1B) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy1017;
				if (yych <= 0x08) goto yy4;
				goto yy1017;
			} else {
				if (yych == '\r') goto yy1017;
				if (yych <= 0x1A) goto yy4;
				goto yy1017;
			}
		} else {
			if (yych <= '%') {
				if (yych <= 0x1F) goto yy4;
				if (yych <= '#') goto yy1017;
				if (yych <= '$') goto yy4;
				goto yy80;
			} else {
				if (yych == '+') goto yy80;
				if (yych <= ',') goto yy1017;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy359;
				goto yy4;
			} else {
				if (yych <= '?') goto yy1017;
				if (yych <= '@') goto yy96;
				if (yych <= 'Z') goto yy46;
				goto yy1017;
			}
		} else {
			if (yych <= '~') {
				if (yych == '`') goto yy1017;
				if (yych <= 'z') goto yy46;
				goto yy1017;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy1017;
				if (yych <= 0xF4) goto yy4;
				goto yy1017;
			}
		}
	}
yy1023:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
yy1024:
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '?') {
		if (yych <= '$') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) goto yy109;
					goto yy3;
				} else {
					if (yych <= '\n') goto yy109;
					if (yych <= '\f') goto yy3;
					goto yy109;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy3;
					goto yy109;
				} else {
					if (yych <= 0x1F) goto yy3;
					if (yych <= '#') goto yy109;
					goto yy3;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '*') {
					if (yych <= '%') goto yy105;
					goto yy109;
				} else {
					if (yych <= '+') goto yy80;
					if (yych <= ',') goto yy109;
					goto yy46;
				}
			} else {
				if (yych <= '7') {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy3;
					goto yy1023;
				} else {
					if (yych <= '9') goto yy1025;
					if (yych <= ':') goto yy148;
					goto yy109;
				}
			}
		}
	} else {
		if (yych <= 'f') {
			if (yych <= 'Z') {
				if (yych <= 'D') {
					if (yych <= '@') goto yy96;
					goto yy196;
				} else {
					if (yych <= 'E') goto yy754;
					if (yych <= 'F') goto yy196;
					goto yy115;
				}
			} else {
				if (yych <= '`') {
					if (yych == '_') goto yy46;
					goto yy109;
				} else {
					if (yych == 'e') goto yy754;
					goto yy196;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= '~') {
					if (yych <= 'z') goto yy115;
					goto yy109;
				} else {
					if (yych <= 0x7F) goto yy3;
					if (yych <= 0xC1) goto yy109;
					goto yy56;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) goto yy57;
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy109;
				}
			}
		}
	}
yy1025:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 32) {
		goto yy103;
	}
	if (yych <= '@') {
		if (yych <= '$') {
			if (yych <= '\r') {
				if (yych <= 0x08) {
					if (yych <= 0x00) goto yy26;
					goto yy3;
				} else {
					if (yych <= '\n') goto yy26;
					if (yych <= '\f') goto yy3;
					goto yy26;
				}
			} else {
				if (yych <= 0x1B) {
					if (yych <= 0x1A) goto yy3;
					goto yy26;
				} else {
					if (yych <= 0x1F) goto yy3;
					if (yych <= '#') goto yy26;
					goto yy3;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych <= '*') {
					if (yych <= '%') goto yy105;
					goto yy26;
				} else {
					if (yych <= '+') goto yy80;
					if (yych <= ',') goto yy26;
					goto yy46;
				}
			} else {
				if (yych <= '9') {
					if (yych <= '.') goto yy266;
					if (yych <= '/') goto yy3;
					goto yy1025;
				} else {
					if (yych <= ':') goto yy148;
					if (yych <= '?') goto yy26;
					goto yy96;
				}
			}
		}
	} else {
		if (yych <= 'f') {
			if (yych <= '^') {
				if (yych <= 'E') {
					if (yych <= 'D') goto yy196;
					goto yy754;
				} else {
					if (yych <= 'F') goto yy196;
					if (yych <= 'Z') goto yy115;
					goto yy26;
				}
			} else {
				if (yych <= '`') {
					if (yych <= '_') goto yy46;
					goto yy26;
				} else {
					if (yych == 'e') goto yy754;
					goto yy196;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= '~') {
					if (yych <= 'z') goto yy115;
					goto yy26;
				} else {
					if (yych <= 0x7F) goto yy3;
					if (yych <= 0xC1) goto yy26;
					goto yy56;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) goto yy57;
					if (yych <= 0xEF) goto yy58;
					goto yy59;
				} else {
					if (yych <= 0xF3) goto yy60;
					if (yych <= 0xF4) goto yy61;
					goto yy26;
				}
			}
		}
	}
yy1026:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy189;
	if (yych == ':') goto yy152;
	goto yy1017;
yy1027:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy1017;
				if (yych <= 0x08) goto yy4;
				goto yy1017;
			} else {
				if (yych == '\r') goto yy1017;
				goto yy4;
			}
		} else {
			if (yych <= '#') {
				if (yych <= 0x1B) goto yy1017;
				if (yych <= 0x1F) goto yy4;
				goto yy1017;
			} else {
				if (yych <= '$') goto yy4;
				if (yych <= ',') goto yy1017;
				if (yych <= '.') goto yy174;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '?') {
				if (yych <= '9') goto yy174;
				if (yych <= ':') goto yy4;
				goto yy1017;
			} else {
				if (yych <= '@') goto yy4;
				if (yych <= 'Z') goto yy174;
				if (yych <= '^') goto yy1017;
				goto yy4;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy1017;
				if (yych <= 'z') goto yy174;
				goto yy1017;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy1017;
				if (yych <= 0xF4) goto yy4;
				goto yy1017;
			}
		}
	}
yy1028:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1017;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy1017;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy1017;
				if (yych == 0x1B) goto yy1017;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy1017;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy1017;
			} else {
				if (yych <= '+') goto yy80;
				if (yych <= ',') goto yy1017;
				if (yych <= '.') goto yy46;
				goto yy4;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '9') goto yy196;
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy1017;
				goto yy96;
			} else {
				if (yych <= 'F') goto yy196;
				if (yych <= 'Z') goto yy115;
				if (yych <= '^') goto yy1017;
				goto yy46;
			}
		} else {
			if (yych <= '~') {
				if (yych <= '`') goto yy1017;
				if (yych <= 'f') goto yy196;
				if (yych <= 'z') goto yy115;
				goto yy1017;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy1017;
				if (yych <= 0xF4) goto yy4;
				goto yy1017;
			}
		}
	}
yy1029:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= 0x1F) {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1017;
				if (yych <= 0x08) goto yy4;
				if (yych <= '\n') goto yy1017;
				goto yy4;
			} else {
				if (yych <= '\r') goto yy1017;
				if (yych == 0x1B) goto yy1017;
				goto yy4;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '#') goto yy1017;
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				goto yy1017;
			} else {
				if (yych <= '+') goto yy194;
				if (yych <= ',') goto yy1017;
				if (yych <= '-') goto yy251;
				goto yy46;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy196;
				if (yych <= ':') goto yy148;
				goto yy1017;
			} else {
				if (yych <= '@') goto yy96;
				if (yych <= 'F') goto yy196;
				if (yych <= 'Z') goto yy115;
				goto yy1017;
			}
		} else {
			if (yych <= 'z') {
				if (yych <= '_') goto yy46;
				if (yych <= '`') goto yy1017;
				if (yych <= 'f') goto yy196;
				goto yy115;
			} else {
				if (yych <= 0x7F) {
					if (yych <= '~') goto yy1017;
					goto yy4;
				} else {
					if (yych <= 0xC1) goto yy1017;
					if (yych <= 0xF4) goto yy4;
					goto yy1017;
				}
			}
		}
	}
yy1030:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[768+yych] & 64) {
		goto yy115;
	}
	if (yych <= '+') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy1017;
				if (yych <= 0x08) goto yy4;
				goto yy1017;
			} else {
				if (yych == '\r') goto yy1017;
				goto yy4;
			}
		} else {
			if (yych <= '#') {
				if (yych <= 0x1B) goto yy1017;
				if (yych <= 0x1F) goto yy4;
				goto yy1017;
			} else {
				if (yych <= '$') goto yy4;
				if (yych <= '%') goto yy80;
				if (yych <= '*') goto yy1017;
				goto yy80;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '/') {
				if (yych <= ',') goto yy1017;
				if (yych <= '.') goto yy46;
				goto yy4;
			} else {
				if (yych <= ':') goto yy148;
				if (yych <= '?') goto yy1017;
				goto yy96;
			}
		} else {
			if (yych <= '~') {
				if (yych == '_') goto yy46;
				goto yy1017;
			} else {
				if (yych <= 0x7F) goto yy4;
				if (yych <= 0xC1) goto yy1017;
				if (yych <= 0xF4) goto yy4;
				goto yy1017;
			}
		}
	}
yy1031:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy1018;
	goto yy1;
yy1032:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy1031;
	goto yy1;
yy1033:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy1031;
	goto yy1;
yy1034:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy1033;
	goto yy1;
yy1035:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy1033;
	goto yy1;
yy1036:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy1033;
	goto yy1;
yy1037:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy354;
			goto yy466;
		} else {
			if (yych <= '/') goto yy354;
			if (yych <= '9') goto yy955;
			goto yy1046;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy354;
			goto yy955;
		} else {
			if (yych <= '`') goto yy354;
			if (yych <= 'f') goto yy955;
			goto yy354;
		}
	}
yy1038:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy982;
				goto yy959;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy982;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy982;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1039:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy1047;
				goto yy959;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy982;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy982;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1040:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '5') goto yy1047;
				goto yy982;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy959;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 'F') goto yy982;
				if (yych <= 'Z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy982;
				goto yy1;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 0xC1) goto yy1;
				yyt2 = YYCURSOR;
				goto yy205;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy206;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1041:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1048;
	goto yy1;
yy1042:
	yych = *++YYCURSOR;
	if (yych == '6') goto yy1049;
	goto yy1;
yy1043:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1050;
	goto yy1;
yy1044:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1051;
	goto yy1;
yy1045:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '-') goto yy1052;
			if (yych <= '.') goto yy46;
			goto yy4;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy46;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy46;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1046:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy1053;
		if (yych <= ':') goto yy957;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy1053;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy1053;
		goto yy1;
	}
yy1047:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt1 = YYCURSOR;
				goto yy297;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy995;
				goto yy959;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy995;
			} else {
				if (yych <= 'Z') goto yy1;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy995;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy1;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1048:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1054;
	goto yy1;
yy1049:
	yych = *++YYCURSOR;
	if (yych == ')') goto yy99;
	goto yy1;
yy1050:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '9') {
			if (yych <= '/') {
				yyt1 = YYCURSOR;
				goto yy1016;
			}
			goto yy1;
		} else {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy1016;
			}
			if (yych <= 0xC1) goto yy1;
			yyt1 = YYCURSOR;
			goto yy1055;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt1 = YYCURSOR;
				goto yy1056;
			}
			if (yych <= 0xEF) {
				yyt1 = YYCURSOR;
				goto yy1057;
			}
			yyt1 = YYCURSOR;
			goto yy1058;
		} else {
			if (yych <= 0xF3) {
				yyt1 = YYCURSOR;
				goto yy1059;
			}
			if (yych <= 0xF4) {
				yyt1 = YYCURSOR;
				goto yy1060;
			}
			goto yy1;
		}
	}
yy1051:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1061;
	goto yy1;
yy1052:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy1062;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1062;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1062;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1053:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy1063;
		if (yych <= ':') goto yy980;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy1063;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy1063;
		goto yy1;
	}
yy1054:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy943;
	if (yych <= '9') goto yy1064;
	goto yy943;
yy1055:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy1016;
	goto yy1;
yy1056:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1;
	if (yych <= 0xBF) goto yy1055;
	goto yy1;
yy1057:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy1055;
	goto yy1;
yy1058:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1;
	if (yych <= 0xBF) goto yy1057;
	goto yy1;
yy1059:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0xBF) goto yy1057;
	goto yy1;
yy1060:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1;
	if (yych <= 0x8F) goto yy1057;
	goto yy1;
yy1061:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1;
	if (yych <= '9') goto yy1065;
	goto yy1;
yy1062:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy1066;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1066;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1066;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1063:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy354;
			goto yy466;
		} else {
			if (yych <= '/') goto yy354;
			if (yych <= '9') goto yy993;
			goto yy1067;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy354;
			goto yy993;
		} else {
			if (yych <= '`') goto yy354;
			if (yych <= 'f') goto yy993;
			goto yy354;
		}
	}
yy1064:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy943;
	if (yych <= '9') goto yy1068;
	goto yy943;
yy1065:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy971;
	if (yych <= '9') goto yy1069;
	goto yy971;
yy1066:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy1070;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1070;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1070;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1067:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1;
		if (yych <= '9') goto yy1071;
		if (yych <= ':') goto yy830;
		goto yy1;
	} else {
		if (yych <= 'F') goto yy1071;
		if (yych <= '`') goto yy1;
		if (yych <= 'f') goto yy1071;
		goto yy1;
	}
yy1068:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy943;
	if (yych <= '9') goto yy1072;
	goto yy943;
yy1069:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy971;
	if (yych <= '9') goto yy1073;
	goto yy971;
yy1070:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy1074;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1074;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1074;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1071:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= '9') goto yy1075;
			if (yych <= ':') goto yy1;
			yyt2 = YYCURSOR;
			goto yy198;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1075;
				goto yy1;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= 'f') goto yy1075;
				goto yy1;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy198;
			}
			if (yych <= 0xC1) goto yy1;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy205;
			}
			yyt2 = YYCURSOR;
			goto yy206;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy1;
			}
		}
	}
yy1072:
	++YYCURSOR;
	goto yy943;
yy1073:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy971;
	if (yych <= '9') goto yy1076;
	goto yy971;
yy1074:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= ',') goto yy4;
			if (yych <= '-') goto yy1077;
			if (yych <= '.') goto yy46;
			goto yy4;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy46;
			if (yych <= '?') goto yy4;
			if (yych <= '@') goto yy96;
			goto yy46;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy4;
				goto yy46;
			} else {
				if (yych <= '`') goto yy4;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1075:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				yyt2 = YYCURSOR;
				goto yy1078;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				if (yych <= '9') goto yy1007;
				goto yy466;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy1007;
			} else {
				if (yych <= 'Z') goto yy354;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy198;
				}
				goto yy1007;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy354;
				yyt2 = YYCURSOR;
				goto yy198;
			} else {
				if (yych <= 0xC1) goto yy354;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy205;
				}
				yyt2 = YYCURSOR;
				goto yy206;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy207;
				}
				yyt2 = YYCURSOR;
				goto yy208;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy209;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy210;
				}
				goto yy354;
			}
		}
	}
yy1076:
	++YYCURSOR;
	goto yy971;
yy1077:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych <= '9') goto yy1079;
				goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1079;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1079;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1078:
	yyaccept = 17;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy199;
		if (yych <= '9') goto yy542;
		goto yy199;
	} else {
		if (yych <= 'F') goto yy542;
		if (yych <= '`') goto yy199;
		if (yych <= 'f') goto yy542;
		goto yy199;
	}
yy1079:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1080;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1080;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1080:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1081;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1081;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1081:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1082;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1082;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1082:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1083;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1083;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1083:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1084;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1084;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1084:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1085;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1085;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1085:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1086;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1086;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1086:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1087;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1087;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1087:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1088;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1088;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1088:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1089;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1089;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1089:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy80;
			if (yych <= '*') goto yy4;
			goto yy80;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy4;
				goto yy46;
			} else {
				if (yych <= '/') goto yy4;
				if (yych >= ':') goto yy4;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy96;
			if (yych <= 'F') goto yy1090;
			if (yych <= 'Z') goto yy46;
			goto yy4;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy46;
				goto yy4;
			} else {
				if (yych <= 'f') goto yy1090;
				if (yych <= 'z') goto yy46;
				goto yy4;
			}
		}
	}
yy1090:
	yyaccept = 35;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm[1024+yych] & 16) {
		goto yy46;
	}
	if (yych <= '%') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych >= 0x01) goto yy4;
			} else {
				if (yych <= '\n') goto yy1091;
				if (yych <= '\f') goto yy4;
			}
		} else {
			if (yych <= 0x1F) {
				if (yych != 0x1B) goto yy4;
			} else {
				if (yych <= '#') goto yy1091;
				if (yych <= '$') goto yy4;
				goto yy80;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '+') {
				if (yych >= '+') goto yy80;
			} else {
				if (yych <= ',') goto yy1091;
				if (yych <= ':') goto yy4;
			}
		} else {
			if (yych <= 0x7F) {
				if (yych <= '@') goto yy96;
				if (yych >= 0x7F) goto yy4;
			} else {
				if (yych <= 0xC1) goto yy1091;
				if (yych <= 0xF4) goto yy4;
			}
		}
	}
yy1091:
#line 234 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_UUID); }
#line 36093 "../../lnav/src/data_scanner_re.cc"
}
#line 276 "../../lnav/src/data_scanner_re.re"

}
